#include "DlgProcs.h"


void UpdateBitmap(HWND hDlg, SPECTRUMPARAM sp);


BOOL CALLBACK InverseFFTDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	static PSPECTRUMPARAM  pSP;
	static float           fSource;
	BITMAP                 bitmap;
	static long            SizeBMP;
	static BYTE            *pBits, *pMemBits;

	switch (message)
		{
		case WM_INITDIALOG:
			pSP = (PSPECTRUMPARAM) lParam;
			GetObject(pSP->hBitmap, sizeof(BITMAP), &bitmap);
			SizeBMP = bitmap.bmWidthBytes * bitmap.bmHeight;
			pBits = (BYTE*) bitmap.bmBits;
			GetObject(pSP->hSpectBMP, sizeof(BITMAP), &bitmap);
			pMemBits = (BYTE*) bitmap.bmBits;
			CopyMemory(pMemBits, pBits, SizeBMP);
			CheckDlgButton(hDlg, IDC_OPT_REAL, BST_CHECKED);
			SetDlgItemInt(hDlg, IDC_ADDOFFSET, 0, FALSE);
			SetDlgItemInt(hDlg, IDC_MULTIPLY, 1, FALSE);
			return TRUE;

		case WM_COMMAND:
			switch (LOWORD (wParam))
			{
			case IDC_OPT_ABS:
			case IDC_OPT_PHASE:
			case IDC_OPT_REAL:
			case IDC_OPT_IMAG:
			case IDC_CHK_SIGNABS:
				UpdateBitmap(hDlg, *pSP);
				return TRUE;

			case IDC_CHK_NORMALIZE:
				EnableWindow(GetDlgItem(hDlg, IDC_MULTIPLY), IsDlgButtonChecked(hDlg, IDC_CHK_NORMALIZE)!=BST_CHECKED);
				UpdateBitmap(hDlg, *pSP);
				return TRUE;

			case IDC_ADDOFFSET:
			case IDC_MULTIPLY:
				if (HIWORD(wParam)==EN_CHANGE)
					UpdateBitmap(hDlg, *pSP);
				return TRUE;

			case IDOK:
				GetObject(pSP->hBitmap, sizeof(BITMAP), &bitmap);
				SaveBitsForUndo(pMemBits, bitmap.bmWidth, bitmap.bmWidthBytes, bitmap.bmHeight, TEXT("Spektrum bearbeiten"));
				UpdateBitmap(hDlg, *pSP);
				EndDialog(hDlg, TRUE);
				return TRUE;

			case IDCANCEL:
				CopyMemory(pBits, pMemBits, SizeBMP);
				InvalidateRect(GetParent(hDlg), NULL, FALSE);
				EndDialog (hDlg, FALSE);
				return TRUE;
			}
			break;
	}
	return FALSE;
}




void UpdateBitmap(HWND hDlg, SPECTRUMPARAM sp)
{
	BITMAP        bitmap;
	BYTE          *pBits;
	long          x, y;
	unsigned long Offset;
	float         fMax, fAddOffset, fMultiply;
	TCHAR         szBuffer[80], *szEndConv;
	BOOL          bTemp;

	GetObject(sp.hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	bTemp = TRUE;
	GetDlgItemText(hDlg, IDC_ADDOFFSET, szBuffer, 80);
	fAddOffset = (float) strtod(szBuffer, &szEndConv);
	if (fabs(fAddOffset)==HUGE_VAL || strlen(szEndConv) != 0)
		bTemp = FALSE;
	GetDlgItemText(hDlg, IDC_MULTIPLY, szBuffer, 80);
	fMultiply = (float) strtod(szBuffer, &szEndConv);
	if (fabs(fMultiply)==HUGE_VAL || strlen(szEndConv) != 0)
		bTemp = FALSE;
	EnableWindow(GetDlgItem(hDlg, IDOK), bTemp);
	if (!bTemp)
	{
		return;
	}

	if (IsDlgButtonChecked(hDlg, IDC_OPT_PHASE)==BST_CHECKED)
	{
		for (y=sp.OffY; y<bitmap.bmHeight+sp.OffY; y++)
		{
			Offset = y*sp.Width+sp.OffX;
			for (x=0; x<bitmap.bmWidth; x++)
			{
				sp.Floats[Offset] = atan2(sp.SpectIm[Offset], sp.SpectRe[Offset]) + (float) Pi;
				Offset++;
			}
		}
	}
	else if (IsDlgButtonChecked(hDlg, IDC_OPT_REAL)==BST_CHECKED)
	{
		for (y=sp.OffY; y<bitmap.bmHeight+sp.OffY; y++)
		{
			Offset = y*sp.Width+sp.OffX;
			for (x=0; x<bitmap.bmWidth; x++)
			{
				sp.Floats[Offset] = sp.SpectRe[Offset];
				Offset++;
			}
		}
	}
	else if (IsDlgButtonChecked(hDlg, IDC_OPT_IMAG)==BST_CHECKED)
	{
		for (y=sp.OffY; y<bitmap.bmHeight+sp.OffY; y++)
		{
			Offset = y*sp.Width+sp.OffX;
			for (x=0; x<bitmap.bmWidth; x++)
			{
				sp.Floats[Offset] = sp.SpectIm[Offset];
				Offset++;
			}
		}
	}
	else
	{
		for (y=sp.OffY; y<bitmap.bmHeight+sp.OffY; y++)
		{
			Offset = y*sp.Width+sp.OffX;
			for (x=0; x<bitmap.bmWidth; x++)
			{
			  sp.Floats[Offset] = (float) hypot(sp.SpectRe[Offset], sp.SpectIm[Offset]);
				Offset++;
			}
		}
	}

	// Offset Addieren...
	if (fAddOffset!=0)
	{
		for (y=sp.OffY; y<bitmap.bmHeight+sp.OffY; y++)
		{
			Offset = y*sp.Width+sp.OffX;
			for (x=0; x<bitmap.bmWidth; x++)
			{
			  sp.Floats[Offset] += fAddOffset;
				Offset++;
			}
		}
	}


	// Betrag bilden?
	if (IsDlgButtonChecked(hDlg, IDC_CHK_SIGNABS))
	{
		for (y=sp.OffY; y<bitmap.bmHeight+sp.OffY; y++)
		{
			Offset = y*sp.Width+sp.OffX;
			for (x=0; x<bitmap.bmWidth; x++)
			{
			  sp.Floats[Offset] = fabs(sp.Floats[Offset]);
				Offset++;
			}
		}
	}

	// Normalisieren?
	if (IsDlgButtonChecked(hDlg, IDC_CHK_NORMALIZE)==BST_CHECKED)
	{
		fMax = sp.Floats[0];
		for (y=sp.OffY; y<bitmap.bmHeight+sp.OffY; y++)
		{
			Offset = y*sp.Width+sp.OffX;
			for (x=0; x<bitmap.bmWidth; x++)
			{
				if (sp.Floats[Offset]>fMax)  fMax = sp.Floats[Offset];
				Offset++;
			}
		}
		if (fMax>0)
		  fMax = (float) 255.0 / fMax;
		else
		  fMax = 255.0;
	}
	else
	{
		// Multiplizieren?
		if (fMultiply!=1)
		{
			for (y=sp.OffY; y<bitmap.bmHeight+sp.OffY; y++)
			{
				Offset = y*sp.Width+sp.OffX;
				for (x=0; x<bitmap.bmWidth; x++)
				{
					sp.Floats[Offset] *= fMultiply;
					Offset++;
				}
			}
		}
    fMax = 1;
	}

	// Bitmap schreiben und darstellen...
	for (y=0; y<bitmap.bmHeight; y++)
	{
		Offset = (y+sp.OffY)*sp.Width+sp.OffX;
		for (x=0; x<bitmap.bmWidth; x++)
		{
			if (sp.Floats[Offset]>=0)
				if (sp.Floats[Offset]*fMax<256)
					pBits[y*bitmap.bmWidthBytes+x] = BYTE(sp.Floats[Offset] * fMax);
				else
					pBits[y*bitmap.bmWidthBytes+x] = 255;
			else
				pBits[y*bitmap.bmWidthBytes+x] = 0;
			Offset++;
		}
	}
	InvalidateRect(GetParent(hDlg), NULL, FALSE);
}