#include "Numerical.h"
#include "DIBs.h"


BOOL CalcBitOp(HBITMAP hBitmap, WORD Operation, BYTE Operand)
{
	BITMAP bitmap;
	long x, y, Offset;
	BYTE *pBits, BitNr, tmpByte;

	GetObject (hBitmap, sizeof (BITMAP), &bitmap); 
	pBits= (BYTE*) bitmap.bmBits;

	switch (Operation)
	{
	case IDM_BITOP_NOT:
		   for (y = 0; y < bitmap.bmHeight; y++)
			 {
				 Offset = y * bitmap.bmWidthBytes;
				 for (x = 0; x < bitmap.bmWidth; x++)
				 {
					 pBits[Offset] = ~pBits[Offset];
					 Offset++;
				 }
			 }
			 return TRUE;

	case IDM_BITOP_AND:
		   for (y = 0; y < bitmap.bmHeight; y++)
			 {
				 Offset = y * bitmap.bmWidthBytes;
				 for (x = 0; x < bitmap.bmWidth; x++)
				 {
					 pBits[Offset] = pBits[Offset] & Operand;
					 Offset++;
				 }
			 }
			 return TRUE;

	case IDM_BITOP_OR:
		   for (y = 0; y < bitmap.bmHeight; y++)
			 {
				 Offset = y * bitmap.bmWidthBytes;
				 for (x = 0; x < bitmap.bmWidth; x++)
				 {
					 pBits[Offset] = pBits[Offset] | Operand;
					 Offset++;
				 }
			 }
			 return TRUE;

	case IDM_BITOP_XOR:
		   for (y = 0; y < bitmap.bmHeight; y++)
			 {
				 Offset = y * bitmap.bmWidthBytes;
				 for (x = 0; x < bitmap.bmWidth; x++)
				 {
					 pBits[Offset] = pBits[Offset] ^ Operand;
					 Offset++;
				 }
			 }
			 return TRUE;

 	case IDM_BITOP_RL:
		   for (y = 0; y < bitmap.bmHeight; y++)
			 {
				 Offset = y * bitmap.bmWidthBytes;
				 for (x = 0; x < bitmap.bmWidth; x++)
				 {
					 pBits[Offset] = (pBits[Offset] << 1) + (pBits[Offset] >> 7);
					 Offset++;
				 }
			 }
			 return TRUE;

 	case IDM_BITOP_RR:
		   for (y = 0; y < bitmap.bmHeight; y++)
			 {
				 Offset = y * bitmap.bmWidthBytes;
				 for (x = 0; x < bitmap.bmWidth; x++)
				 {
					 pBits[Offset] = (pBits[Offset] >> 1) + (pBits[Offset] << 7);
					 Offset++;
				 }
			 }
			 return TRUE;

 	case IDM_BITOP_SL:
		   for (y = 0; y < bitmap.bmHeight; y++)
			 {
				 Offset = y * bitmap.bmWidthBytes;
				 for (x = 0; x < bitmap.bmWidth; x++)
				 {
					 pBits[Offset] = pBits[Offset] << 1;
					 Offset++;
				 }
			 }
			 return TRUE;

 	case IDM_BITOP_SR:
		   for (y = 0; y < bitmap.bmHeight; y++)
			 {
				 Offset = y * bitmap.bmWidthBytes;
				 for (x = 0; x < bitmap.bmWidth; x++)
				 {
					 pBits[Offset] = pBits[Offset] >> 1;
					 Offset++;
				 }
			 }
			 return TRUE;

 	case IDM_BITOP_BI:
		   for (y = 0; y < bitmap.bmHeight; y++)
			 {
				 Offset = y * bitmap.bmWidthBytes;
				 for (x = 0; x < bitmap.bmWidth; x++)
				 {
					 tmpByte = pBits[Offset];
					 pBits[Offset] = 0;
					 for (BitNr=0; BitNr<8; BitNr++)
					   pBits[Offset] += ((tmpByte >> BitNr) & 1) << (7-BitNr);
					 Offset++;
				 }
			 }
			 return TRUE;

	default:
		   return FALSE;
	}
	
}



BOOL CalcNormalize(HBITMAP hBitmap)
{
	BITMAP bitmap;
	long x, y, Offset;
	BYTE Min, Max, *pBits;
	double f;

	if (!hBitmap)  return FALSE;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	Min = 255;
	Max = 0;
	for (y=0; y<bitmap.bmHeight; y++)
	{
		Offset = y*bitmap.bmWidthBytes;
		for (x=0; x<bitmap.bmWidth; x++)
		{
			if (pBits[Offset]>Max)  Max=pBits[Offset];
			if (pBits[Offset]<Min)  Min=pBits[Offset];
			Offset++;
		}
	}
	if (Min!=Max)
	{
		f=255.0/(Max-Min);
		for (y=0; y<bitmap.bmHeight; y++)
		{
			Offset = y*bitmap.bmWidthBytes;
			for (x=0; x<bitmap.bmWidth; x++)
			{
				pBits[Offset] = BYTE((pBits[Offset]-Min)*f);
				Offset++;
			}
		}
		return TRUE;
	}
	return FALSE;
}



#define SWAP(a,b) temp=(a);(a)=(b);(b)=temp

int SelectKthOrder(int arr[], unsigned int n, unsigned int k)
{
	unsigned int i, ir, j, l, mid;
	int a, temp;

	l = 1;
	ir = n;
	for(;;)
	{
		if (ir <= l+1)
		{
			if (ir == l+1 && arr[ir] < arr[l])
			{
				SWAP(arr[l], arr[ir]);
			}
			return arr[k];
		}
		else
		{
			mid = (l+ir) >> 1;
			SWAP(arr[mid], arr[l+1]);
			if (arr[l] > arr[ir])
			{
				SWAP(arr[l], arr[ir]);
			}
			if (arr[l+1]>arr[ir])
			{
				SWAP(arr[l+1], arr[ir]);
			}
			if (arr[l] > arr[l+1])
			{
				SWAP(arr[l], arr[l+1]);
			}
			i = l+1;
			j = ir;
			a = arr[l+1];
			for(;;)
			{
				do i++; while (arr[i] < a);
				do j--; while (arr[j] > a);
				if (j < i) break;
				SWAP(arr[i], arr[j]);
			}
			arr[l+1] = arr[j];
			arr[j] = a;
			if (j >= k)  ir = j-1;
			if (j <= k)   l = i;
		}
	}
}


BOOL CalcLocalRankOrder(HBITMAP hBitmap, int Width, int Height, int Rank)
{
	BITMAP bitmap;
	int    x, y, i, il, ir, j, jl, jr, n, m, Offset, *pArray;
	BYTE   *pBits, *pMemBits;

	if (Width<1  ||  Height<1)  return FALSE;
	if (Width<3  &&  Height<3)  return FALSE;
	if ((Width-1)%2  ||  (Height-1)%2)  return FALSE;
	if (Rank<1  ||  Rank>Width*Height)  return FALSE;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	if (Width>bitmap.bmWidth  ||  Height>bitmap.bmHeight)  return FALSE;
	pBits = (BYTE*) bitmap.bmBits;
	pMemBits = (BYTE*) malloc(bitmap.bmWidthBytes*bitmap.bmHeight);
	if (!pMemBits)  return FALSE;
	pArray = (int*) malloc(sizeof(int)*Width*Height);
	if (!pArray)
	{
		free(pMemBits);
		return FALSE;
	}
	CopyMemory(pMemBits, pBits, bitmap.bmWidthBytes*bitmap.bmHeight);

  for (y = 0; y < bitmap.bmHeight; y++)
	{
		for (x = 0; x < bitmap.bmWidth; x++)
		{
			// Kopieren pMemBits in pArray
			Offset = 0;
			jl = y-Height/2;
			jr = y+Height/2;
			for (j = jl; j <= jr; j++)
			{
				m=abs(j);
				if (m>=bitmap.bmHeight)  m = (bitmap.bmHeight-2) - (m-bitmap.bmHeight);
				m *= bitmap.bmWidthBytes;
				il = x-Width/2;
				ir = x+Width/2;
				for (i = il; i <= ir; i++)
				{
					n=abs(i);
					if (n>=bitmap.bmWidth)  n = (bitmap.bmWidth-2) - (n-bitmap.bmWidth);
					pArray[Offset] = pMemBits[m+n];
					Offset++;
				}
			}
			// Sortieren in pArray, dann in pBits speichern
			pBits[y*bitmap.bmWidthBytes+x] = SelectKthOrder(pArray-1, Width*Height, Rank);
		}
	}
	return TRUE;
}




BOOL LoadLinFilter(HWND hwnd, LINFILTER *LinFilter)
{
	HANDLE hFile;
	char   *szBuf, *szTemp, *szEndPtr;
	DWORD  dwLength, Offset, Counter;
	int    x, y, i, j;
	float  CoeffSum;
	BOOL   bFilterOk, bTmp, bDivByGiven;

	LinFilter->FilterMatrix = NULL;
	LinFilter->FilterWidth = 0;
	LinFilter->FilterHeight = 0;
	LinFilter->BorderParam = 0;
	LinFilter->DivBy = 1;
	LinFilter->AddOffset = 0;
	LinFilter->SignParam = 0;
	LinFilter->NormalizeParam = 0;
	LinFilter->szFilterName[0] = '\0';
	hFile = CreateFile(LinFilter->szFilterFile, GENERIC_READ, 0, NULL, OPEN_EXISTING,
		                 FILE_SHARE_READ, NULL);
	if (hFile == INVALID_HANDLE_VALUE)  return FALSE;
	Offset = GetFileSize(hFile, NULL);
	if (Offset == 0xFFFFFFFF)   // Fehler?
	{
		CloseHandle(hFile);
		return FALSE;
	}
	if (Offset<80)  Offset=80;
	szBuf = (char*) malloc(Offset+1);
	if (!szBuf)
	{
		CloseHandle(hFile);
		return FALSE;
	}
	szTemp = (char*) malloc(Offset+1);
	if (!szBuf)
	{
		CloseHandle(hFile);
		free(szBuf);
		return FALSE;
	}
	if (!ReadFile(hFile, szTemp, Offset, &dwLength, NULL))
	{
		CloseHandle(hFile);
		free(szBuf);
		free(szTemp);
		return FALSE;
	}
	CloseHandle(hFile);
	if (dwLength != Offset)
	{
		free(szBuf);
		free(szTemp);
		return FALSE;
	}
	Counter=0;
	Offset=0;
	bTmp = TRUE;
	while (Counter<dwLength)
	{
		if (szTemp[Counter]=='/'  &&  Counter<dwLength-1)
		{
			if (szTemp[Counter+1]=='*')
			{
				bTmp = FALSE;
				szBuf[Offset] = ' ';
				Offset++;
				Counter++;
			}
		}
		if (bTmp)
		{
			szBuf[Offset] = szTemp[Counter];
			Offset++;
		}
		if (!bTmp  &&  szTemp[Counter]=='*'  &&  Counter<dwLength-1)
		{
			if (szTemp[Counter+1]=='/')
			{
				bTmp = TRUE;
				Counter++;
			}
		}
		Counter++;
	}
	dwLength = Offset;
	szBuf[dwLength] = '\0';
	if (strncmp(szBuf, TEXT("Particle Tracer - Linear Filter:"), 32) != 0)
	{
		MessageBox(hwnd, TEXT("Datei wurde nicht eindeutig als lineares Filter von Particle Tracer erkannt. Die Kennung (in der ersten Zeile) fehlt!"), TEXT("Lineares Filter - FEHLER"), MB_ICONERROR | MB_OK);
		free(szBuf);
		free(szTemp);
		return FALSE;
	}
	Offset = 32;
	i = 0;
	while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t')  Offset++;
	while (szBuf[Offset]!='\r' &&  szBuf[Offset]!='\n')
	{
		if (i<76)
		{
			LinFilter->szFilterName[i]=szBuf[Offset];
			i++;
		}
		else
		{
			LinFilter->szFilterName[76]='.';
			LinFilter->szFilterName[77]='.';
			LinFilter->szFilterName[78]='.';
			i = 79;
		}
		Offset++;
		if (Offset>=dwLength)
		{
			free(szBuf);
			free(szTemp);
			return FALSE;
		}
	}
	LinFilter->szFilterName[i]='\0';
	bFilterOk = FALSE;
	bDivByGiven = FALSE;
	while (Offset<dwLength)
	{
		while ((szBuf[Offset]=='\r'  ||  szBuf[Offset]=='\n'  ||  szBuf[Offset]==' '  ||  szBuf[Offset]=='\t')  &&  Offset<dwLength)
			Offset++;
		i = 0;
		while (szBuf[Offset]!='\r'  &&  szBuf[Offset]!='\n'  &&  szBuf[Offset]!=' '  &&  szBuf[Offset]!='\t'  &&  Offset<dwLength)
		{
      szTemp[i]=szBuf[Offset];
			i++;
			Offset++;
			if (szTemp[i-1]==':')  break;
		}
		szTemp[i]='\0';
		if (i==0)  break;
		if (strncmp(szTemp, TEXT("Matrix:"), strlen(szTemp)) == 0)
		{
      while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t'  ||  szBuf[Offset]=='\r'  ||  szBuf[Offset]=='\n')
			{
				Offset++;
				if (Offset>=dwLength)
				{
					MessageBox(hwnd, TEXT("Definition der Matrix fehlt!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			i = 0;
      while (szBuf[Offset]>='0'  &&  szBuf[Offset]<='9')
			{
				szTemp[i]=szBuf[Offset];
				i++;
				Offset++;
				if (Offset>=dwLength)
				{
					MessageBox(hwnd, TEXT("Definition der Matrix unvollstndig!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			szTemp[i]='\0';
			LinFilter->FilterHeight = atoi(szTemp);
			if (LinFilter->FilterHeight<1  ||  LinFilter->FilterHeight%2 != 1)
			{
				MessageBox(hwnd, TEXT("Hhe der Matrix ungltig!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
			bTmp = FALSE;
      while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t'  ||  szBuf[Offset]=='*'  ||  szBuf[Offset]=='X'  ||  szBuf[Offset]=='x')
			{
				if (szBuf[Offset]=='*'  ||  szBuf[Offset]=='X'  ||  szBuf[Offset]=='x')
				{
					if (bTmp)
					{
						MessageBox(hwnd, TEXT("Definition der Dimension der Matrix ungltig!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
						free(szBuf);
						free(szTemp);
						return FALSE;
					}
					else
						bTmp = TRUE;
				}
				Offset++;
				if (Offset>=dwLength)
				{
					MessageBox(hwnd, TEXT("Definition der Matrix unvollstndig!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			i = 0;
      while (szBuf[Offset]>='0'  &&  szBuf[Offset]<='9')
			{
				szTemp[i]=szBuf[Offset];
				i++;
				Offset++;
				if (Offset>=dwLength)
				{
					MessageBox(hwnd, TEXT("Definition der Matrix unvollstndig!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			szTemp[i]='\0';
			LinFilter->FilterWidth = atoi(szTemp);
			if (LinFilter->FilterWidth<1  ||  LinFilter->FilterWidth%2 != 1)
			{
				MessageBox(hwnd, TEXT("Breite der Matrix ungltig!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
			if (!(LinFilter->FilterMatrix = (float*) malloc(LinFilter->FilterHeight * LinFilter->FilterWidth * sizeof(float))))
			{
				MessageBox(hwnd, TEXT("Nicht geggend Speicher fr die Matrix!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
			bTmp = FALSE;
			while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t'  ||  szBuf[Offset]=='\r'  ||  szBuf[Offset]=='\n'  ||  szBuf[Offset]==':')
			{
				if (szBuf[Offset]==':')
				{
					if (bTmp)
					{
						MessageBox(hwnd, TEXT("Definition der Matrix ungltig!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
						free(szBuf);
						free(szTemp);
						return FALSE;
					}
					else
						bTmp = TRUE;
				}
				Offset++;
				if (Offset>=dwLength)
				{
					MessageBox(hwnd, TEXT("Definition der Matrix fehlt!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			for (y=LinFilter->FilterHeight-1; y>=0; y--)
			{
				for (x=0; x<LinFilter->FilterWidth; x++)
				{
					LinFilter->FilterMatrix[y*LinFilter->FilterWidth+x] = (float) strtod(szBuf+Offset, &szEndPtr);
					if (szEndPtr==szBuf+Offset  ||  fabs(LinFilter->FilterMatrix[y*LinFilter->FilterWidth+x])==HUGE_VAL)
					{
						MessageBox(hwnd, TEXT("Definition der Matrix ungltig!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
						free(szBuf);
						free(szTemp);
						return FALSE;
					}
          Offset = szEndPtr-szBuf;
					while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t')
					{
						Offset++;
						if (Offset>=dwLength)
						{
							MessageBox(hwnd, TEXT("Definition der Matrix fehlt!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
							free(szBuf);
							free(szTemp);
							return FALSE;
						}
					}
					if ((szBuf[Offset]=='\r'  ||  szBuf[Offset]=='\n')  &&  x<LinFilter->FilterWidth-1)
					{
						MessageBox(hwnd, TEXT("Definition der Matrix ungltig! Return vor Ende einer Zeile gefunden."), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
						free(szBuf);
						free(szTemp);
						return FALSE;
					}
				}
				if (szBuf[Offset]!='\r'  &&  szBuf[Offset]!='\n')
				{
					MessageBox(hwnd, TEXT("Definition der Matrix ungltig! Erwarte Return am Ende jeder Zeile."), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
				if (szBuf[Offset]=='\r')  Offset++;
				if (szBuf[Offset]=='\n')  Offset++;
				if (y>0  &&  (szBuf[Offset]=='\r'  ||  szBuf[Offset]=='\n'))
				{
					MessageBox(hwnd, TEXT("Erwarte nchste Zeile der Matrix! Stattdessen Return gefunden."), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			bFilterOk = TRUE;
		}
		else if (strncmp(szTemp, TEXT("Border:"), strlen(szTemp)) == 0)
		{
			while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t')
			{
				Offset++;
				if (Offset>=dwLength)
				{
					MessageBox(hwnd, TEXT("Parameter zu 'Border' fehlt!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			i = 0;
			while (szBuf[Offset]!=' '  &&  szBuf[Offset]!='\t'  &&  szBuf[Offset]!='\r'  &&  szBuf[Offset]!='\n')
			{
				szTemp[i] = szBuf[Offset];
				i++;
				Offset++;
				if (Offset>=dwLength)
				{
					MessageBox(hwnd, TEXT("Parameter zu 'Border' fehlt!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			if (i==0)
			{
				MessageBox(hwnd, TEXT("Parameter zu 'Border' fehlt!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
			szTemp[i] = '\0';
			if (strncmp(szTemp, TEXT("FlipInc"), strlen(szTemp)) == 0)
			{
				LinFilter->BorderParam = -1;
			}
			else if (strncmp(szTemp, TEXT("FlipExc"), strlen(szTemp)) == 0)
			{
				LinFilter->BorderParam = -2;
			}
			else if (strncmp(szTemp, TEXT("Periodic"), strlen(szTemp)) == 0)
			{
				LinFilter->BorderParam = -3;
			}
			else if (strncmp(szTemp, TEXT("LocalMean"), strlen(szTemp)) == 0)
			{
				LinFilter->BorderParam = -4;
			}
			else if (strncmp(szTemp, TEXT("CoeffSum"), strlen(szTemp)) == 0)
			{
				LinFilter->BorderParam = -5;
			}
			else if (strncmp(szTemp, TEXT("FillWith"), strlen(szTemp)) == 0)
			{
				while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t')
				{
					Offset++;
					if (Offset>=dwLength)
					{
						MessageBox(hwnd, TEXT("Parameter zu 'Border: FillWith' fehlt (Grauwert 0-255  oder  'Avg')!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
						free(szBuf);
						free(szTemp);
						return FALSE;
					}
				}
				if (strncmp(szBuf+Offset, TEXT("Avg"), 3) == 0)
				{
					LinFilter->BorderParam = 256;
					Offset += 3;
				}
				else
				{
					if (szBuf[Offset]<'0'  ||  szBuf[Offset]>'9')
					{
						MessageBox(hwnd, TEXT("Parameter zu 'Border: FillWith' fehlerhaft! Erwarte ganze Zahl (0-255) oder 'Avg'."), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
						free(szBuf);
						free(szTemp);
						return FALSE;
					}
					LinFilter->BorderParam = strtoul(szBuf+Offset, &szEndPtr, 10);
					if (szEndPtr==szBuf+Offset  ||  LinFilter->BorderParam<0  ||  LinFilter->BorderParam>255)
					{
						MessageBox(hwnd, TEXT("Parameter zu 'Border: FillWith' fehlerhaft! Erwarte ganze Zahl (0-255) oder 'Avg'."), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
						free(szBuf);
						free(szTemp);
						return FALSE;
					}
					Offset = szEndPtr-szBuf;
				}
				if (szBuf[Offset]!=' '  &&  szBuf[Offset]!='\t'  &&  szBuf[Offset]!='\r'  &&  szBuf[Offset]!='\n'  &&  szBuf[Offset]!='\0'  &&  Offset<dwLength)
				{
					MessageBox(hwnd, TEXT("Parameter zu 'Border: FillWith' fehlerhaft! Erwarte ganze Zahl (0-255) oder 'Avg'."), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			else
			{
				MessageBox(hwnd, TEXT("Unbekannter Parameter fr 'Border' angegeben!\n")
					               TEXT("Mgliche Parameter sind:\n")
												 TEXT("\tFlipExc\t\t(Spiegeln ohne Randpixel)\n")
												 TEXT("\tFlipInc\t\t(Spiegeln inklusive Randpixel)\n")
												 TEXT("\tPeriodic\t\t(Bitmap periodisch wiederholen)\n")
												 TEXT("\tLocalMean\t(Lokale Mittelwerte)\n")
												 TEXT("\tCoeffSum\t\t(Durch Koeff.-Summe dividieren)\n")
												 TEXT("\tFillWith [Grauwert]\t(Mit angegebenem Grauwert fllen)\n")
												 TEXT("\tFillWith Avg\t(Mit mittlerem Grauwert fllen)"),
												 TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
		}
		else if (strncmp(szTemp, TEXT("DivBy:"), strlen(szTemp)) == 0)
		{
			while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t')
			{
				Offset++;
				if (Offset>=dwLength)
				{
					MessageBox(hwnd, TEXT("Parameter zu 'DivBy:' fehlt! Fliekommazahl ungleich 0 erwartet!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			if (szBuf[Offset]=='\r'  ||  szBuf[Offset]=='\n')
			{
				MessageBox(hwnd, TEXT("Parameter zu 'DivBy:' fehlt! Fliekommazahl ungleich 0 erwartet!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
			LinFilter->DivBy = (float) strtod(szBuf+Offset, &szEndPtr);
			if (szEndPtr==szBuf+Offset  ||  fabs(LinFilter->DivBy)==HUGE_VAL)
			{
				MessageBox(hwnd, TEXT("Fehlerhafter Parameter zu 'DivBy:'! Fliekommazahl ungleich 0 erwartet!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
			bDivByGiven = TRUE;
	    Offset = szEndPtr-szBuf;
			if (szBuf[Offset]!=' '  &&  szBuf[Offset]!='\t'  &&  szBuf[Offset]!='\r'  &&  szBuf[Offset]!='\n'  &&  szBuf[Offset]!='\0'  &&  Offset<dwLength)
			{
				MessageBox(hwnd, TEXT("Fehlerhafter Parameter zu 'DivBy:'! Fliekommazahl ungleich 0 erwartet!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
		}
		else if (strncmp(szTemp, TEXT("AddOffset:"), strlen(szTemp)) == 0)
		{
			while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t')
			{
				Offset++;
				if (Offset>=dwLength)
				{
					MessageBox(hwnd, TEXT("Parameter zu 'AddOffset:' fehlt! Fliekommazahl ungleich 0 erwartet!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			if (szBuf[Offset]=='\r'  ||  szBuf[Offset]=='\n')
			{
				MessageBox(hwnd, TEXT("Parameter zu 'AddOffset:' fehlt! Fliekommazahl ungleich 0 erwartet!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
			LinFilter->AddOffset = (float) strtod(szBuf+Offset, &szEndPtr);
			if (szEndPtr==szBuf+Offset  ||  fabs(LinFilter->AddOffset)==HUGE_VAL)
			{
				MessageBox(hwnd, TEXT("Fehlerhafter Parameter zu 'AddOffset:'! Fliekommazahl ungleich 0 erwartet!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
      Offset = szEndPtr-szBuf;
			if (szBuf[Offset]!=' '  &&  szBuf[Offset]!='\t'  &&  szBuf[Offset]!='\r'  &&  szBuf[Offset]!='\n'  &&  szBuf[Offset]!='\0'  &&  Offset<dwLength)
			{
				MessageBox(hwnd, TEXT("Fehlerhafter Parameter zu 'AddOffset:'! Fliekommazahl ungleich 0 erwartet!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
		}
		else if (strncmp(szTemp, TEXT("Sign:"), strlen(szTemp)) == 0)
		{
			while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t')
			{
				Offset++;
				if (Offset>=dwLength)
				{
					MessageBox(hwnd, TEXT("Parameter zu 'Sign:' fehlt!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			if (szBuf[Offset]=='\r'  ||  szBuf[Offset]=='\n')
			{
				MessageBox(hwnd, TEXT("Parameter zu 'Sign:' fehlt!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
			i = 0;
			while (szBuf[Offset]!=' '  &&  szBuf[Offset]!='\t'  &&  szBuf[Offset]!='\r'  &&  szBuf[Offset]!='\n'  &&  Offset<dwLength)
			{
				szTemp[i] = szBuf[Offset];
				Offset++;
				i++;
			}
			szTemp[i] = '\0';
      if (strncmp(szTemp, TEXT("Keep"), strlen(szTemp)) == 0)
			{
				LinFilter->SignParam = 0;
			}
			else if (strncmp(szTemp, TEXT("ABS"), strlen(szTemp)) == 0)
			{
				LinFilter->SignParam = 1;
			}
			else if (strncmp(szTemp, TEXT("NegativeToZero"), strlen(szTemp)) == 0)
			{
				LinFilter->SignParam = 2;
			}
			else
			{
				MessageBox(hwnd, TEXT("Unbekannter Parameter fr 'Sign:' angegeben!\n")
					               TEXT("Mgliche Parameter sind:\n")
												 TEXT("\tKeep\t\t(Vorzeichen beibehalten)\n")
												 TEXT("\tNegativeToZero\t(Negative Werte gleich Null setzen)\n")
												 TEXT("\tABS\t\t(Betrag bilden)\n"),
												 TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
		}
		else if (strncmp(szTemp, TEXT("Normalize:"), strlen(szTemp)) == 0)
		{
			while (szBuf[Offset]==' '  ||  szBuf[Offset]=='\t')
			{
				Offset++;
				if (Offset>=dwLength)
				{
					MessageBox(hwnd, TEXT("Parameter zu 'Normalize:' fehlt!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
			if (szBuf[Offset]=='\r'  ||  szBuf[Offset]=='\n')
			{
				MessageBox(hwnd, TEXT("Parameter zu 'Normalize:' fehlt!"), TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
			i = 0;
			while (szBuf[Offset]!=' '  &&  szBuf[Offset]!='\t'  &&  szBuf[Offset]!='\r'  &&  szBuf[Offset]!='\n'  &&  Offset<dwLength)
			{
				szTemp[i] = szBuf[Offset];
				Offset++;
				i++;
			}
			szTemp[i] = '\0';
      if (strncmp(szTemp, TEXT("DoNot"), strlen(szTemp)) == 0)
			{
				LinFilter->NormalizeParam = 0;
			}
			else if (strncmp(szTemp, TEXT("Global"), strlen(szTemp)) == 0)
			{
				LinFilter->NormalizeParam = 1;
			}
			else
			{
				MessageBox(hwnd, TEXT("Unbekannter Parameter fr 'Normalize:' angegeben!\n")
					               TEXT("Mgliche Parameter sind:\n")
												 TEXT("\tGlobal\t(Global normalisieren)\n")
												 TEXT("\tDoNot\t(Nicht normalisieren)\n"),
												 TEXT("Lineares Filter - FEHLER"), MB_ICONEXCLAMATION | MB_OK);
				free(szBuf);
				free(szTemp);
				return FALSE;
			}
		}
		else
		{
			wsprintf(szBuf, TEXT("Unbekanntes Schlsselwort: '%s'"), szTemp);
			MessageBox(hwnd, szBuf, TEXT("Lineares Filter - FEHLER"), MB_ICONERROR | MB_OK);
			free(szBuf);
			free(szTemp);
			return FALSE;
		}
	}
	if (!bFilterOk)  MessageBox(hwnd, TEXT("Filter-Matrix nicht ordnungsgem definiert!"), TEXT("Lineares Filter - FEHLER"), MB_ICONERROR | MB_OK);
	if (LinFilter->BorderParam==-5)
	{
		// Immer CoeffSum!=0 ???
		for (y=0; y<LinFilter->FilterHeight; y++)
		{
			for (x=0; x<LinFilter->FilterWidth; x++)
			{
				CoeffSum = 0;
				for (j= ((y<=(LinFilter->FilterHeight>>1)) ? y : 0);  j < ((y>(LinFilter->FilterHeight>>1)) ? y : LinFilter->FilterHeight);  j++)
				{
					for (i= ((x<=(LinFilter->FilterWidth>>1)) ? x : 0);  i < ((x>(LinFilter->FilterWidth>>1)) ? x : LinFilter->FilterWidth);  i++)
					{
						CoeffSum+=LinFilter->FilterMatrix[j*LinFilter->FilterWidth+i];
					}
				}
				if (CoeffSum==0)
				{
					MessageBox(hwnd, TEXT("Bei der angegebenen Filter-Matrix lt sich 'CoeffSum' ")
						               TEXT("nicht als Parameter zu 'Border:' angeben! Die Koeffizientensumme ")
													 TEXT("kann Null werden, so da sich nicht dadurch dividieren lt!")
						             , TEXT("Lineares Filter - FEHLER"), MB_ICONERROR | MB_OK);
					free(szBuf);
					free(szTemp);
					return FALSE;
				}
			}
		}
		if (bDivByGiven)
		{
			if (MessageBox(hwnd, TEXT("WARNUNG:\tObwohl als Parameter zu Border 'CoeffSum' angegeben ist,\n")
										             TEXT("\t\twurde zustzlich 'DivBy' definiert! Diese Option steht\n")
																 TEXT("\t\taber bei 'CoeffSum' nicht zur Verfgung, denn 'DivBy' wird\n")
																 TEXT("\t\twird dann INTERN dynamisch berechnet.\n\n")
													 TEXT("Wollen Sie die Filter-Operation trotzdem ausfhren?\n")
											     TEXT("(Der angegebene Wert fr 'DivBy' wird dabei ignoriert)")
				                 , TEXT("Lineares Filter - WARNUNG"), MB_ICONEXCLAMATION | MB_YESNO | MB_DEFBUTTON2)
				    != IDYES)
			{
				bFilterOk=FALSE;
			}
		}
		LinFilter->BorderParam = 0;
		LinFilter->DivBy       = 0;
	}
	free(szBuf);
	free(szTemp);
	return bFilterOk;
}




BOOL CalcLinFilter(HBITMAP hBitmap, LINFILTER  LinFilter)
{
	HBITMAP hMemBMP;
	BITMAP  bitmap, MemBMP;
	long    x, y, i, j, FilterWidth, FilterHeight, MemWidth, MemHeight, OffBits, OffMem;
	BYTE    *pBits, *pMemBits;
	float   *pFloats, fMax, fMin;
	double  f, CoeffSum;

  FilterWidth = LinFilter.FilterWidth;
	FilterHeight= LinFilter.FilterHeight;
	if (FilterWidth%2!=1)  return FALSE;
	if (FilterHeight%2!=1)  return FALSE;
	FilterWidth  >>= 1;
	FilterHeight >>= 1;
	if (!hBitmap)  return FALSE;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	MemWidth  = bitmap.bmWidth  + (FilterWidth<<1);
	MemHeight = bitmap.bmHeight + (FilterHeight<<1);
	hMemBMP = hBitmap;
	if (LinFilter.DivBy == 0)  LinFilter.BorderParam = 0;  // Auf jeden Fall mit Nullen fllen, wenn DivBy=0 (CoeffSum)
	if (!ChangeBitmapSize(hMemBMP, FALSE, MemWidth, MemHeight, FilterWidth, FilterHeight, LinFilter.BorderParam))  return FALSE;
  GetObject(hMemBMP, sizeof(BITMAP), &MemBMP);
	pMemBits = (BYTE*) MemBMP.bmBits;
	pFloats = (float*) malloc(bitmap.bmWidth*bitmap.bmHeight*sizeof(float));
	if (!pFloats)
	{
		DeleteObject(hMemBMP);
		return FALSE;
	}

	// Filter anwenden
	if (LinFilter.DivBy != 0)
	{
		for (y=0; y<bitmap.bmHeight; y++)
		{
			for (x=0; x<bitmap.bmWidth; x++)
			{
				f = 0;
				OffBits = 0;
				for (j=0; j<=(FilterHeight<<1); j++)
				{
					OffMem = (y+j)*MemBMP.bmWidthBytes+x;
					for (i=0; i<=(FilterWidth<<1); i++)
					{
						f += LinFilter.FilterMatrix[OffBits] * pMemBits[OffMem];
						OffBits++;
						OffMem++;
					}
				}
				f /= LinFilter.DivBy;
				f += LinFilter.AddOffset;
				pFloats[y*bitmap.bmWidth+x] = (float) f;
			}
		}
	}
	else  // Bei DivBy=0 Sonderbehandlung: Koeffizientensumme
	{
		for (y=0; y<bitmap.bmHeight; y++)
		{
			for (x=0; x<bitmap.bmWidth; x++)
			{
				f = 0;
				OffBits = 0;
				CoeffSum = 0;
				for (j=0; j<=(FilterHeight<<1); j++)
				{
					OffMem = (y+j)*MemBMP.bmWidthBytes+x;
					for (i=0; i<=(FilterWidth<<1); i++)
					{
						if (x+i>=FilterWidth  &&  y+j>=FilterHeight
							     &&  x+i<bitmap.bmWidth+FilterWidth  &&  y+j<bitmap.bmHeight+FilterHeight)
						{
							f += LinFilter.FilterMatrix[OffBits] * pMemBits[OffMem];
							CoeffSum += LinFilter.FilterMatrix[OffBits];
						}
						OffBits++;
						OffMem++;
					}
				}
				if (CoeffSum==0)
				{
					DeleteObject(hMemBMP);
					free(pFloats);
					return FALSE;
				}
				f /= CoeffSum;
				f += LinFilter.AddOffset;
				pFloats[y*bitmap.bmWidth+x] = (float) f;
			}
		}
	}

	// Vorzeichen???
	switch (LinFilter.SignParam)
	{
		case 1:
			y = bitmap.bmWidth*bitmap.bmHeight;
			for (x=0; x<y; x++)  pFloats[x]=fabs(pFloats[x]);
			break;

		case 2:
			y = bitmap.bmWidth*bitmap.bmHeight;
			for (x=0; x<y; x++)  if (pFloats[x]<0.0)  pFloats[x]=0.0;
			break;

		default:
			break;
	}

	// Normalisieren???
	switch (LinFilter.NormalizeParam)
	{
		case 1:
			fMin=fMax=pFloats[0];
			y = bitmap.bmWidth*bitmap.bmHeight;
			for (x=1; x<y; x++)
			{
				if (pFloats[x]>fMax)  fMax=pFloats[x];
				if (pFloats[x]<fMin)  fMin=pFloats[x];
			}
			if (fMin!=fMax)
			{
				f=255.0/(fMax-fMin);
				for (x=0; x<y; x++)  pFloats[x] = float((pFloats[x]-fMin)*f);
			}
			break;

		default:
			break;
	}

	// Bitmap schreiben
  OffMem = 0;
	for (y=0; y<bitmap.bmHeight; y++)
	{
		OffBits= y*bitmap.bmWidthBytes;
		for (x=0; x<bitmap.bmWidth; x++)
		{
			if  (pFloats[OffMem]<0)   pFloats[OffMem]=0;
			if (pFloats[OffMem]>255) pFloats[OffMem]=255;
			pBits[OffBits]=(BYTE) pFloats[OffMem];
			OffBits++;
			OffMem++;
		}
	}
	free(pFloats);
	DeleteObject(hMemBMP);
  return TRUE;
}



BOOL CalcLocalNormalize(HBITMAP hBitmap, int Width, int Height, int MaxMinTresh, int Compare)
{
	BITMAP bitmap;
	int    x, y, i, imin, imax, j, jmin, jmax, Offset;
	BYTE   *pBits, *pMemBits, Max, Min;

	if (Width<1  ||  Height<1)  return FALSE;
	if ((Width-1)%2  ||  (Height-1)%2)  return FALSE;

	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	if (Width>bitmap.bmWidth  ||  Height>bitmap.bmHeight)  return FALSE;
	pBits = (BYTE*) bitmap.bmBits;
	pMemBits = (BYTE*) malloc(bitmap.bmWidthBytes*bitmap.bmHeight);
	if (!pMemBits)  return FALSE;
	CopyMemory(pMemBits, pBits, bitmap.bmWidthBytes*bitmap.bmHeight);

  for (y = 0; y < bitmap.bmHeight; y++)
	{
		for (x = 0; x < bitmap.bmWidth; x++)
		{
			// Minimum und Maximum in Mindestbreite und -hhe suchen
			Min=255;
			Max=0;
			jmin = y-(Height>>1);
			if (jmin<0)  jmin=0;
			jmax = y+(Height>>1);
			if (jmax>bitmap.bmHeight-1)  jmax=bitmap.bmHeight-1;
			imin = x-(Width>>1);
			if (imin<0)  imin=0;
			imax = x+(Width>>1);
			if (imax>bitmap.bmWidth-1)  imax=bitmap.bmWidth-1;
			for (j=jmin; j<=jmax; j++)
			{
				Offset=j*bitmap.bmWidthBytes+imin;
				for (i=imin; i<=imax; i++)
				{
					if (pMemBits[Offset]>Max)  Max=pMemBits[Offset];
					if (pMemBits[Offset]<Min)  Min=pMemBits[Offset];
					Offset++;
				}
			}

/*			while (Max-Min<Param  &&  (imin>0  ||  imax<bitmap.bmWidth-1  ||  jmin>0  ||  jmax<bitmap.bmHeight-1))
			{
				if (jmin>0)
				{
					jmin--;
					Offset=jmin*bitmap.bmWidthBytes+imin;
					for (i=imin; i<=imax; i++)
					{
						if (pMemBits[Offset]>Max)  Max=pMemBits[Offset];
						if (pMemBits[Offset]<Min)  Min=pMemBits[Offset];
						Offset++;
					}
				}
				if (jmax<bitmap.bmHeight-1)
				{
					jmax++;
					Offset=jmax*bitmap.bmWidthBytes+imin;
					for (i=imin; i<=imax; i++)
					{
						if (pMemBits[Offset]>Max)  Max=pMemBits[Offset];
						if (pMemBits[Offset]<Min)  Min=pMemBits[Offset];
						Offset++;
					}
				}
				if (imin>0)
				{
					imin--;
					Offset=jmin*bitmap.bmWidthBytes+imin;
					for (j=jmin; j<=jmax; j++)
					{
						if (pMemBits[Offset]>Max)  Max=pMemBits[Offset];
						if (pMemBits[Offset]<Min)  Min=pMemBits[Offset];
						Offset += bitmap.bmWidthBytes;
					}
				}
				if (imax<bitmap.bmWidth-1)
				{
					imax++;
					Offset=jmin*bitmap.bmWidthBytes+imax;
					for (j=jmin; j<=jmax; j++)
					{
						if (pMemBits[Offset]>Max)  Max=pMemBits[Offset];
						if (pMemBits[Offset]<Min)  Min=pMemBits[Offset];
						Offset += bitmap.bmWidthBytes;
					}
				}
			}*/
			// Neuen Grauwert berechnen und speichern in pBits
			if (Max-Min<MaxMinTresh)
			{
				if (pMemBits[y*bitmap.bmWidthBytes+x]<Compare)
					pBits[y*bitmap.bmWidthBytes+x] = 0;
				else
					pBits[y*bitmap.bmWidthBytes+x] = 255;
			}
			else
				pBits[y*bitmap.bmWidthBytes+x] = BYTE(((double) pMemBits[y*bitmap.bmWidthBytes+x]-Min)/(Max-Min)*255);
		} 
	} 
	free (pMemBits);
	return TRUE;
}


BOOL CalcGradientFilter(HBITMAP hBitmap)
{
	BITMAP bitmap;
	BYTE   *pBits;
	long   x, y, x1, x2, y1, y2, OffBits, OffGrad;
	double *Grad, MaxGrad;

	if (!hBitmap)  return FALSE;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	Grad = (double*) malloc(bitmap.bmWidth*bitmap.bmHeight*sizeof(double));
	if (!Grad)  return FALSE;
	if (bitmap.bmWidth<=2  ||  bitmap.bmHeight<=2)
	{
		free(Grad);
		return FALSE;
	}
	OffGrad = 0;
	MaxGrad = 0;
	for (y=0; y<bitmap.bmHeight; y++)
	{
		y1 = y-1;
		y2 = y+1;
		if (y1<0)  y1=0;
		if (y2>=bitmap.bmHeight)  y2=bitmap.bmHeight-1;
		for (x=0; x<bitmap.bmWidth; x++)
		{
			x1 = x-1;
			x2 = x+1;
			if (x1<0)  x1=0;
			if (x2>=bitmap.bmWidth)  x2=bitmap.bmWidth-1;
			Grad[OffGrad] = hypot(( (double) pBits[y*bitmap.bmWidthBytes+x2] - (double) pBits[y*bitmap.bmWidthBytes+x1]) / (x2-x1),
				                        ( (double) pBits[y2*bitmap.bmWidthBytes+x] - (double) pBits[y1*bitmap.bmWidthBytes+x]) / (y2-y1) );
			if (Grad[OffGrad]>MaxGrad)  MaxGrad=Grad[OffGrad];
			OffGrad++;
		}
	}
	OffGrad = 0;
	for (y=0; y<bitmap.bmHeight; y++)
	{
		OffBits = y*bitmap.bmWidthBytes;
		for (x=0; x<bitmap.bmWidth; x++)
		{
			pBits[OffBits] = BYTE(Grad[OffGrad]/MaxGrad*255.0);
			OffBits++;
			OffGrad++;
		}
	}
	free(Grad);
	return TRUE;
}


void fft2d(float DataRe[], float DataIm[], unsigned long Width, unsigned long Height, int iSign)
{
	int           idim;
	unsigned long i1, i2, i3, i2rev, i3rev, ip1, ip2, ip3, ifp1, ifp2,
		            ibit, k1, k2, n, nprev, nrem, ntot;
	float         tempi, tempr, temp;
	double        theta, wi, wpi, wpr, wr, wtemp;
	
	ntot = Width*Height;
	nprev = 1;
	for (idim=2; idim>=1; idim--)
	{
		n = (idim==1) ? Width : Height;
		nrem = ntot / (n*nprev);
		ip1 = nprev;
		ip2 = ip1*n;
		ip3 = ip2*nrem;
		i2rev = 0;
		for (i2=0; i2<ip2; i2+=ip1)
		{
			if (i2 < i2rev)
			{
				for (i1=i2; i1<i2+ip1; i1++);
				{
					for (i3=i1; i3<=ip3; i3+=ip2)
					{
						i3rev = i2rev+i3-i2;
						SWAP(DataRe[i3], DataRe[i3rev]);
						SWAP(DataIm[i3], DataIm[i3rev]);
					}
				}
			}
			ibit = ip2;
			while (ibit>=ip1  &&  i2rev>ibit)
			{
				i2rev -= ibit;
				ibit >>= 1;
			}
			i2rev += ibit;
		}
		ifp1 = ip1;
		while (ifp1<ip2)
		{
			ifp2 = ifp1 << 1;
			theta = -iSign*2.0*Pi/(ifp2/ip1);
			wtemp = sin(0.5*theta);
			wpr = -2.0*wtemp*wtemp;
			wpi = sin(theta);
			wr = 1.0;
			wi = 0.0;
			for (i3=0; i3<ifp1; i3+=ip1)
			{
				for (i1=i3; i1<i3+ip1; i1++)
				{
					for (i2=i1; i2<=ip3; i2+=ifp2)
					{
						k1 = i2;
						k2 = k1 + ifp1;
						tempr = (float)wr*DataRe[k2]-(float)wi*DataIm[k2];
						tempi = (float)wr*DataIm[k2]+(float)wi*DataRe[k2];
						DataRe[k2] = DataRe[k1] - tempr;
						DataIm[k2] = DataIm[k1] - tempi;
						DataRe[k1] += tempr;
						DataIm[k1] += tempi;
					}
				}
				wr = (wtemp=wr)*wpr-wi*wpi+wr;
				wi = wi*wpr+wtemp*wpi+wi;
			}
			ifp1=ifp2;
		}
		nprev *= n;
	}
	
	if (iSign>0)    // Wenn Hintransformation:   Geteilt durch Width*Height  (ntot)
	{
		for (k1=0; k1<ntot; k1++)
		{
			DataRe[k1] /= ntot;
			DataIm[k1] /= ntot;
		}
	}
}


void DFT2D(float DataRe[], float DataIm[], unsigned long Width, unsigned long Height, int iSign)
{
	float	 *TmpDatRe, *TmpDatIm;
	float  *eRe, *eIm;
	unsigned long   x, y, n, k, nOffset, kOffset;
	
	nOffset   =  (Width>Height) ? Width : Height;
	TmpDatRe = (float*) malloc(nOffset*sizeof(float));
	TmpDatIm = (float*) malloc(nOffset*sizeof(float));
	eRe      = (float*) malloc(nOffset*sizeof(float));
	eIm      = (float*) malloc(nOffset*sizeof(float));

	
	// 1. Transformation:     HORIZONTAL
	eRe[0] = 1.0;
	eIm[0] = 0.0;
	eRe[1] = float( cos(2.0*Pi/Width));
	eIm[1] = float(-sin(2.0*Pi/Width)*iSign);
	for (n=2; n<Width; n++)
	{
		eRe[n] = eRe[n-1]*eRe[1] - eIm[n-1]*eIm[1];
		eIm[n] = eRe[n-1]*eIm[1] + eIm[n-1]*eRe[1];
	}
	
	for (y=0; y<Height; y++)
	{
		for (k=0; k<Width; k++)
		{
			kOffset = 0;
			TmpDatRe[k] = 0;
			TmpDatIm[k] = 0;
			for (x = 0; x<Width; x++)
			{
				nOffset = y*Width + x;
				TmpDatRe[k] += DataRe[nOffset]*eRe[kOffset] - DataIm[nOffset]*eIm[kOffset];
				TmpDatIm[k] += DataRe[nOffset]*eIm[kOffset] + DataIm[nOffset]*eRe[kOffset];
				kOffset += k;
				kOffset %= Width;
			}
		}
		CopyMemory(DataRe+y*Width, TmpDatRe, Width*sizeof(float));
		CopyMemory(DataIm+y*Width, TmpDatIm, Width*sizeof(float));
	}

	// 2. Transformation:     VERTIKAL
	eRe[1] = float( cos(2.0*Pi/Height));
	eIm[1] = float(-sin(2.0*Pi/Height)*iSign);
	for (n=2; n<Height; n++)
	{
		eRe[n] = eRe[n-1]*eRe[1] - eIm[n-1]*eIm[1];
		eIm[n] = eRe[n-1]*eIm[1] + eIm[n-1]*eRe[1];
	}
	
	for (x=0; x<Width; x++)
	{
		for (k=0; k<Height; k++)
		{
			kOffset = 0;
			TmpDatRe[k] = 0;
			TmpDatIm[k] = 0;
			for (y = 0; y<Height; y++)
			{
				nOffset = y*Width + x;
				TmpDatRe[k] += DataRe[nOffset]*eRe[kOffset] - DataIm[nOffset]*eIm[kOffset];
				TmpDatIm[k] += DataRe[nOffset]*eIm[kOffset] + DataIm[nOffset]*eRe[kOffset];
				kOffset += k;
				kOffset %= Height;
			}
		}
		for (y=0; y<Height; y++)   //Ergebnis kopieren
		{
			DataRe[y*Width+x] = TmpDatRe[y];
			DataIm[y*Width+x] = TmpDatIm[y];
		}
	}

	if (iSign>0)      // Durch die Zahl der Pixel dividieren, wenn Hintransformation
	{
		for (x=0; x<Width; x++)
		{
			for (y=0; y<Height; y++)
			{
				DataRe[y*Width+x] /= Width*Height;
		    DataIm[y*Width+x] /= Width*Height;
			}
		}
	}

  // Speicher wieder freigeben
	free(eRe);
	free(eIm);
	free(TmpDatRe);
	free(TmpDatIm);
}



void FFT(float data[], unsigned long nn, int iSign)
{
	unsigned long n, mmax, m, j, istep, i;
	double        temp, wtemp, tempi, wr, wpr, wpi, wi, theta;
	float         fltemp;

	n = nn<<1;
	j = 1;
	for (i=1; i<n; i+=2)
	{
		if (j>i) 
		{
			fltemp=data[j];   data[j]  =data[i];   data[i]  =fltemp;
			fltemp=data[j+1]; data[j+1]=data[i+1]; data[i+1]=fltemp;
		}
		m = n>>1;
		while (m>=2 && j>m)
		{
			j -= m;
			m >>= 1;
		}
		j += m;
	}

	mmax = 2;
	while (n>mmax)
	{
		istep = mmax<<1;
		theta = -iSign*(2*Pi/mmax);
		wtemp = sin(0.5*theta);
		wpr   = -2*wtemp*wtemp;
		wpi   = sin(theta);
		wr    = 1.0;
		wi    = 0.0;
		for (m=1; m<mmax; m+=2)
		{
			for (i=m; i<=n; i+=istep)
			{
				j = i+mmax;
				temp  = wr*data[j]   - wi*data[j+1];
				tempi = wr*data[j+1] + wi*data[j];
				data[j]  = float(data[i]   - temp);
				data[j+1]= float(data[i+1] - tempi);
				data[i]   += (float) temp;
				data[i+1] += (float) tempi;
			}
			wr = (wtemp=wr)*wpr - wi*wpi + wr;
			wi = wi*wpr + wtemp*wpi + wi;
		}
		mmax = istep;
	}
}


void FFT2D(float DataRe[], float DataIm[], unsigned long Width, unsigned long Height, int iSign)
{
	unsigned long x, y, Offset;
	float *Data, f;

	Offset = (Height>Width) ? Height : Width;
	Data = (float*) malloc(2*sizeof(float)*Offset);

	for (y=0; y<Height; y++)
	{
		Offset = y*Width;
		for (x=0; x<Width; x++)
		{
			Data[2*x]   = DataRe[Offset+x];
      Data[2*x+1] = DataIm[Offset+x];
		}
//		DFT(Data, Width, iSign);
		FFT(Data-1, Width, iSign);
		for (x=0; x<Width; x++)
		{
			DataRe[Offset+x] = Data[2*x];
      DataIm[Offset+x] = Data[2*x+1];
		}
	}

	for (x=0; x<Width; x++)
	{
		for (y=0; y<Height; y++)
		{
			Data[2*y]   = DataRe[y*Width+x];
      Data[2*y+1] = DataIm[y*Width+x];
		}
//		DFT(Data, Height, iSign);
		FFT(Data-1, Height, iSign);
		for (y=0; y<Height; y++)
		{
			DataRe[y*Width+x] = Data[2*y];
      DataIm[y*Width+x] = Data[2*y+1];
		}
	}
	if (iSign>0)
	{
		f = float(Width*Height);
		for (x=0; x<Width; x++)
		{
			for (y=0; y<Height; y++)
			{
				DataRe[y*Width+x] /= f;
		    DataIm[y*Width+x] /= f;
			}
		}
	}
	free(Data);
}


BOOL CalcSpectrum(HWND hwnd, HBITMAP hBitmap, int MaxFreq)
{
	BITMAP bitmap;
	int    x, y, i, j, xmax, ymax, OffsetX, OffsetY;
	BYTE   *pBits;
	float  *SpectRe, *SpectIm, f;

	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	xmax = 2;
	while (xmax<=bitmap.bmWidth) xmax<<=1;   //Nchste 2er-Potenz suchen
	ymax = 2;
	while (ymax<=bitmap.bmHeight) ymax<<=1;  //Nchste 2er-Potenz suchen
	if (xmax<ymax)  xmax=ymax;  else  ymax=xmax;   //Nur quadratischen Bereich benutzen
xmax=ymax=2048;
	OffsetX = (xmax - bitmap.bmWidth);
	OffsetY = (ymax - bitmap.bmHeight);
	SpectRe = (float*) malloc(xmax*ymax*sizeof(float));
	SpectIm = (float*) malloc(xmax*ymax*sizeof(float));

	for (y=0; y<ymax; y++)
	{
		j = abs(y-OffsetY);
		if (j>=bitmap.bmHeight)  j = bitmap.bmHeight - (j-bitmap.bmHeight) - 2;
		for (x=0; x<xmax; x++)
		{
			i = abs(x-OffsetX);
			if (i>=bitmap.bmWidth)  i = bitmap.bmWidth - (i-bitmap.bmWidth) - 2;
			if (x-OffsetX<0  ||  x-OffsetX>=bitmap.bmWidth  ||  y-OffsetY<0  ||  y-OffsetY>=bitmap.bmHeight)
				SpectRe[y*xmax+x] = 0;
			else
				SpectRe[y*xmax+x] = pBits[j*bitmap.bmWidthBytes+i];
			SpectIm[y*xmax+x] = 0;
		}
	}
  i = GetTickCount();
	FFT2D(SpectRe, SpectIm, xmax, ymax, 1);
  i = GetTickCount()-i;
	char szBuf[80];
	MessageBox(hwnd, itoa(i, szBuf, 10), TEXT("MILLISEKUNDEN"), MB_OK);
	for (y=0; y<ymax; y++)
	{
		for (x=0; x<xmax; x++)
		{
			if (sqrt((double) x*x+y*y)<MaxFreq  ||  sqrt((double) (xmax-x)*(xmax-x)+y*y)<MaxFreq
				   ||  sqrt((double) x*x+(ymax-y)*(ymax-y))<MaxFreq  ||  sqrt((double) (xmax-x)*(xmax-x)+(ymax-y)*(ymax-y))<MaxFreq)
			{
			SpectRe[y*xmax+x] = 0;
			SpectIm[y*xmax+x] = 0;
			}
		}
	}

	FFT2D(SpectRe, SpectIm, xmax, ymax, -1);
	f=1;
	for (y=0; y<ymax; y++)
	{
		for (x=0; x<xmax; x++)
		{
			SpectRe[y*xmax+x]= ((sqrt(SpectRe[y*xmax+x]*SpectRe[y*xmax+x]+SpectIm[y*xmax+x]*SpectIm[y*xmax+x])));
			//if (x==0 && y==0) SpectRe[y*xmax+x]=0;
		}
	}
  f=0;
	for (y=0; y<ymax; y++)
	{
		for (x=0; x<xmax; x++)
		{
			if (f<SpectRe[y*xmax+x])  f=SpectRe[y*xmax+x];
		}
	}
	f/=255;
	for (y=0; y<ymax; y++)
	{
		for (x=0; x<xmax; x++)
		{
			SpectRe[y*xmax+x] = SpectRe[y*xmax+x]/f;
			if (SpectRe[y*xmax+x]>255)  SpectRe[y*xmax+x]=255;
			if (SpectRe[y*xmax+x]<0)  SpectRe[y*xmax+x]=0;
		}
	}

	for (y=0; y<bitmap.bmHeight; y++)
	{
		for (x=0; x<bitmap.bmWidth; x++)
		{
			f = (SpectRe[(y+OffsetY)*xmax+(x+OffsetX)]);
			if (f<0)  f=-f;
			if (f>255) f=255;
			pBits[y*bitmap.bmWidthBytes+x] = (BYTE) (f+0.01);
		}
	}
 
/*	f = 0;
	for (y=0; y<ymax; y++)
	{
		Offset = y*xmax;
		for (x=0; x<xmax; x++)
			SpectRe[Offset+x] = SpectRe[Offset+x]*SpectRe[Offset+x] + SpectIm[Offset+x]*SpectIm[Offset+x];
	}
	for (y=0; y<ymax; y++)
	{
		Offset = y*xmax;
		for (x=0; x<xmax; x++)
			if (f<SpectRe[Offset+x])  f = SpectRe[Offset+x];
	}
	f = f/2;
	for (y=0; y<ymax; y++)
	{
		Offset = y*xmax;
		for (x=0; x<xmax; x++)
			if (SpectRe[Offset+x]>f)  SpectRe[Offset+x]=f;
	}
*/ /*	for (y=0; y<ymax; y++)
	{
		Offset = y*xmax;
		for (x=0; x<xmax; x++)
		{
			pBits[y*bitmap.bmWidthBytes+x] = SpectRe[Offset+x];
			/*pBits[(ymax/2-y-1)*bitmap.bmWidthBytes+x+xmax/2] = SpectRe[(ymax-y-1)*xmax+x];
			pBits[(y+ymax/2)*bitmap.bmWidthBytes+xmax/2-x-1] = SpectRe[Offset+xmax-x-1];
			pBits[(ymax/2-y-1)*bitmap.bmWidthBytes+xmax/2-x-1] = SpectRe[(ymax-y-1)*xmax+xmax-x-1];*/
/*		}
	}
/**/
	free(SpectRe);
	free(SpectIm);
	return TRUE;
}


long AreaBinarisation(HBITMAP hBitmap, BYTE GrayCutoff, long *Areas)
{
	BITMAP bitmap;
	BYTE   *pBits;
	long   x, y, Offset;

	if (!hBitmap)  return 0;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	for (y=0; y<bitmap.bmHeight; y++)
	{
		Offset = y*bitmap.bmWidthBytes;
	  for (x=0; x<bitmap.bmWidth; x++)
		{
	    if (pBits[Offset]<GrayCutoff)  Areas[Offset]=0;  else  Areas[Offset]=-1;
			Offset++;
		}
	  for (x=bitmap.bmWidth; x<bitmap.bmWidthBytes; x++)
		{
	    Areas[Offset]=0;
			Offset++;
		}
	}
  return 0;
}


long SearchMaxima(HBITMAP hBitmap, BYTE GrayCutoff, long *Areas)   // Markiere lokale Maxima > GrayCutoff als "-1" in Areas[],
{																																	 // flle den Rest mit "0"
	BITMAP bitmap;
	BYTE   *pBits, g;
	long   x, y, Offset;
	BOOL   IsMax, Ready;

	if (!hBitmap)  return 0;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	for (x=0; x<bitmap.bmWidthBytes*bitmap.bmHeight; x++)  Areas[x]=0;

	// Erster Durchlauf: lokale Maxima suchen. Markiere alle Pixel mit
	// Grauwert >= Grauwert der Nachbarn
	for (y=0; y<bitmap.bmHeight; y++)
	{
		Offset = y*bitmap.bmWidthBytes;
	  for (x=0; x<bitmap.bmWidth; x++)
		{
			g = pBits[Offset];     // Zu diskutierender Grauwert g: Ist dies ein Maximum?
			if (g<GrayCutoff)
				IsMax = FALSE;       // Wenn g kleiner als Grauwertgrenze, irrelevant
			else                   // ansonsten:
			{
				IsMax = TRUE;        // Zunchst Annahme: JA, dies knnte ein Maximum sein
				if (x>0)             // Untersuche alle Nachbarpixel...
					if (pBits[Offset-1] > g)  IsMax=FALSE;
				if (x<bitmap.bmWidth-1)
					if (pBits[Offset+1]> g)  IsMax=FALSE;
				if (y>0)
				{
					if (pBits[Offset-bitmap.bmWidthBytes] > g)  IsMax=FALSE;
					if (x>0)
						if (pBits[Offset-bitmap.bmWidthBytes-1] > g)  IsMax=FALSE;
					if (x<bitmap.bmWidth-1)
						if (pBits[Offset-bitmap.bmWidthBytes+1] > g)  IsMax=FALSE;
				}
				if (y<bitmap.bmHeight-1)
				{
					if (pBits[Offset+bitmap.bmWidthBytes] > g)  IsMax=FALSE;
					if (x>0)
						if (pBits[Offset+bitmap.bmWidthBytes-1] > g)  IsMax=FALSE;
					if (x<bitmap.bmWidth-1)
						if (pBits[Offset+bitmap.bmWidthBytes+1] > g)  IsMax=FALSE;
				}
			}
			if (IsMax)   // g tatschlich grer oder gleich den Grauwerten der Nachbarpixel
			{            // Hat einer der bereits untersuchten Nachbarpixel denselben Grauwert,
				if (x>0)   // ist aber kein Maximum? Wenn ja, kann g kein Maximum sein!
					if (pBits[Offset-1] == g  &&  !Areas[Offset-1])  IsMax=FALSE;
				if (y>0)
				{
					if (pBits[Offset-bitmap.bmWidthBytes] == g  &&  !Areas[Offset-bitmap.bmWidthBytes])  IsMax=FALSE;
					if (x>0)
						if (pBits[Offset-bitmap.bmWidthBytes-1] == g  &&  !Areas[Offset-bitmap.bmWidthBytes-1])  IsMax=FALSE;
					if (x<bitmap.bmWidth-1)
						if (pBits[Offset-bitmap.bmWidthBytes+1] == g  &&  !Areas[Offset-bitmap.bmWidthBytes+1])  IsMax=FALSE;
				}
			}
			if (IsMax)   // Wenn immernoch Maximum, speichere Grauwert in Areas[Offset]
				Areas[Offset] = -1;
			Offset++;
		}
	}

	// Zweiter Durchlauf: (Hin und her) unntig markierte Pixel wieder entfernen, bis alle entfernt
	do
	{
		Ready = TRUE;
		for (y=bitmap.bmHeight-1; y>=0; y--)
		{
			Offset = y*bitmap.bmWidthBytes+bitmap.bmWidth-1;
			for (x=bitmap.bmWidth-1; x>=0; x--)
			{
				if (Areas[Offset])
				{
					g = pBits[Offset];
					IsMax = TRUE;
					if (x<bitmap.bmWidth-1)
						if (pBits[Offset+1] == g  &&  !Areas[Offset+1])  IsMax=FALSE;
					if (y<bitmap.bmHeight-1)
					{
						if (pBits[Offset+bitmap.bmWidthBytes] == g  &&  !Areas[Offset+bitmap.bmWidthBytes])  IsMax=FALSE;
						if (x<bitmap.bmWidth-1)
							if (pBits[Offset+bitmap.bmWidthBytes+1] == g  &&  !Areas[Offset+bitmap.bmWidthBytes+1])  IsMax=FALSE;
						if (x>0)
							if (pBits[Offset+bitmap.bmWidthBytes-1] == g  &&  !Areas[Offset+bitmap.bmWidthBytes-1])  IsMax=FALSE;
					}
					if (!IsMax)    // Wenn doch kein Maximum:
					{
						Areas[Offset] = 0;              // Lsche Markierung wieder
						Ready = FALSE;
					}
				}
				Offset--;
			}
		}

		if (!Ready)
		{
			Ready = TRUE;
			for (y=0; y<bitmap.bmHeight; y++)
			{
				Offset = y*bitmap.bmWidthBytes;
				for (x=0; x<bitmap.bmWidth; x++)
				{
					if (Areas[Offset])
					{
						g = pBits[Offset];
						IsMax = TRUE;
						if (x>0)
							if (pBits[Offset-1] == g  &&  !Areas[Offset-1])  IsMax=FALSE;
						if (y>0)
						{
							if (pBits[Offset-bitmap.bmWidthBytes] == g  &&  !Areas[Offset-bitmap.bmWidthBytes])  IsMax=FALSE;
							if (x<bitmap.bmWidth-1)
								if (pBits[Offset-bitmap.bmWidthBytes+1] == g  &&  !Areas[Offset-bitmap.bmWidthBytes+1])  IsMax=FALSE;
							if (x>0)
								if (pBits[Offset-bitmap.bmWidthBytes-1] == g  &&  !Areas[Offset-bitmap.bmWidthBytes-1])  IsMax=FALSE;
						}
						if (!IsMax)    // Wenn doch kein Maximum:
						{
							Areas[Offset] = 0;              // Lsche Markierung wieder
							Ready = FALSE;
						}
					}
					Offset++;
				}
			}
		}
	} while (!Ready);

	return 0;
}


long AreaSeparation(HBITMAP hBitmap, long *Areas)  // Numeriere zusammenhngende Gebiete in Areas[]
{
	BITMAP bitmap;
	long   x, y, Offset, MaxNr;
	BOOL   Ready;

	if (!hBitmap)  return 0;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);

	// Erster Durchlauf:
	// Pixel zeilenweise von unten links nach oben rechts durchnumerieren
	for (y=0; y<bitmap.bmHeight; y++)
	{
		Offset = y*bitmap.bmWidthBytes;
	  for (x=0; x<bitmap.bmWidth; x++)
		{
			if (Areas[Offset])	Areas[Offset]=Offset+1;
			Offset++;
		}
	}

	// Zweiter Durchlauf: (Hin und her) benachbarte Pixel durch maximale Nr gleich benennen
	do
	{
		Ready = TRUE;
		for (y=bitmap.bmHeight-1; y>=0; y--)
		{
			Offset = y*bitmap.bmWidthBytes+bitmap.bmWidth-1;
			for (x=bitmap.bmWidth-1; x>=0; x--)
			{
				if (Areas[Offset])
				{
					MaxNr = Areas[Offset];
					if (x<bitmap.bmWidth-1)
						if (Areas[Offset+1]>MaxNr)  MaxNr = Areas[Offset+1];
					if (y<bitmap.bmHeight-1)
					{
						if (Areas[Offset+bitmap.bmWidthBytes]>MaxNr)  MaxNr = Areas[Offset+bitmap.bmWidthBytes];
						if (x<bitmap.bmWidth-1)
							if (Areas[Offset+bitmap.bmWidthBytes+1]>MaxNr)  MaxNr = Areas[Offset+bitmap.bmWidthBytes+1];
						if (x>0)
							if (Areas[Offset+bitmap.bmWidthBytes-1]>MaxNr)  MaxNr = Areas[Offset+bitmap.bmWidthBytes-1];
					}
					if (MaxNr>Areas[Offset])    // Wenn aktuelles Pixel umbenannt werden mu:
					{
						Areas[Offset] = MaxNr;              // Benenne Pixel mit maximaler PartikelNr
						Ready = FALSE;
					}
				}
				Offset--;
			}
		}

		if (!Ready)
		{
			Ready = TRUE;
			for (y=0; y<bitmap.bmHeight; y++)
			{
				Offset = y*bitmap.bmWidthBytes;
				for (x=0; x<bitmap.bmWidth; x++)
				{
					if (Areas[Offset])
					{
						MaxNr = Areas[Offset];
						if (x>0)
							if (Areas[Offset-1]>MaxNr)  MaxNr = Areas[Offset-1];
						if (y>0)
						{
							if (Areas[Offset-bitmap.bmWidthBytes]>MaxNr)  MaxNr = Areas[Offset-bitmap.bmWidthBytes];
							if (x<bitmap.bmWidth-1)
								if (Areas[Offset-bitmap.bmWidthBytes+1]>MaxNr)  MaxNr = Areas[Offset-bitmap.bmWidthBytes+1];
							if (x>0)
								if (Areas[Offset-bitmap.bmWidthBytes-1]>MaxNr)  MaxNr = Areas[Offset-bitmap.bmWidthBytes-1];
						}
						if (MaxNr>Areas[Offset])    // Wenn aktuelles Pixel umbenannt werden mu:
						{
							Areas[Offset] = MaxNr;              // Benenne Pixel mit maximaler PartikelNr
							Ready = FALSE;
						}
					}
					Offset++;
				}
			}
		}
	} while (!Ready);   // Wiederhole, bis nichts mehr umzubenennen ist.

	MaxNr = 0;
	x = bitmap.bmWidthBytes*bitmap.bmHeight;
	for (Offset=0; Offset<x; Offset++)
	{
		if (Areas[Offset] == Offset+1)  MaxNr++;
	}
	return MaxNr;
}

long AreaErosion(HBITMAP hBitmap, long *Areas, PARTICLEDATA *Particles, long NumParticles)
{
	BITMAP bitmap;
	BYTE   *pBits;
	long   Offset, i, Nr;

	if (!hBitmap)  return 0;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
/*
	// Nur erstes gefundenes Pixel einer bestimmten Nummer lassen, den Rest derselben Nummer lschen:
	for (y=0; y<bitmap.bmHeight; y++)
	{
		Offset = y*bitmap.bmWidthBytes;
	  for (x=0; x<bitmap.bmWidth; x++)
		{
			if (Areas[Offset])
			{
				Nr = Areas[Offset];
				for (i=Offset+1; i<Nr; i++)
					if (Areas[i]==Nr)  Areas[i]=0;
			}
			Offset++;
		}
	}
*/

	Nr=0;
	i=bitmap.bmWidthBytes*bitmap.bmHeight;
	for (Offset=0; Offset<i; Offset++)
	{
		if (Areas[Offset]==Offset+1)
		{
			Particles[Nr].X = Offset % bitmap.bmWidthBytes;
			Particles[Nr].Y = Offset / bitmap.bmWidthBytes;
			Particles[Nr].Area = 1;
			Particles[Nr].GrayMax = pBits[Offset];
			Particles[Nr].GrayAvg = pBits[Offset];
			Nr++;
			if (Nr>NumParticles)  return Nr;
		}

	}

	return Nr;
}


long AreaCenter(HBITMAP hBitmap, long *Areas, PARTICLEDATA *Particles, long NumParticles)
{
	BITMAP bitmap;
	BYTE   *pBits;
	long   x, y, i, Offset, Nr;

	if (!hBitmap)  return 0;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	// Neue Nummern aufsteigend verteilen und Flchenschwerpunkte damit numerieren:
  Nr = 0;
	for (y=bitmap.bmHeight-1; y>=0; y--)
	{
		Offset = y*bitmap.bmWidthBytes+bitmap.bmWidth-1;
		for (x=bitmap.bmWidth-1; x>=0; x--)
		{
			if (Areas[Offset]==Offset+1)
			{
				Nr++;
				if (Nr>NumParticles) return Nr;
				Areas[Offset] = Nr;
			}
			else if (Areas[Offset])
			{
				Areas[Offset] = Areas[Areas[Offset]-1];
			}
			Offset--;
		}
	}
	if (Nr<NumParticles)  return Nr;

  // Initialisierung der Partikeldaten
	for (i=0; i<NumParticles; i++)
	{
		Particles[i].X = 0;
		Particles[i].Y = 0;
		Particles[i].BoundRect.left   = bitmap.bmWidth;
		Particles[i].BoundRect.right  = -1;
		Particles[i].BoundRect.bottom = bitmap.bmHeight;
		Particles[i].BoundRect.top    = -1;
		Particles[i].Area    = 0;
		Particles[i].GrayAvg = 0;
		Particles[i].GrayMax = 0;
	}
	// Partikeldaten aufsummieren fr Mittelwerte
	for (y=0; y<bitmap.bmHeight; y++)
	{
		Offset = y*bitmap.bmWidthBytes;
		for (x=0; x<bitmap.bmWidth; x++)
		{
			if (Areas[Offset])
			{
				Particles[Areas[Offset]-1].X += x * (double) pBits[Offset];
				Particles[Areas[Offset]-1].Y += y * (double) pBits[Offset];
				if (Particles[Areas[Offset]-1].BoundRect.left>x)
					Particles[Areas[Offset]-1].BoundRect.left=x;
				if (Particles[Areas[Offset]-1].BoundRect.right<x)
					Particles[Areas[Offset]-1].BoundRect.right=x;
				if (Particles[Areas[Offset]-1].BoundRect.top<y)
					Particles[Areas[Offset]-1].BoundRect.top=y;
				if (Particles[Areas[Offset]-1].BoundRect.bottom>y)
					Particles[Areas[Offset]-1].BoundRect.bottom=y;
				Particles[Areas[Offset]-1].Area += 1;
				Particles[Areas[Offset]-1].GrayAvg += pBits[Offset];
				if (pBits[Offset] > Particles[Areas[Offset]-1].GrayMax)  Particles[Areas[Offset]-1].GrayMax= pBits[Offset];
			}
			Offset++;
		}
	}

	for (i=0; i<NumParticles; i++)
	{
		Particles[i].X /= (double) Particles[i].GrayAvg;
		Particles[i].Y /= (double) Particles[i].GrayAvg;
		Particles[i].GrayAvg /= (double) Particles[i].Area;
	}

	return 0;
}


long CheckAreaSize(PARTICLEDATA *Particles, long &NumParticles, long MinArea, long MaxArea)
{
	long i,j;

	j=0;
	for (i=0; i<NumParticles; i++)
	{
		if (Particles[i].Area >= MinArea  &&  Particles[i].Area <= MaxArea)
		{
			Particles[j] = Particles[i];
			j++;
		}
	}

	NumParticles = j;
	return NumParticles;
}


long ShowParticleCoord(HBITMAP hBitmap, PARTICLEDATA *Particles, long NumParticles)
{
	BITMAP bitmap;
	BYTE   *pBits;
	long   x, y, i, Offset, Size;

	if (!hBitmap)  return 0;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	Size = bitmap.bmWidthBytes*bitmap.bmHeight;
  for (Offset=0; Offset<Size; Offset++)
		if  (pBits[Offset]==255)  pBits[Offset]=254;

	for (i=0; i<NumParticles; i++)
	{
		x = long(Particles[i].X + 0.5);  // auf (long) runden
		y = long(Particles[i].Y + 0.5);  // auf (long) runden
		Offset = y*bitmap.bmWidthBytes+x;
		pBits[Offset]=255;
		if (x<bitmap.bmWidth-1)  pBits[Offset+1]=255;
		if (x>0)  pBits[Offset-1]=255;
		if (y<bitmap.bmHeight-1)  pBits[Offset+bitmap.bmWidthBytes]=255;
		if (y>0)  pBits[Offset-bitmap.bmWidthBytes]=255;
	}

	return 0;
}


long ExportParticleData(PARTICLEDATAINFO ParticleDataInfo)
{
	TCHAR  szSaveFile[MAX_PATH];
	HANDLE hFile;
	TCHAR  szBuf[1024], szTmp[128];
	long   i, j, k, Success = 0;
	DWORD  iResult = 0;

	strcpy(szSaveFile, ParticleDataInfo.szFileName);
	strcat(szSaveFile, TEXT(".bmp"));
	if (!WriteDibSectionToDibFile (ParticleDataInfo.hBitmap, szSaveFile))  Success |= 1;

	strcpy(szSaveFile, ParticleDataInfo.szFileName);
	strcat(szSaveFile, TEXT(".dat"));

	hFile = CreateFile(szSaveFile, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
		                 FILE_ATTRIBUTE_ARCHIVE | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		Success |= 2;
		return Success;
	}
	else
	{
		strcpy(szBuf, TEXT("x\ty"));
		if (ParticleDataInfo.Algorithm==IDM_PCOUNT_STREAKS)
			strcat(szBuf, TEXT("\tMax. Hhe\tMax. Breite"));
		strcat(szBuf, TEXT("\tFlche\tMax. GW\tMittl. GW\r\n"));
		if (!WriteFile(hFile, szBuf, strlen(szBuf), &iResult, NULL))
		{
			CloseHandle(hFile);
			Success |= 2;
			return Success;
		}
		for (i=ParticleDataInfo.NumParticles-1; i>=0; i--)
		{
			strcpy(szBuf, gcvt(ParticleDataInfo.Particles[i].X, 10, szTmp));
			strcat(szBuf, TEXT("\t"));
			strcat(szBuf, gcvt(ParticleDataInfo.Particles[i].Y, 10, szTmp));
			if (ParticleDataInfo.Algorithm==IDM_PCOUNT_STREAKS)
			{
				strcat(szBuf, TEXT("\t"));
				strcat(szBuf, ltoa(ParticleDataInfo.Particles[i].BoundRect.top-ParticleDataInfo.Particles[i].BoundRect.bottom+1, szTmp, 10));
				strcat(szBuf, TEXT("\t"));
				strcat(szBuf, ltoa(ParticleDataInfo.Particles[i].BoundRect.right-ParticleDataInfo.Particles[i].BoundRect.left+1, szTmp, 10));
			}
			strcat(szBuf, TEXT("\t"));
      strcat(szBuf, ltoa(ParticleDataInfo.Particles[i].Area, szTmp, 10));
			strcat(szBuf, TEXT("\t"));
      strcat(szBuf, itoa((int) ParticleDataInfo.Particles[i].GrayMax, szTmp, 10));
			strcat(szBuf, TEXT("\t"));
			strcat(szBuf, gcvt(ParticleDataInfo.Particles[i].GrayAvg, 10, szTmp));
			strcat(szBuf, TEXT("\r\n"));
			j=k=0;
			while (szBuf[j])
			{
				if (j!=k)  szBuf[k] = szBuf[j];
				if (szBuf[j]=='.')
				{
					if (ParticleDataInfo.ExportGermanNumberFormat)  szBuf[j]=',';
					if (szBuf[j+1]>='0'  &&  szBuf[j+1]<='9')  k++;
				}
				else k++;
				j++;
			}
			szBuf[k] = '\0';
			if (!WriteFile(hFile, szBuf, strlen(szBuf), &iResult, NULL))
			{
				CloseHandle(hFile);
				Success |= 2;
				return Success;
			}
		}
		CloseHandle(hFile);
	}

	strcpy(szSaveFile, ParticleDataInfo.szFileName);
	strcat(szSaveFile, TEXT(".txt"));

	hFile = CreateFile(szSaveFile, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
		                 FILE_ATTRIBUTE_ARCHIVE | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		Success |= 4;
		return Success;
	}
	else
	{
		switch (ParticleDataInfo.Algorithm)
		{
		case IDM_PCOUNT_MAXIMA:
			strcpy(szBuf, TEXT("Algorithmus:\tMaxima suchen\r\n"));
			break;
		case IDM_PCOUNT_BINARISATION:
			strcpy(szBuf, TEXT("Algorithmus:\tBinarisation.\r\n"));
			break;
		case IDM_PCOUNT_STREAKS:
			strcpy(szBuf, TEXT("Algorithmus:\tStreifen zhlen (mit Binarisation).\r\n"));
			break;
		default:
			strcpy(szBuf, TEXT("Algorithmus:\tName unbekannt - ID-Nr konnte nicht zugeordnet werden!\r\n"));
			break;
		}
		if (!WriteFile(hFile, szBuf, strlen(szBuf), &iResult, NULL))
		{
			CloseHandle(hFile);
			Success |= 4;
			return Success;
		}
		strcpy(szBuf, TEXT("Minimaler Grauwert:\t"));
    strcat(szBuf, itoa((int) ParticleDataInfo.GrayCutoff, szTmp, 10));
		strcat(szBuf, TEXT("\r\n"));
		if (!WriteFile(hFile, szBuf, strlen(szBuf), &iResult, NULL))
		{
			CloseHandle(hFile);
			Success |= 4;
			return Success;
		}
    if ( (ParticleDataInfo.Algorithm == IDM_PCOUNT_MAXIMA) 
			     ||  (ParticleDataInfo.Algorithm == IDM_PCOUNT_BINARISATION)
					 ||  (ParticleDataInfo.Algorithm == IDM_PCOUNT_STREAKS) )
		{
			strcpy(szBuf, TEXT("Minimale Partikelflche:\t"));
		  strcat(szBuf, itoa((int) ParticleDataInfo.MinArea, szTmp, 10));
			strcat(szBuf, TEXT(" Pixel\r\n"));
			if (!WriteFile(hFile, szBuf, strlen(szBuf), &iResult, NULL))
			{
				CloseHandle(hFile);
				Success |= 4;
				return Success;
			}
			strcpy(szBuf, TEXT("Maximale Partikelflche:\t"));
		  strcat(szBuf, itoa((int) ParticleDataInfo.MaxArea, szTmp, 10));
			strcat(szBuf, TEXT(" Pixel\r\n"));
			if (!WriteFile(hFile, szBuf, strlen(szBuf), &iResult, NULL))
			{
				CloseHandle(hFile);
				Success |= 4;
				return Success;
			}
		}
		strcpy(szBuf, TEXT("\r\nZahl der gefundenen Partikel:\t"));
	  strcat(szBuf, ltoa(ParticleDataInfo.NumParticles, szTmp, 10));
		strcat(szBuf, TEXT("\r\n"));
		if (!WriteFile(hFile, szBuf, strlen(szBuf), &iResult, NULL))
		{
			CloseHandle(hFile);
			Success |= 4;
			return Success;
		}
		CloseHandle(hFile);
	}

	return Success;
}
