/*-------------------------------------------------------------
   PTracer.CPP  Dies ist das Hauptprogramm. Die Programmausfhrung
	               beginnt mit der Funktion WinMain, wo das Fenster
								 registriert und angezeigt wird. In dieser Prozedur
								 befindet sich auerdem die Nachrichtenschleife
                (c) Sven Biermann, 2000
  -------------------------------------------------------------*/

#define  WINVER 0x0500
#define  _WIN32_WINNT 0x0500
#include <windows.h>
#include <math.h>
#include <commdlg.h>
#include "DlgProcs.h"
#include "resource.h"
#include "DIBs.h"
#include "FileSys.h"
#include "readIMX.h"

#define UNTITLED  TEXT ("(unbenannt)")


// Funktionsprototypen
LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM);

void SetCaption (HWND hwnd, TCHAR *szTitleName, BOOL bChanged);
BOOL CheckBorder(int &xShow, int &yShow, int cxClient, int cyClient, long ShowWidth, long ShowHeight);
BOOL PaintTheBitmap(HDC hdc, HBITMAP hBitmap, int xShow, int yShow, long ShowWidth, long ShowHeight, int cxClient, int cyClient);
void DrawSelectionRect(HDC hdc, BITMAP bitmap, int xShow, int yShow, long ShowWidth, long ShowHeight, RECT SelRect, BOOL bSelecting);
void DrawLine(HBITMAP hBitmap, RECT SelRect, RECT DrawRect, int PenWidth, int GrayVal, BYTE *pMemBits);


// Globale Variablen
TCHAR szAppName[] = TEXT ("PTRACER") ;
TCHAR szLongAppName[] = TEXT ("Particle Tracer");
UINT  MSWHEEL_ROLLMSG = 0;
HWND  hInfoBox = 0;


int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow)
{
     HWND     hwnd;
		 MSG      msg;
     WNDCLASS wndclass;
		 HACCEL   hAccel;

     wndclass.style         = 0;
     wndclass.lpfnWndProc   = WndProc;
     wndclass.cbClsExtra    = 0;
     wndclass.cbWndExtra    = 0;
     wndclass.hInstance     = hInstance;
     wndclass.hIcon         = LoadIcon (hInstance, MAKEINTRESOURCE(IDI_ICON));
     wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
     wndclass.hbrBackground = (HBRUSH) GetStockObject (GRAY_BRUSH);
     wndclass.lpszMenuName  = szAppName;
     wndclass.lpszClassName = szAppName;

     if (!RegisterClass (&wndclass))
     {    // UNICODE-Compilierung ist die einzige realistische Fehlermglichkeit 
          MessageBox (NULL, TEXT ("Programm arbeitet mit Unicode und setzt Windows NT voraus!"), 
                      szAppName, MB_ICONERROR);
          return 0;
     }

     hwnd = CreateWindow (szAppName, szLongAppName,
                          WS_OVERLAPPEDWINDOW,
                          CW_USEDEFAULT, CW_USEDEFAULT,
                          CW_USEDEFAULT, CW_USEDEFAULT, 
                          NULL, NULL, hInstance, NULL);
		 hInfoBox = CreateDialog (hInstance, TEXT("InfoBox"), hwnd, InfoBoxProc);
//HKEY_LOCAL_MACHINE\Software\PTracer

     ShowWindow (hwnd, iCmdShow);
     UpdateWindow (hwnd);

		 if (szCmdLine[0]=='"')  szCmdLine++;
		 if (szCmdLine[strlen(szCmdLine)-1]=='"')  szCmdLine[strlen(szCmdLine)-1]='\0';
		 if (strlen(szCmdLine)>4)
		 {
       if (!_stricmp(szCmdLine+strlen(szCmdLine)-4,TEXT(".BMP")))
		     PostMessage(hwnd, WM_COMMAND,IDM_FILE_OPENONSTARTUP,(LPARAM) szCmdLine);
			 else if (!_stricmp(szCmdLine+strlen(szCmdLine)-4,TEXT(".IMG")))
		     PostMessage(hwnd, WM_COMMAND,IDM_FILE_IMPORTONSTARTUP,(LPARAM) szCmdLine);
			 else if (!_stricmp(szCmdLine+strlen(szCmdLine)-4,TEXT(".IMX")))
		     PostMessage(hwnd, WM_COMMAND,IDM_FILE_IMPORTONSTARTUP,(LPARAM) szCmdLine);
		 }

		 hAccel = LoadAccelerators(hInstance, TEXT("MyAccelerators"));

		 if (GetSystemMetrics(SM_MOUSEWHEELPRESENT))
			 MSWHEEL_ROLLMSG = 0;
		 else
			 MSWHEEL_ROLLMSG = RegisterWindowMessage(TEXT("MSWHEEL_ROLLMSG"));

     while (GetMessage (&msg, NULL, 0, 0))
     {
			 if (!TranslateAccelerator(hwnd, hAccel, &msg))
			 {
				 if (hInfoBox == 0 || !IsDialogMessage (hInfoBox, &msg))
				 {
           TranslateMessage (&msg);
           DispatchMessage (&msg);
				 }
			 }

     }
     return msg.wParam;
}





/****************************************************************/
/*                                                              */
/*  Fenster-Prozedur des Hauptfensters, Nachrichtenbearbeitung  */
/*                                                              */
/****************************************************************/

LRESULT CALLBACK WndProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	   static HBITMAP      hBitmap = NULL;
		 static BOOL         bGrayDIB = FALSE;
		 HBITMAP             hOtherBitmap;
     static BYTE				 *pBits, *pMemBits;
     static BITMAPINFO*	 pbmi;
		 static RECT         SelRect, XORSelRect, OldSelRect;
		 static int          cxClient, cyClient, xShow, yShow, xMousePos, yMousePos;
     static float				 ShowSizeFactor;
     static long				 ShowWidth, ShowHeight;
		 static boolean			 bChanged = FALSE, bScroll = FALSE, bZoom = FALSE,
			                   bDrawLB = FALSE, bDrawRB = FALSE, bDrawingLine = FALSE,
												 bSelecting = FALSE, bDelOldSelRect = FALSE, bSelMoving = FALSE;
		 static OPENFILENAME ofn;
		 static TCHAR				 szAppPath[MAX_PATH];
     static TCHAR        szFileName [MAX_PATH], szTitleName [MAX_PATH];
		 static TCHAR        szOtherFileName[MAX_PATH];
		 static TCHAR        szRGBPalName[MAX_PATH] = TEXT("");
		 static TCHAR        szRGBPalFilter[] = TEXT ("RGB-Palette (*.RGB)\0*.rgb\0")
                                            TEXT ("Alle Dateien (*.*)\0*.*\0\0");
     static TCHAR        szFilter[] = TEXT ("Bitmap-Dateien (*.BMP)\0*.bmp\0")
                                      TEXT ("Alle Dateien (*.*)\0*.*\0\0");
     static TCHAR        szImportFilter[] = TEXT ("DaVis-Dateien (*.IMX; *.IMG)\0*.imx; *.img\0")
                                            TEXT ("Alle Dateien (*.*)\0*.*\0\0");
     static TCHAR        szLFFilter[] = TEXT ("Lineare Filter-Dateien (*.LFT)\0*.lft\0")
                                        TEXT ("Alle Dateien (*.*)\0*.*\0\0");
		 static HINSTANCE		 hInstance;
		 HMENU							 hMenu;
		 MENUITEMINFO        mii;
     static BITMAP       bitmap;
		 BITMAP              OtherBitmap;
		 HDC                 hdc;
		 RECT                rect;
     PAINTSTRUCT         ps;

		 static SELMANUALPARAM  SelManualParam;
		 static PENCOLORPARAM   PenColorParam;
		 static BITOPPARAM	    BitOpParam;
		 static PALETTEPARAM    PaletteParam;
		 static LOCOPSIZEPARAM  LocOpSizeParam = {0,0,-1,-1,-1,-1};
		 static SPECTRUMPARAM   SpectrumParam;
		 static PARTICLEPARAM   ParticleParam;
		 static LINFILTER       LinFilter;
		 static BITMAPSIZEPARAM ManBitmapSize;
		 static BITMAPSIZEPARAM NewBitmapSize;
		 PPARTICLEDATA          Particles;
		 BOOL   						    bTemp;
		 int								    iResult;
		 static int             PenWidth;
		 double							    Factor;
		 long								    x, y, i, j;
		 long                   *Areas;
		 #define MAXSZTEMP 80
		 TCHAR                  szTemp[MAXSZTEMP];

     switch (message)
     {
     case WM_CREATE:
          hInstance = ((LPCREATESTRUCT) lParam)->hInstance;
			    ofn.lStructSize       = sizeof (OPENFILENAME) ;
          ofn.hwndOwner         = hwnd ;
          ofn.hInstance         = NULL ;
          ofn.lpstrFilter       = szFilter ;
          ofn.lpstrCustomFilter = NULL ;
          ofn.nMaxCustFilter    = 0 ;
          ofn.nFilterIndex      = 0 ;
          ofn.lpstrFile         = szFileName ;
          ofn.nMaxFile          = MAX_PATH ;
          ofn.lpstrFileTitle    = szTitleName ;
          ofn.nMaxFileTitle     = MAX_PATH ;
          ofn.lpstrInitialDir   = NULL ;
          ofn.lpstrTitle        = NULL ;
          ofn.Flags             = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST;
          ofn.nFileOffset       = 0 ;
          ofn.nFileExtension    = 0 ;
          ofn.lpstrDefExt       = TEXT ("bmp") ;
          ofn.lCustData         = 0 ;
          ofn.lpfnHook          = NULL ;
          ofn.lpTemplateName    = NULL ;
					LinFilter.szFilterFile[0] = '\0';
          PenColorParam.LButtonGrayVal = 255;
					PenColorParam.RButtonGrayVal = 0;
					PenWidth = 2;
					NewBitmapSize.OffX       = 0;
					NewBitmapSize.OffY       = 0;
					NewBitmapSize.OffXCenter = TRUE;
					NewBitmapSize.OffYCenter = TRUE;
					NewBitmapSize.Border     = 0;
				  ManBitmapSize.hBitmap   = 0;
					ManBitmapSize.NewWidth  = 1024;
					ManBitmapSize.NewHeight = 1024;
					ManBitmapSize.OffX      = 0;
					ManBitmapSize.OffY      = 0;
					ParticleParam.MinGrayVal = 50;
					ParticleParam.MinArea = 1;
					ParticleParam.MaxArea = 10;
          return 0 ;

     case WM_SIZE:
          bTemp = FALSE;
					if (LOWORD(lParam)>cxClient)  bTemp = TRUE;
					if (HIWORD(lParam)>cyClient)  bTemp = TRUE;
					cxClient = LOWORD (lParam);
          cyClient = HIWORD (lParam);
					if (hBitmap)
						if (CheckBorder(xShow, yShow, cxClient, cyClient, ShowWidth, ShowHeight))
							bTemp = TRUE;
					if (bTemp)  InvalidateRect(hwnd, NULL, FALSE);
          return 0 ;

     case WM_INITMENUPOPUP:
			    if (HIWORD(lParam) == 0)
					{
						mii.cbSize = sizeof(mii);
						mii.fMask = MIIM_TYPE;
						mii.fType = MFT_STRING;
						mii.dwTypeData = MenuUndoText(hBitmap);
						SetMenuItemInfo((HMENU) wParam, IDM_EDIT_UNDO, FALSE, &mii);
						
						EnableMenuItem ((HMENU) wParam, IDM_EDIT_PASTE,
							              IsClipboardFormatAvailable(CF_DIB) ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_EDIT_COPY,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_FILE_SAVE,
							               hBitmap && szFileName[0] ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_FILE_SAVEAS,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_EDIT_UNDO,
							               CanUndo(hBitmap) ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_MOVE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_SELECT,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_DRAW,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_SETORIGIN,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_SELMANUAL,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_SELMEASURE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_PEN_1x1,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_PEN_2x2,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_PEN_3x3,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_PEN_4x4,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_PEN_5x5,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_PEN_6x6,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_PEN_7x7,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_PEN_8x8,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_PEN_9x9,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						CheckMenuItem((HMENU) wParam, IDM_TOOLS_PEN_1x1, (PenWidth == 1) ? MF_CHECKED : MF_UNCHECKED);
				    CheckMenuItem((HMENU) wParam, IDM_TOOLS_PEN_2x2, (PenWidth == 2) ? MF_CHECKED : MF_UNCHECKED);
				    CheckMenuItem((HMENU) wParam, IDM_TOOLS_PEN_3x3, (PenWidth == 3) ? MF_CHECKED : MF_UNCHECKED);
				    CheckMenuItem((HMENU) wParam, IDM_TOOLS_PEN_4x4, (PenWidth == 4) ? MF_CHECKED : MF_UNCHECKED);
				    CheckMenuItem((HMENU) wParam, IDM_TOOLS_PEN_5x5, (PenWidth == 5) ? MF_CHECKED : MF_UNCHECKED);
				    CheckMenuItem((HMENU) wParam, IDM_TOOLS_PEN_6x6, (PenWidth == 6) ? MF_CHECKED : MF_UNCHECKED);
				    CheckMenuItem((HMENU) wParam, IDM_TOOLS_PEN_7x7, (PenWidth == 7) ? MF_CHECKED : MF_UNCHECKED);
				    CheckMenuItem((HMENU) wParam, IDM_TOOLS_PEN_8x8, (PenWidth == 8) ? MF_CHECKED : MF_UNCHECKED);
				    CheckMenuItem((HMENU) wParam, IDM_TOOLS_PEN_9x9, (PenWidth == 9) ? MF_CHECKED : MF_UNCHECKED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_PEN_COLOR,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_FLIPHOR,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_FLIPVER,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_ROTLEFT,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_ROTRIGHT,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_COLORS_SETGRAYPAL,
							               (hBitmap && !bGrayDIB) ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_COLORS_GRAYFROMINTENSITY,
							               (hBitmap && !bGrayDIB) ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_COLORS_REDMAXIMUM,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_COLORS_HEIGHTFIELD,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_COLORS_RAINBOW,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_COLORS_DOUBLERAINBOW,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_COLORS_LOADRGBPAL,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_COLORS_SAVERGBPAL,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_TOOLS_CHANGEBITMAPSIZE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_BITOP_NOT,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_BITOP_AND,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_BITOP_OR,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_BITOP_XOR,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_BITOP_RL,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_BITOP_RR,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_BITOP_SL,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_BITOP_SR,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_BITOP_BI,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_PALETTE,
							               (hBitmap && bGrayDIB) ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_NORMALIZE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_ANDBMPFILE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_ORBMPFILE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_XORBMPFILE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_ADDBMPFILE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_MIXBMPFILE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_SUBBMPFILE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_SUBFROMBMPFILE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_MULBMPFILE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						bTemp = FALSE;
						if (hBitmap  &&  IsClipboardFormatAvailable(CF_DIB))
						{
							GetDIBSizeFromClipboard(hwnd, x, y);
							if (x == bitmap.bmWidth  &&  y == bitmap.bmHeight)  bTemp = TRUE;
						}
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_ANDBMPCLIP,
							               bTemp ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_ORBMPCLIP,
							               bTemp ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_XORBMPCLIP,
							               bTemp ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_ADDBMPCLIP,
							               bTemp ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_MIXBMPCLIP,
							               bTemp ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_SUBBMPCLIP,
							               bTemp ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_SUBFROMBMPCLIP,
							               bTemp ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PNTOP_MULBMPCLIP,
							               bTemp ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_NORMALIZE,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_RANKORDER,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_GRADIENT,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_LINFILTER,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_RECT3X3,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_RECT5X5,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_GAUSS3X3,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_GAUSS5X5,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_UNSHARP3X3,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_UNSHARP5X5,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_DERIVHOR,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_DERIVVER,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_LOCOP_DERIVDIAG,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_GLOBAL_SPECTRUM,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PCOUNT_MAXIMA,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PCOUNT_BINARISATION,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						EnableMenuItem ((HMENU) wParam, IDM_PCOUNT_STREAKS,
							               hBitmap ? MF_ENABLED : MF_GRAYED);
						return 0;
					}
			    break;

		 case WM_COMMAND:
          switch (LOWORD (wParam))
          {
          case IDM_FILE_OPEN:
							 if (hBitmap)
							 {
								 if (bChanged)
								 {
									 if (MessageBox(hwnd, TEXT("Die Datei wurde bearbeitet und noch nicht gespeichert! Wollen Sie wirklich eine andere Datei ffnen?"),
										              TEXT("ParticleTracer - Datei ffnen"), MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2) != IDYES)
													return 0;
								 }
							 }

							 // Standarddialog zum ffnen von Dateien
							 ofn.lpstrFilter = szFilter;
							 ofn.lpstrFile = szFileName;
               ofn.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST;
               if (!GetOpenFileName(&ofn))  return 0 ;
               
               // Wenn momentan ein Bitmap geladen ist, Freigabe
               if (hBitmap)
               {
                    DeleteObject (hBitmap) ;
                    hBitmap = NULL ;
               }

					case IDM_FILE_OPENONSTARTUP:
						   if (LOWORD(wParam)==IDM_FILE_OPENONSTARTUP)  strcpy(szFileName,(LPCSTR) lParam);

							 // DIB-Datei einlesen und DIB Section erzeugen
               SetCursor (LoadCursor (NULL, IDC_WAIT)) ;
               ShowCursor (TRUE) ;

               hBitmap = CreateDibSectionFromDibFile (szFileName, bGrayDIB) ;

               ShowCursor (FALSE) ;
               SetCursor (LoadCursor (NULL, IDC_ARROW)) ;

               // Neuzeichnen des Anwendungsbereichs erzwingen
               InvalidateRect (hwnd, NULL, FALSE) ;
							 ShowSizeFactor=0;
							 bChanged = FALSE;
							 ClrUndoBuf();
							 
               if (hBitmap == NULL)
               {
                 MessageBox (hwnd, TEXT ("Fehler beim Lesen der DIB-Datei. Mglicherweise falsches Format? Dieses Programm verarbeitet nur unkomprimierte Windows-Bitmaps mit 256 Graustufen!"), 
                             TEXT("Particle Tracer - Fehler beim ffnen"), MB_OK | MB_ICONEXCLAMATION) ;
								 szFileName[0] = NULL;
								 SetCaption(hwnd, NULL, bChanged);
								 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_ARROW) );
								 ShowWidth = 0;
								 ShowHeight = 0;
								 return 0;
               }
							 
							 SetCaption (hwnd, szFileName, bChanged);
               GetObject (hBitmap, sizeof (BITMAP), &bitmap) ;
							 pBits = (BYTE*) bitmap.bmBits;
							 ShowWidth = bitmap.bmWidth;
							 ShowHeight = bitmap.bmHeight;
							 xShow=(cxClient-ShowWidth)/2;
							 yShow=(cyClient-ShowHeight)/2;
							 SelRect.left = SelRect.bottom = 0;
							 SelRect.right = bitmap.bmWidth-1;
							 SelRect.top = bitmap.bmHeight-1;
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(hInstance, TEXT("crosscursor")) );
							 return 0 ;
          
          case IDM_FILE_IMPORT:
							 if (hBitmap)
							 {
								 if (bChanged)
								 {
									 if (MessageBox(hwnd, TEXT("Die Datei wurde bearbeitet und noch nicht gespeichert! Wollen Sie wirklich eine andere Datei ffnen?"),
										              TEXT("ParticleTracer - Datei ffnen"), MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2) != IDYES)
													return 0;
								 }
							 }

							 // Standarddialog zum ffnen von Dateien
							 ofn.lpstrFilter = szImportFilter;
							 ofn.lpstrFile = szFileName;
               ofn.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST;
               if (!GetOpenFileName(&ofn))  return 0 ;

					case IDM_FILE_IMPORTONSTARTUP:
						   if (LOWORD(wParam)==IDM_FILE_IMPORTONSTARTUP)  strcpy(szFileName,(LPCSTR) lParam);

               // Sanduhr-Cursor anzeigen
               SetCursor (LoadCursor (NULL, IDC_WAIT)) ;
               ShowCursor (TRUE) ;

							 // IMX- bzw. IMG-Datei in DaVisBuffer einlesen
							 {
							   BufferType DaVisBuffer;
  							 int err = 0;
								 err = ReadIMX (szFileName, &DaVisBuffer, NULL);
								 if (err)
								 {
									 MessageBox(hwnd,TEXT("Fehler beim Einladen der IMX-Datei! Mglicherweise falsches Format?"),
										               TEXT("Particle Tracer - Fehler beim ffnen"),MB_OK);
                   ShowCursor (FALSE) ;
                   SetCursor (LoadCursor (NULL, IDC_ARROW)) ;
									 return 0;
								 }

  							 // Wenn momentan ein Bitmap geladen ist, Freigabe
                 if (hBitmap)  DeleteObject(hBitmap);
                 hBitmap = NewDIB(DaVisBuffer.nx,DaVisBuffer.ny,NULL); bGrayDIB = TRUE;
                 if (hBitmap == NULL)
                 {
                   MessageBox (hwnd, TEXT ("Fehler beim Erstellen der DIB-Datei. Mglicherweise zu wenig freier Arbeitsspeicher?"), 
                                     TEXT("Particle Tracer - Fehler beim ffnen"), MB_OK | MB_ICONEXCLAMATION) ;
					  			 szFileName[0] = NULL;
						  		 SetCaption(hwnd, NULL, bChanged);
								   SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_ARROW) );
								   ShowWidth = 0;
								   ShowHeight = 0;
								   return 0;
                 }
	   						 GetObject(hBitmap, sizeof(BITMAP), &bitmap);
		  					 pBits = (BYTE*) bitmap.bmBits;

								 // Bitmap aus Buffer-Daten erstellen
								 if (DaVisBuffer.isFloat)  //Sind Buffer-Daten array of floats oder words?
								 {
									 float MinFloat = DaVisBuffer.floatArray[0];  // Deklariere MinFloat und initialisiere mit erstem Array-Eintrag
									 float MaxFloat = MinFloat;                   // Deklariere MaxFloat und initialisiere entsprechend
  							   for (y=0; y<DaVisBuffer.ny; y++)             // Durchlaufe nun gesamtes Array 
	  							   for (x=0; x<DaVisBuffer.nx; x++)           // und suche Minimum und Maximum
										 {
											 if (DaVisBuffer.floatArray[(DaVisBuffer.ny-1-y)*DaVisBuffer.nx+x]<MinFloat)
												 MinFloat = DaVisBuffer.floatArray[(DaVisBuffer.ny-1-y)*DaVisBuffer.nx+x];
											 if (DaVisBuffer.floatArray[(DaVisBuffer.ny-1-y)*DaVisBuffer.nx+x]>MaxFloat)
												 MaxFloat = DaVisBuffer.floatArray[(DaVisBuffer.ny-1-y)*DaVisBuffer.nx+x];
										 }
									 if (MaxFloat!=MinFloat)                        // Wenn Maximum und Minimum verschieden,
   							     for (y=0; y<DaVisBuffer.ny; y++)             // durchlaufe nun gesamtes Array 
	  							     for (x=0; x<DaVisBuffer.nx; x++)           // und strecke bzw. stauche Wertebereich auf 0..255
		  						       pBits[y*bitmap.bmWidthBytes+x] = BYTE((DaVisBuffer.floatArray[(DaVisBuffer.ny-1-y)*DaVisBuffer.nx+x]-MinFloat)
											                                       * 255.0f / (MaxFloat-MinFloat));
									 else                                           // ansonsten
   							     for (y=0; y<DaVisBuffer.ny; y++)             // durchlaufe gesamtes Array 
	  							     for (x=0; x<DaVisBuffer.nx; x++)           // und setze Farbwert auf 0
		  						       pBits[y*bitmap.bmWidthBytes+x] = 0;
                   free(DaVisBuffer.floatArray);  //DaVisBuffer wieder freigegeben
								 }
	               else
								 {
									 Word MinWord = DaVisBuffer.wordArray[0];  // Deklariere MinWord und initialisiere mit erstem Array-Eintrag
									 Word MaxWord = MinWord;                   // Deklariere MaxWord und initialisiere entsprechend
  							   for (y=0; y<DaVisBuffer.ny; y++)          // Durchlaufe nun gesamtes Array 
	  							   for (x=0; x<DaVisBuffer.nx; x++)        // und suche Minimum und Maximum
										 {
											 if (DaVisBuffer.wordArray[(DaVisBuffer.ny-1-y)*DaVisBuffer.nx+x]<MinWord)
												 MinWord = DaVisBuffer.wordArray[(DaVisBuffer.ny-1-y)*DaVisBuffer.nx+x];
											 if (DaVisBuffer.wordArray[(DaVisBuffer.ny-1-y)*DaVisBuffer.nx+x]>MaxWord)
												 MaxWord = DaVisBuffer.wordArray[(DaVisBuffer.ny-1-y)*DaVisBuffer.nx+x];
										 }
									 if (MaxWord!=MinWord)                        // Wenn Maximum und Minimum verschieden,
   							     for (y=0; y<DaVisBuffer.ny; y++)             // durchlaufe nun gesamtes Array 
	  							     for (x=0; x<DaVisBuffer.nx; x++)           // und strecke bzw. stauche Wertebereich auf 0..255
		  						       pBits[y*bitmap.bmWidthBytes+x] = BYTE((DaVisBuffer.wordArray[(DaVisBuffer.ny-1-y)*DaVisBuffer.nx+x]-MinWord)
											                                       * (255.0f / (MaxWord-MinWord)));
									 else                                           // ansonsten
   							     for (y=0; y<DaVisBuffer.ny; y++)             // durchlaufe gesamtes Array 
	  							     for (x=0; x<DaVisBuffer.nx; x++)           // und setze Farbwert auf 0
		  						       pBits[y*bitmap.bmWidthBytes+x] = 0;
									 free(DaVisBuffer.wordArray);  //DaVisBuffer wieder freigegeben
								 }
							 } // Gltigkeitsbereich von DaVisBuffer endet hier

               ShowCursor (FALSE) ;
               SetCursor (LoadCursor (NULL, IDC_ARROW)) ;

               // Neuzeichnen des Anwendungsbereichs erzwingen
               InvalidateRect (hwnd, NULL, FALSE) ;
							 ShowSizeFactor=0;
							 bChanged = FALSE;
							 ClrUndoBuf();
							 							 
							 SetCaption (hwnd, szFileName, bChanged);
               GetObject (hBitmap, sizeof (BITMAP), &bitmap) ;
							 pBits = (BYTE*) bitmap.bmBits;
							 ShowWidth = bitmap.bmWidth;
							 ShowHeight = bitmap.bmHeight;
							 xShow=(cxClient-ShowWidth)/2;
							 yShow=(cyClient-ShowHeight)/2;
							 SelRect.left = SelRect.bottom = 0;
							 SelRect.right = bitmap.bmWidth-1;
							 SelRect.top = bitmap.bmHeight-1;
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(hInstance, TEXT("crosscursor")) );
							 return 0;

					case IDM_FILE_SAVEAS:
						   ofn.lpstrFilter = szFilter;
						   ofn.lpstrFile = szFileName;
		           ofn.Flags = OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT;
						   if (!GetSaveFileName (&ofn))  return 0;

					case IDM_FILE_SAVE:
						   if (WriteDibSectionToDibFile (hBitmap, szFileName))
							 {
								 bChanged = FALSE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
							 else
								 MessageBox (hwnd, TEXT("Speichern gescheitert! berprfen Sie den Dateinamen bzw. den freien Datentrgerspeicher!"),
								               TEXT("Particle Tracer - Fehler beim Speichern"), MB_OK | MB_ICONEXCLAMATION);
						   return 0;

					case IDM_EDIT_COPY:
						   if (!hBitmap)  return 0;
               CopyDIBtoClipboard(hwnd, hBitmap, SelRect);
						   return 0;
					
					case IDM_EDIT_PASTE:
							 if (hBitmap)
							 {
								 if (bChanged)
								 {
									 if (MessageBox(hwnd, TEXT("Die Datei wurde bearbeitet und noch nicht gespeichert! Wollen Sie wirklich eine andere Datei aus der Zwischenablage einfgen?"),
										              TEXT("ParticleTracer - Aus der Zwischenablage einfgen"), MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2) != IDYES)
													return 0;
								 }
							 }
               
               
               // Wenn momentan ein Bitmap geladen ist, Freigabe
               if (hBitmap)
               {
                    DeleteObject (hBitmap) ;
                    hBitmap = NULL ;
               }

               // DIB-Datei einlesen und DIB Section erzeugen
               SetCursor (LoadCursor (NULL, IDC_WAIT)) ;
               ShowCursor (TRUE) ;

               hBitmap = PasteDIBfromClipboard (hwnd, bGrayDIB) ;

               ShowCursor (FALSE) ;
               SetCursor (LoadCursor (NULL, IDC_ARROW)) ;

               // Neuzeichnen des Anwendungsbereichs erzwingen
               InvalidateRect (hwnd, NULL, FALSE) ;
							 ShowSizeFactor=0;
							 bChanged = FALSE;
							 ClrUndoBuf();
							 szFileName[0] = NULL;
								 
               if (hBitmap == NULL)
               {
                 MessageBox (hwnd, TEXT ("Fehler beim Lesen der DIB-Datei aus der Zwischenablage. Mglicherweise falsches Format? Dieses Programm verarbeitet nur unkomprimierte Windows-Bitmaps mit 256 Graustufen!"), 
                             TEXT("Particle Tracer - Fehler beim Kopieren aus der Zwischenablage"), MB_OK | MB_ICONEXCLAMATION) ;

								 SetCaption(hwnd, NULL, bChanged);
							   SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_ARROW) );
								 ShowWidth = 0;
								 ShowHeight = 0;
								 return 0;
               }
							 
							 SetCaption (hwnd, szFileName, bChanged);
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(hInstance, TEXT("crosscursor")) );
               GetObject (hBitmap, sizeof (BITMAP), &bitmap) ;
							 pBits = (BYTE*) bitmap.bmBits;
							 ShowWidth = bitmap.bmWidth;
							 ShowHeight = bitmap.bmHeight;
							 xShow=(cxClient-ShowWidth)/2;
							 yShow=(cyClient-ShowHeight)/2;
							 SelRect.left = SelRect.bottom = 0;
							 SelRect.right = bitmap.bmWidth-1;
							 SelRect.top = bitmap.bmHeight-1;
							 return 0;

					case IDM_EDIT_UNDO:
						   Undo(hBitmap);
							 {
								 RGBQUAD RGBPal[256];
								 GetRGBPalette(hBitmap, RGBPal);
								 bGrayDIB = TRUE;
								 for (i=0; i<256; i++)
								 {
									 if (RGBPal[i].rgbRed   != i)  break;
									 if (RGBPal[i].rgbGreen != i)  break;
									 if (RGBPal[i].rgbBlue  != i)  break;
								 }
								 if (i<256)  bGrayDIB = FALSE;
							 }
							 InvalidateRect(hwnd, NULL, FALSE);
						   return 0;
					
					case IDM_TOOLS_MOVE:
						   hMenu = GetMenu(hwnd);
						   CheckMenuItem(hMenu, IDM_TOOLS_MOVE, MF_CHECKED);
						   CheckMenuItem(hMenu, IDM_TOOLS_DRAW, MF_UNCHECKED);
						   CheckMenuItem(hMenu, IDM_TOOLS_SELECT, MF_UNCHECKED);
							 CheckMenuItem(hMenu, IDM_TOOLS_SETORIGIN, MF_UNCHECKED);
						   return 0;

					case IDM_TOOLS_SELECT:
							 hMenu = GetMenu(hwnd);
						   CheckMenuItem(hMenu, IDM_TOOLS_MOVE, MF_UNCHECKED);
						   CheckMenuItem(hMenu, IDM_TOOLS_DRAW, MF_UNCHECKED);
						   CheckMenuItem(hMenu, IDM_TOOLS_SELECT, MF_CHECKED);
							 CheckMenuItem(hMenu, IDM_TOOLS_SETORIGIN, MF_UNCHECKED);
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(hInstance, TEXT("crosscursor")) );
						   return 0;
					
					case IDM_TOOLS_DRAW:
						   hMenu = GetMenu(hwnd);
						   CheckMenuItem(hMenu, IDM_TOOLS_MOVE, MF_UNCHECKED);
						   CheckMenuItem(hMenu, IDM_TOOLS_DRAW, MF_CHECKED);
						   CheckMenuItem(hMenu, IDM_TOOLS_SELECT, MF_UNCHECKED);
							 CheckMenuItem(hMenu, IDM_TOOLS_SETORIGIN, MF_UNCHECKED);
						   return 0;

					case IDM_TOOLS_SETORIGIN:
						   hMenu = GetMenu(hwnd);
						   CheckMenuItem(hMenu, IDM_TOOLS_MOVE, MF_UNCHECKED);
						   CheckMenuItem(hMenu, IDM_TOOLS_DRAW, MF_UNCHECKED);
						   CheckMenuItem(hMenu, IDM_TOOLS_SELECT, MF_UNCHECKED);
							 CheckMenuItem(hMenu, IDM_TOOLS_SETORIGIN, MF_CHECKED);
						   return 0;

					case IDM_TOOLS_SELMANUAL:
							 OldSelRect = SelRect;
							 SelManualParam.hBitmap = hBitmap;
							 SelManualParam.bFFT    = FALSE;
							 SelManualParam.pRect = &SelRect;  // SelManualParam.LastRect=SelRect;
							 if (!DialogBoxParam(hInstance, TEXT("SelectManualDlg"), hwnd, SelectManualDlgProc, (LPARAM) &SelManualParam))
							 {
								 SelRect = OldSelRect;
								 InvalidateRect(hwnd, NULL, FALSE);
							 }
							 else
							 {
								 SelManualParam.LastRect = SelRect;
							 }
						   return 0;
					
					case IDM_TOOLS_SELMEASURE:
						{
						   int    decimal, sign;
							 BYTE   Min, Max, Act;
							 double Mean, Sum;
							 TCHAR  szMsg[250], szNum[20], *szDbl;
							 szMsg[0]='\0';    szNum[0] = '\0';

							 i = SelRect.right-SelRect.left+1;
						   j = SelRect.top-SelRect.bottom+1;
							 strncat(szMsg, TEXT("Hhe:\t"), 100);
							 ltoa(j, szNum, 10);
							 strncat(szMsg, szNum, 100);
							 strncat(szMsg, TEXT("\nBreite:\t"), 100);
							 ltoa(i, szNum, 10);
							 strncat(szMsg, szNum, 100);
							 strncat(szMsg, TEXT("\nFlche:\t"), 100);
							 ltoa(i*j, szNum, 10);
							 strncat(szMsg, szNum, 100);
							 strncat(szMsg, TEXT("\nDiag.:\t"), 100);
							 szDbl = fcvt(hypot((double) i, (double) j), 2, &decimal, &sign);
							 strncat(szMsg, szDbl, decimal);
							 strncat(szMsg, TEXT("."), 1);
							 strncat(szMsg, szDbl+decimal, 100);

							 if (bGrayDIB)
							 {
								 Mean=0; Min=Max=pBits[SelRect.bottom*bitmap.bmWidthBytes+SelRect.left];
								 for (x=SelRect.left; x<=SelRect.right; x++)
								 {
									 Sum=0;
									 for (y=SelRect.bottom; y<=SelRect.top; y++)
									 {
										 Act = pBits[y*bitmap.bmWidthBytes+x];
										 Min = (Act<Min) ? Act : Min;
										 Max = (Act>Max) ? Act : Max;
										 Sum+= (double) Act;
									 }
									 Sum/=(double) j;
									 Mean+=Sum;
								 }
								 Mean/=(double) i;
								 strncat(szMsg, TEXT("\n\nGrauwerte:\nMinimal:\t"), 100);
								 ltoa(Min, szNum, 10);
								 strncat(szMsg, szNum, 100);
								 strncat(szMsg, TEXT("\nMittel:\t"), 100);
								 szDbl = fcvt(Mean, 2, &decimal, &sign);
								 strncat(szMsg, szDbl, decimal);
								 strncat(szMsg, TEXT("."), 1);
								 strncat(szMsg, szDbl+decimal, 100);
								 strncat(szMsg, TEXT("\nMaximal:\t"), 100);
								 ltoa(Max, szNum, 10);
								 strncat(szMsg, szNum, 100);
							 }
							 else
								 strncat(szMsg, TEXT("\n\nKeine reine Grauwertetabelle!"), 100);
							 MessageBox(hwnd, szMsg, TEXT("Particle Tracer - Maxima zhlen"), MB_OK | MB_ICONINFORMATION);
						   return 0;
						}

					case IDM_TOOLS_PEN_1x1:
						   PenWidth = 1;
						   return 0;
					case IDM_TOOLS_PEN_2x2:
						   PenWidth = 2;
						   return 0;
					case IDM_TOOLS_PEN_3x3:
						   PenWidth = 3;
						   return 0;
					case IDM_TOOLS_PEN_4x4:
						   PenWidth = 4;
						   return 0;
					case IDM_TOOLS_PEN_5x5:
						   PenWidth = 5;
						   return 0;
					case IDM_TOOLS_PEN_6x6:
						   PenWidth = 6;
						   return 0;
					case IDM_TOOLS_PEN_7x7:
						   PenWidth = 7;
						   return 0;
					case IDM_TOOLS_PEN_8x8:
						   PenWidth = 8;
						   return 0;
					case IDM_TOOLS_PEN_9x9:
						   PenWidth = 9;
						   return 0;

					case IDM_TOOLS_PEN_COLOR:
						   DialogBoxParam(hInstance, TEXT("PenColorDlg"), hwnd, PenColorDlgProc, (LPARAM) &PenColorParam);
						   return 0;

					case IDM_TOOLS_FLIPHOR:
							 if (!hBitmap)  return 0;
							 ClrUndoBuf();
							 FlipHorizontal(hBitmap);
							 x = SelRect.left;
							 SelRect.left  = bitmap.bmWidth - 1 - SelRect.right;
							 SelRect.right = bitmap.bmWidth - 1 - x;
						   if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
							 InvalidateRect(hwnd, NULL, FALSE);
							 return 0;

					case IDM_TOOLS_FLIPVER:
							 if (!hBitmap)  return 0;
							 ClrUndoBuf();
							 FlipVertical(hBitmap);
							 y = SelRect.top;
							 SelRect.top  = bitmap.bmHeight - 1 - SelRect.bottom;
							 SelRect.bottom = bitmap.bmHeight - 1 - y;
						   if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
							 InvalidateRect(hwnd, NULL, FALSE);
							 return 0;

					case IDM_TOOLS_ROTLEFT:
							 if (!hBitmap)  return 0;
							 if (RotateLeft(hBitmap))
							 {
								 GetObject(hBitmap, sizeof(BITMAP), &bitmap);
								 pBits = (BYTE*) bitmap.bmBits;
								 x = ShowWidth;
								 ShowWidth = ShowHeight;
								 ShowHeight = x;
								 x = xShow;
								 if (cxClient<ShowWidth)
									 xShow = yShow + (cxClient-cyClient)/2;
								 else
									 xShow = xShow + (ShowHeight-ShowWidth)/2;
								 if (cyClient<ShowHeight)
									 yShow = cxClient-x-ShowHeight + (cyClient-cxClient)/2;
								 else
									 yShow = yShow + (ShowWidth-ShowHeight)/2;
								 CheckBorder(xShow, yShow, cxClient, cyClient, ShowWidth, ShowHeight);
								 ClrUndoBuf();
								 OldSelRect = SelRect;
								 SelRect.bottom = OldSelRect.left;
								 SelRect.right  = bitmap.bmWidth - OldSelRect.bottom - 1;
								 SelRect.top    = OldSelRect.right;
								 SelRect.left   = bitmap.bmWidth - OldSelRect.top - 1;
								 OldSelRect = SelRect;
								 if (!bChanged)
								 {
									 bChanged = TRUE;
									 SetCaption(hwnd, szFileName, bChanged);
								 }
								 InvalidateRect(hwnd, NULL, FALSE);
							 }
						   return 0;

					case IDM_TOOLS_ROTRIGHT:
							 if (!hBitmap)  return 0;
							 if (RotateRight(hBitmap))
							 {
								 GetObject(hBitmap, sizeof(BITMAP), &bitmap);
								 pBits = (BYTE*) bitmap.bmBits;
								 x = ShowWidth;
								 ShowWidth = ShowHeight;
								 ShowHeight = x;
								 x = xShow;
								 if (cxClient<ShowWidth)
									 xShow = cyClient-yShow-ShowWidth + (cxClient-cyClient)/2;
								 else
									 xShow = xShow + (ShowHeight-ShowWidth)/2;
								 if (cyClient<ShowHeight)
									 yShow = x + (cyClient-cxClient)/2;
								 else
									 yShow = yShow + (ShowWidth-ShowHeight)/2;
								 CheckBorder(xShow, yShow, cxClient, cyClient, ShowWidth, ShowHeight);
								 ClrUndoBuf();
								 OldSelRect = SelRect;
								 SelRect.bottom = bitmap.bmHeight - OldSelRect.right - 1;
								 SelRect.right  = OldSelRect.top;
								 SelRect.top    = bitmap.bmHeight - OldSelRect.left - 1;
								 SelRect.left   = OldSelRect.bottom;
								 OldSelRect = SelRect;
								 if (!bChanged)
								 {
									 bChanged = TRUE;
									 SetCaption(hwnd, szFileName, bChanged);
								 }
								 InvalidateRect(hwnd, NULL, FALSE);
							 }
						   return 0;

					case IDM_TOOLS_ROT180DEG:
					case IDM_TOOLS_FLIPPOINT:
							 if (!hBitmap)  return 0;
							 ClrUndoBuf();
							 Rotate180Degrees(hBitmap);
							 OldSelRect = SelRect;
							 x = SelRect.left;
							 SelRect.left  = bitmap.bmWidth - 1 - SelRect.right;
							 SelRect.right = bitmap.bmWidth - 1 - x;
							 y = SelRect.top;
							 SelRect.top  = bitmap.bmHeight - 1 - SelRect.bottom;
							 SelRect.bottom = bitmap.bmHeight - 1 - y;
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
							 InvalidateRect(hwnd, NULL, FALSE);
						   return 0;

					case IDM_COLORS_SETGRAYPAL:
							 if (!hBitmap)  return 0;
							 SaveForUndo(hBitmap, TEXT("Grauwerte-Palette einsetzen"));
							 SetStandardPalette(hBitmap);
							 InvalidateRect(hwnd, NULL, FALSE);
							 bGrayDIB = TRUE;
						   return 0;

 					case IDM_COLORS_GRAYFROMINTENSITY:
							 if (!hBitmap)  return 0;
							 SaveForUndo(hBitmap, TEXT("Grauwerte aus Intensitt berechnen"));
						   // Paletten vergleichen
							 {
								 // Paletten vergleichen...
								 RGBQUAD ActPal[256];
								 BYTE    PaletteEntry[256];
								 GetRGBPalette(hBitmap, ActPal);
								 for (i=0; i<256; i++)
								 {
								 	 PaletteEntry[i] = (BYTE) (((unsigned long) ActPal[i].rgbBlue
									 											    + (unsigned long) ActPal[i].rgbGreen
																				    + (unsigned long) ActPal[i].rgbRed) / 3);
								 } 
								 // Pixeldaten kopieren
								 for (i=0; i<bitmap.bmWidthBytes*bitmap.bmHeight; i++)
								 {
									 pBits[i] = PaletteEntry[pBits[i]];
								 }
								 SetStandardPalette(hBitmap);
							 }
							 InvalidateRect(hwnd, NULL, FALSE);
							 bGrayDIB = TRUE;
						   return 0;

					case IDM_COLORS_REDMAXIMUM:
							 if (!hBitmap)  return 0;
							 SaveForUndo(hBitmap, TEXT("Obersten Palettenwert rot markieren"));
							 SetSingleRGBPaletteEntry(hBitmap, 255, RGB(255,0,0));							 
							 InvalidateRect(hwnd, NULL, FALSE);
							 bGrayDIB = FALSE;
						   return 0;

					case IDM_COLORS_RAINBOW:
							 if (!hBitmap)  return 0;
							 SaveForUndo(hBitmap, TEXT("Regenbogen"));
							 {
								 RGBQUAD NewPal[256];
								 for (i=0; i<64; i++)
								 {
									 NewPal[i].rgbBlue  = BYTE(i * 4);
									 NewPal[i].rgbGreen = 0;
									 NewPal[i].rgbRed   = 0;
									 NewPal[i].rgbReserved = 0;
								 }
								 for (i=64; i<128; i++)
								 {
									 NewPal[i].rgbBlue  = BYTE(255-4*(i-64));
									 NewPal[i].rgbGreen = BYTE((i-64)*4);
									 NewPal[i].rgbRed   = 0;
									 NewPal[i].rgbReserved = 0;
								 }
								 for (i=128; i<192; i++)
								 {
									 NewPal[i].rgbBlue  = 0;
									 NewPal[i].rgbGreen = 255;
									 NewPal[i].rgbRed   = BYTE((i-128)*4);
									 NewPal[i].rgbReserved = 0;
								 }
								 for (i=192; i<256; i++)
								 {
									 NewPal[i].rgbBlue  = 0;
									 NewPal[i].rgbGreen = BYTE(255-(i-192)*4);
									 NewPal[i].rgbRed   = 255;
									 NewPal[i].rgbReserved = 0;
								 }
								 SetRGBPalette(hBitmap, NewPal);							 
							 }
							 InvalidateRect(hwnd, NULL, FALSE);
							 bGrayDIB = FALSE;
						   return 0;

					case IDM_COLORS_DOUBLERAINBOW:
							 if (!hBitmap)  return 0;
							 SaveForUndo(hBitmap, TEXT("Doppelter Regenbogen"));
							 {
								 RGBQUAD NewPal[256];
								 for (i=0; i<32; i++)
								 {
									 NewPal[i+128].rgbBlue  = NewPal[i].rgbBlue  = BYTE(i * 8);
									 NewPal[i+128].rgbGreen = NewPal[i].rgbGreen = 0;
									 NewPal[i].rgbRed = 0;
									 NewPal[i+128].rgbRed = BYTE(255-8*i);
									 NewPal[i+128].rgbReserved = NewPal[i].rgbReserved = 0;
								 }
								 for (i=32; i<64; i++)
								 {
									 NewPal[i+128].rgbBlue  = NewPal[i].rgbBlue  = BYTE(255-8*(i-32));
									 NewPal[i+128].rgbGreen = NewPal[i].rgbGreen = BYTE((i-32)*8);
									 NewPal[i+128].rgbRed   = NewPal[i].rgbRed   = 0;
									 NewPal[i+128].rgbReserved = NewPal[i].rgbReserved = 0;
								 }
								 for (i=64; i<96; i++)
								 {
									 NewPal[i+128].rgbBlue  = NewPal[i].rgbBlue  = 0;
									 NewPal[i+128].rgbGreen = NewPal[i].rgbGreen = 255;
									 NewPal[i+128].rgbRed   = NewPal[i].rgbRed   = BYTE((i-64)*8);
									 NewPal[i+128].rgbReserved = NewPal[i].rgbReserved = 0;
								 }
								 for (i=96; i<128; i++)
								 {
									 NewPal[i+128].rgbBlue  = NewPal[i].rgbBlue  = 0;
									 NewPal[i+128].rgbGreen = NewPal[i].rgbGreen = BYTE(255-(i-96)*8);
									 NewPal[i+128].rgbRed   = NewPal[i].rgbRed   = 255;
									 NewPal[i+128].rgbReserved = NewPal[i].rgbReserved = 0;
								 }
								 SetRGBPalette(hBitmap, NewPal);							 
							 }
							 InvalidateRect(hwnd, NULL, FALSE);
							 bGrayDIB = FALSE;
						   return 0;

					case IDM_COLORS_HEIGHTFIELD:
							 if (!hBitmap)  return 0;
							 SaveForUndo(hBitmap, TEXT("Hhenlinien"));
							 {
								 RGBQUAD NewPal[256];
								 for (i=0; i<256; i++)
								 {
									 NewPal[i].rgbBlue  = BYTE(i);
									 NewPal[i].rgbGreen = BYTE(i);
									 NewPal[i].rgbRed   = BYTE(i);
									 NewPal[i].rgbReserved = 0;
								 }
								 j=0;
								 for (i=0; i<12; i++)
								 {
									 NewPal[253-i*20].rgbRed   = NewPal[254-i*20].rgbRed   = NewPal[255-i*20].rgbRed   = (j==0 || j==4 || j==5) ? 255 : 0;
									 NewPal[253-i*20].rgbGreen = NewPal[254-i*20].rgbGreen = NewPal[255-i*20].rgbGreen = (j==1 || j==3 || j==5) ? 255 : 0;
									 NewPal[253-i*20].rgbBlue  = NewPal[254-i*20].rgbBlue  = NewPal[255-i*20].rgbBlue  = (j==2 || j==3 || j==4) ? 255 : 0;
									 j++; j%=6;
								 }
								 SetRGBPalette(hBitmap, NewPal);							 
							 }
							 InvalidateRect(hwnd, NULL, FALSE);
							 bGrayDIB = FALSE;
						   return 0;

					case IDM_COLORS_LOADRGBPAL:
							 if (!hBitmap)  return 0;
							 // Welche Datei...?
							 ofn.lpstrFilter = szRGBPalFilter;
							 ofn.lpstrFile = szRGBPalName;
		           ofn.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST;
			         if (!GetOpenFileName(&ofn))
				         return 0 ;
							 // Wenn nicht abgebrochen...
							 {
								 RGBQUAD NewRGBPal[256];
	
								 // Versuche einzuladen
								 if (!LoadRGBPalette(szRGBPalName, NewRGBPal))
								 {
									 // Wenn Fehler: Meldung anzeigen und Routine verlassen
									 MessageBox(hwnd, TEXT("RGB-Palette einladen fehlgeschlagen! Falsches Dateiformat?"),
									                  TEXT("Fehler beim Einladen der RGB-Palette"), MB_OK | MB_ICONEXCLAMATION);
									 return 0;
								 }

								 // Ansonsten:
								 SaveForUndo(hBitmap, TEXT("Falschfarb-Palette einladen"));
								 SetRGBPalette(hBitmap, NewRGBPal);
							 }
							 InvalidateRect(hwnd, NULL, FALSE);
							 bGrayDIB = FALSE;
						   return 0;

					case IDM_COLORS_SAVERGBPAL:
							 if (!hBitmap)  return 0;
							 // Welche Datei...?
							 ofn.lpstrFilter = szRGBPalFilter;
							 ofn.lpstrFile = szRGBPalName;
		           ofn.Flags = OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT;
						   if (!GetSaveFileName (&ofn))  return 0;
							 // Wenn nicht abgebrochen, versuche Palette abzuspeichern...
							 if (!SaveRGBPalette(hBitmap, szRGBPalName))  // Wenn Fehler: Meldung anzeigen und Routine verlassen
								 MessageBox(hwnd, TEXT("Exportieren der RGB-Palette fehlgeschlagen! Falsches Dateiformat?"),
									                TEXT("Fehler beim Exportieren der RGB-Palette"), MB_OK | MB_ICONEXCLAMATION);
						   return 0;

					case IDM_TOOLS_CHANGEBITMAPSIZE:
						   if (hBitmap == NULL) return 0;
							 NewBitmapSize.hBitmap    = hBitmap;
							 NewBitmapSize.NewWidth   = bitmap.bmWidth;
							 NewBitmapSize.NewHeight  = bitmap.bmHeight;
							 if (!DialogBoxParam(hInstance, TEXT("BitmapSizeDlg"), hwnd, BitmapSizeDlgProc, (LPARAM) &NewBitmapSize))  return 0;
							 x = NewBitmapSize.OffXCenter ? (NewBitmapSize.NewWidth -bitmap.bmWidth )/2 : NewBitmapSize.OffX;
							 y = NewBitmapSize.OffYCenter ? (NewBitmapSize.NewHeight-bitmap.bmHeight)/2 : NewBitmapSize.OffY;
							 if (ChangeBitmapSize(hBitmap, TRUE,
																						 NewBitmapSize.NewWidth, NewBitmapSize.NewHeight,
								                             x, y,
																						 NewBitmapSize.Border))
							 {
								 GetObject(hBitmap, sizeof(BITMAP), &bitmap);
								 pBits = (BYTE*) bitmap.bmBits;
								 Factor = ((double) (cxClient/2-xShow)) / ((double) ShowWidth);
								 ShowWidth = ShowSizeFactor<0 ? long(-bitmap.bmWidth/(ShowSizeFactor-1)) : long(bitmap.bmWidth*(ShowSizeFactor+1));
								 Factor = ((double) (cyClient/2-yShow)) / ((double) ShowHeight);
								 ShowHeight = ShowSizeFactor<0 ? long(-bitmap.bmHeight/(ShowSizeFactor-1)) : long(bitmap.bmHeight*(ShowSizeFactor+1));
							   xShow=(cxClient-ShowWidth)/2;
							   yShow=(cyClient-ShowHeight)/2;
								 ClrUndoBuf();
								 SelRect.bottom += y;
								 SelRect.right  += x;
								 SelRect.top    += y;
								 SelRect.left   += x;
								 if (SelRect.bottom<0  ||  SelRect.left<0
									       ||  SelRect.right>=bitmap.bmWidth  ||  SelRect.top>=bitmap.bmHeight)
								 {
									 SelRect.bottom = 0;
									 SelRect.right  = bitmap.bmWidth -1;
									 SelRect.top    = bitmap.bmHeight-1;
									 SelRect.left   = 0;
								 }
								 OldSelRect = SelRect;
								 if (!bChanged)
								 {
									 bChanged = TRUE;
									 SetCaption(hwnd, szFileName, bChanged);
								 }
								 InvalidateRect(hwnd, NULL, FALSE);
							 }
							 return 0;

					case IDM_BITOP_AND:
					case IDM_BITOP_OR:
					case IDM_BITOP_XOR:
						   if (hBitmap == NULL) return 0;
							 BitOpParam.hBitmap = hBitmap;
							 BitOpParam.Operation = LOWORD(wParam);
							 GetMenuString(GetMenu(hwnd), LOWORD(wParam), BitOpParam.szMenuString, 80, MF_BYCOMMAND);
							 if (DialogBoxParam(hInstance, TEXT("BitsDlg"), hwnd, BitsDlgProc, (LPARAM) &BitOpParam))
							   if (!bChanged)
								 {
									 bChanged = TRUE;
									 SetCaption(hwnd, szFileName, bChanged);
								 }
							 return 0;

					case IDM_BITOP_NOT:
					case IDM_BITOP_RR:
					case IDM_BITOP_RL:
					case IDM_BITOP_SR:
					case IDM_BITOP_SL:
					case IDM_BITOP_BI:
						   if (hBitmap == NULL)  return 0;
							 GetMenuString(GetMenu(hwnd), LOWORD(wParam), szTemp, MAXSZTEMP, MF_BYCOMMAND);
							 SaveForUndo(hBitmap, szTemp);
							 if (CalcBitOp(hBitmap, LOWORD(wParam), 0))
							 {
								 if (!bChanged)
								 {
									 bChanged = TRUE;
									 SetCaption(hwnd, szFileName, bChanged);
								 }
								 InvalidateRect(hwnd,0,FALSE);
							 }
							 return 0;

					case IDM_PNTOP_PALETTE:
						   if (hBitmap == NULL) return 0;
							 PaletteParam.hBitmap = hBitmap;
							 PaletteParam.hInfoBox = hInfoBox;
						   if (DialogBoxParam(hInstance, TEXT("PaletteDlg"), hwnd, PaletteDlgProc, (LPARAM) &PaletteParam))
							 {
								 if (!bChanged)
								 {
									 bChanged = TRUE;
									 SetCaption(hwnd, szFileName, bChanged);
								 }
							 };
							 return 0;

					case IDM_PNTOP_NORMALIZE:
						   if (!hBitmap)  return 0;
							 SaveForUndo(hBitmap, TEXT("Normalisieren"));
							 if (!CalcNormalize(hBitmap))  return 0;
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
							 return 0;

					case IDM_PNTOP_ANDBMPFILE:
					case IDM_PNTOP_ORBMPFILE:
					case IDM_PNTOP_XORBMPFILE:
					case IDM_PNTOP_ADDBMPFILE:
					case IDM_PNTOP_MIXBMPFILE:
					case IDM_PNTOP_SUBBMPFILE:
					case IDM_PNTOP_SUBFROMBMPFILE:
					case IDM_PNTOP_MULBMPFILE:
							 if (!hBitmap)  return 0;
						   // Standarddialog zum ffnen von Dateien
							 ofn.lpstrFilter = szFilter;
						   ofn.lpstrFile = szOtherFileName;
               ofn.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST;
               if (!GetOpenFileName(&ofn))
                    return 0;
               // DIB-Datei einlesen und DIB Section erzeugen
               SetCursor (LoadCursor (NULL, IDC_WAIT)) ;
               ShowCursor (TRUE) ;
               hOtherBitmap = CreateDibSectionFromDibFile (szOtherFileName, bTemp);
							 if (!bTemp)
							 {
								 if (MessageBox(hwnd, TEXT("Achtung: Die ausgewhlte Bitmap besteht nicht nur aus Graustufen!\n")
									                    TEXT("Gerechnet wird nur mit den Palettenindizes und NICHT mit den ")
																			TEXT("tatschlichen Farben!\n\nWollen Sie trotzdem fortfahren?"), TEXT("Bild einladen - Keine Graustufen"), MB_YESNO | MB_ICONEXCLAMATION) != IDYES)
								 {
									 ShowCursor(FALSE);
									 return 0;
								 }
							 }
               ShowCursor (FALSE) ;
               SetCursor (LoadCursor (NULL, IDC_ARROW)) ;
					case IDM_PNTOP_ANDBMPCLIP:
					case IDM_PNTOP_ORBMPCLIP:
					case IDM_PNTOP_XORBMPCLIP:
					case IDM_PNTOP_ADDBMPCLIP:
					case IDM_PNTOP_MIXBMPCLIP:
					case IDM_PNTOP_SUBBMPCLIP:
					case IDM_PNTOP_SUBFROMBMPCLIP:
					case IDM_PNTOP_MULBMPCLIP:
						   if (!hBitmap)  return 0;
						   if (LOWORD(wParam)==IDM_PNTOP_ANDBMPCLIP  ||  LOWORD(wParam)==IDM_PNTOP_ORBMPCLIP  ||  LOWORD(wParam)==IDM_PNTOP_XORBMPCLIP
								     ||  LOWORD(wParam)==IDM_PNTOP_SUBFROMBMPCLIP  ||  LOWORD(wParam)==IDM_PNTOP_SUBBMPCLIP  ||  LOWORD(wParam)==IDM_PNTOP_ADDBMPCLIP
								     ||  LOWORD(wParam)==IDM_PNTOP_MULBMPCLIP  ||  LOWORD(wParam)==IDM_PNTOP_MIXBMPCLIP)
							 {
								 hOtherBitmap = PasteDIBfromClipboard(hwnd, bTemp);
								 if (!bTemp)
								 {
									 if (MessageBox(hwnd, TEXT("Achtung: Die ausgewhlte Bitmap besteht nicht nur aus Graustufen!\n")
										                    TEXT("Gerechnet wird nur mit den Palettenindizes und NICHT mit den ")
																				TEXT("tatschlichen Farben!\n\nWollen Sie trotzdem fortfahren?"), TEXT("Bild aus Zwischenablage - Keine Graustufen"), MB_YESNO | MB_ICONEXCLAMATION) != IDYES)
										 return 0;
								 }
							 }
               if (hOtherBitmap == NULL)
               {
                 MessageBox (hwnd, TEXT ("Fehler beim Lesen der DIB-Datei. Mglicherweise falsches Format? Dieses Programm verarbeitet nur unkomprimierte Windows-Bitmaps mit 256 Graustufen!"), 
                             TEXT("Particle Tracer - Fehler beim ffnen"), MB_OK | MB_ICONEXCLAMATION) ;
								 return 0;
               }
							 // Richtige Gre?
							 GetObject(hOtherBitmap, sizeof(BITMAP), &OtherBitmap);
							 if (bitmap.bmWidth!=OtherBitmap.bmWidth  ||  bitmap.bmHeight!=OtherBitmap.bmHeight)
							 {
								 DeleteObject(hOtherBitmap);
                 MessageBox (hwnd, TEXT ("Die beiden Bitmaps haben unterschiedliche Gren! Addition, Subtraktion usw. dieser Bitmaps ist daher nicht mglich!"), 
                             TEXT("Particle Tracer - Fehler beim ffnen"), MB_OK | MB_ICONEXCLAMATION) ;
								 return 0;
							 }
							 pMemBits = (BYTE*) OtherBitmap.bmBits;
							 // Fr Undo speichern und Berechnung durchfhren
							 if (LOWORD(wParam)==IDM_PNTOP_ANDBMPFILE  ||  LOWORD(wParam)==IDM_PNTOP_ANDBMPCLIP)
							 {
								 SaveForUndo(hBitmap, TEXT("AND mit anderer Bitmap"));
								 for (y=0; y<bitmap.bmHeight; y++)
								 {
									 i = y*bitmap.bmWidthBytes;
									 for (x=0; x<bitmap.bmWidth; x++)
									 {
										 pBits[i] = pBits[i] & pMemBits[i];
										 i++;
									 }
								 }
							 }
							 if (LOWORD(wParam)==IDM_PNTOP_ORBMPFILE  ||  LOWORD(wParam)==IDM_PNTOP_ORBMPCLIP)
							 {
								 SaveForUndo(hBitmap, TEXT("OR mit anderer Bitmap"));
								 for (y=0; y<bitmap.bmHeight; y++)
								 {
									 i = y*bitmap.bmWidthBytes;
									 for (x=0; x<bitmap.bmWidth; x++)
									 {
										 pBits[i] = pBits[i] | pMemBits[i];
										 i++;
									 }
								 }
							 }
							 if (LOWORD(wParam)==IDM_PNTOP_XORBMPFILE  ||  LOWORD(wParam)==IDM_PNTOP_XORBMPCLIP)
							 {
								 SaveForUndo(hBitmap, TEXT("XOR mit anderer Bitmap"));
								 for (y=0; y<bitmap.bmHeight; y++)
								 {
									 i = y*bitmap.bmWidthBytes;
									 for (x=0; x<bitmap.bmWidth; x++)
									 {
										 pBits[i] = pBits[i] ^ pMemBits[i];
										 i++;
									 }
								 }
							 }
							 if (LOWORD(wParam)==IDM_PNTOP_ADDBMPFILE  ||  LOWORD(wParam)==IDM_PNTOP_ADDBMPCLIP)
							 {
								 SaveForUndo(hBitmap, TEXT("Andere Bitmap addieren"));
								 for (y=0; y<bitmap.bmHeight; y++)
								 {
									 i = y*bitmap.bmWidthBytes;
									 for (x=0; x<bitmap.bmWidth; x++)
									 {
										 iResult =  pBits[i];
										 iResult += pMemBits[i];
										 if (iResult>255)  iResult = 255;
										 pBits[i] = iResult;
										 i++;
									 }
								 }
							 }
							 if (LOWORD(wParam)==IDM_PNTOP_MIXBMPFILE  ||  LOWORD(wParam)==IDM_PNTOP_MIXBMPCLIP)
							 {
								 SaveForUndo(hBitmap, TEXT("Mit anderer Bitmap mischen"));
								 for (y=0; y<bitmap.bmHeight; y++)
								 {
									 i = y*bitmap.bmWidthBytes;
									 for (x=0; x<bitmap.bmWidth; x++)
									 {
										 iResult =  pBits[i];
										 iResult += pMemBits[i];
										 pBits[i] = iResult/2;
										 i++;
									 }
								 }
							 }
							 if (LOWORD(wParam)==IDM_PNTOP_SUBBMPFILE  ||  LOWORD(wParam)==IDM_PNTOP_SUBBMPCLIP)
							 {
								 SaveForUndo(hBitmap, TEXT("Andere Bitmap subtrahieren"));
								 for (y=0; y<bitmap.bmHeight; y++)
								 {
									 i = y*bitmap.bmWidthBytes;
									 for (x=0; x<bitmap.bmWidth; x++)
									 {
										 iResult =  pBits[i];
										 iResult -= pMemBits[i];
										 if (iResult<0)  iResult = 0;
										 pBits[i] = iResult;
										 i++;
									 }
								 }
							 }
							 if (LOWORD(wParam)==IDM_PNTOP_SUBFROMBMPFILE  ||  LOWORD(wParam)==IDM_PNTOP_SUBFROMBMPCLIP)
							 {
								 SaveForUndo(hBitmap, TEXT("Andere Bitmap subtrahieren"));
								 for (y=0; y<bitmap.bmHeight; y++)
								 {
									 i = y*bitmap.bmWidthBytes;
									 for (x=0; x<bitmap.bmWidth; x++)
									 {
										 iResult  = pMemBits[i];
										 iResult -= pBits[i];
										 if (iResult<0)  iResult = 0;
										 pBits[i] = iResult;
										 i++;
									 }
								 }
							 }
							 if (LOWORD(wParam)==IDM_PNTOP_MULBMPFILE  ||  LOWORD(wParam)==IDM_PNTOP_MULBMPCLIP)
							 {
								 SaveForUndo(hBitmap, TEXT("Mit anderer Bitmap multiplizieren"));
								 for (y=0; y<bitmap.bmHeight; y++)
								 {
									 i = y*bitmap.bmWidthBytes;
									 for (x=0; x<bitmap.bmWidth; x++)
									 {
										 Factor = (double) pMemBits[i] / 255.0;
										 pBits[i] = BYTE(pBits[i]*Factor +0.3);
										 i++;
									 }
								 }
							 }
							 // Andere Bitmap wieder lschen und Anzeige aktualisieren
               DeleteObject (hOtherBitmap);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
               InvalidateRect (hwnd, NULL, FALSE);
						   return 0;

					case IDM_LOCOP_NORMALIZE:
						   if (!hBitmap) return 0;
							 LocOpSizeParam.hBitmap = hBitmap;
							 LocOpSizeParam.Operation = IDM_LOCOP_NORMALIZE;
							 if (!DialogBoxParam(hInstance, TEXT("LocOpSizeDlg"), hwnd, LocOpSizeDlgProc, (LPARAM) &LocOpSizeParam))  return 0;
							 SaveForUndo(hBitmap, TEXT("Lokal normalisieren"));
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_WAIT));
							 SetCursor(LoadCursor(NULL, IDC_WAIT));
							 CalcLocalNormalize(hBitmap, LocOpSizeParam.Width, LocOpSizeParam.Height, LocOpSizeParam.Param, LocOpSizeParam.Compare);
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_ARROW));
							 SetCursor(LoadCursor(NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
						   return 0;

					case IDM_LOCOP_RANKORDER:
						   if (hBitmap == NULL) return 0;
							 LocOpSizeParam.hBitmap = hBitmap;
							 LocOpSizeParam.Operation = IDM_LOCOP_RANKORDER;
							 if (!DialogBoxParam(hInstance, TEXT("LocOpSizeDlg"), hwnd, LocOpSizeDlgProc, (LPARAM) &LocOpSizeParam))  return 0;
							 SaveForUndo(hBitmap, TEXT("Rangordnungsfilter"));
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_WAIT));
							 SetCursor(LoadCursor(NULL, IDC_WAIT));
							 CalcLocalRankOrder(hBitmap, LocOpSizeParam.Width, LocOpSizeParam.Height, LocOpSizeParam.Param);
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_ARROW));
							 SetCursor(LoadCursor(NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
							 return 0;
					
					case IDM_LOCOP_GRADIENT:
						   if (!hBitmap)  return 0;
							 SaveForUndo(hBitmap, TEXT("Gradientenfilter"));
							 if (!CalcGradientFilter(hBitmap)) return 0;
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
						   return 0;

					case IDM_LOCOP_LINFILTER:
							 if (!hBitmap)  return 0;
						   // Standarddialog zum ffnen von Dateien
						   ofn.lpstrFilter = szLFFilter;
							 //LinFilter.szFilterFile[0]='\0';
							 ofn.lpstrFile = LinFilter.szFilterFile;
               ofn.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST;
               if (!GetOpenFileName(&ofn))
                    return 0;
               // DIB-Datei einlesen und DIB Section erzeugen
               SetCursor (LoadCursor (NULL, IDC_WAIT));
               ShowCursor (TRUE);
               if (!LoadLinFilter(hwnd, &LinFilter))
							 {
								 if (LinFilter.FilterMatrix)  free(LinFilter.FilterMatrix);
								 return 0;
							 }
							 SaveForUndo(hBitmap, LinFilter.szFilterName);
							 CalcLinFilter(hBitmap, LinFilter);
							 free(LinFilter.FilterMatrix);
               ShowCursor (FALSE);
               SetCursor (LoadCursor (NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
							 return 0;
					
					case IDM_LOCOP_RECT3X3:
							 if (!hBitmap)  return 0;
							 strcpy(LinFilter.szFilterName, TEXT("Rechteck-Filter (3x3)"));
							 LinFilter.FilterHeight = 3;
							 LinFilter.FilterWidth  = 3;
							 LinFilter.FilterMatrix = (float*) malloc(LinFilter.FilterHeight*LinFilter.FilterWidth*sizeof(float));
							 if (!LinFilter.FilterMatrix)  return 0;
               SetCursor (LoadCursor (NULL, IDC_WAIT));
               ShowCursor (TRUE);
							 for (i=0; i<LinFilter.FilterWidth*LinFilter.FilterHeight; i++)  LinFilter.FilterMatrix[i]=1;
							 LinFilter.szFilterFile[0] = '\0';
							 LinFilter.BorderParam = -2;
							 LinFilter.DivBy = (float) LinFilter.FilterWidth*LinFilter.FilterHeight;
							 LinFilter.AddOffset = 0;
							 LinFilter.NormalizeParam = 0;
							 LinFilter.SignParam = 0;
							 SaveForUndo(hBitmap, LinFilter.szFilterName);
							 CalcLinFilter(hBitmap, LinFilter);
							 free(LinFilter.FilterMatrix);
               ShowCursor (FALSE);
               SetCursor (LoadCursor (NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
						   return 0;

					case IDM_LOCOP_RECT5X5:
							 if (!hBitmap)  return 0;
							 strcpy(LinFilter.szFilterName, TEXT("Rechteck-Filter (5x5)"));
							 LinFilter.FilterHeight = 5;
							 LinFilter.FilterWidth  = 5;
							 LinFilter.FilterMatrix = (float*) malloc(LinFilter.FilterHeight*LinFilter.FilterWidth*sizeof(float));
							 if (!LinFilter.FilterMatrix)  return 0;
               SetCursor (LoadCursor (NULL, IDC_WAIT));
               ShowCursor (TRUE);
							 for (i=0; i<LinFilter.FilterWidth*LinFilter.FilterHeight; i++)  LinFilter.FilterMatrix[i]=1;
							 LinFilter.szFilterFile[0] = '\0';
							 LinFilter.BorderParam = -2;
							 LinFilter.DivBy = (float) LinFilter.FilterWidth*LinFilter.FilterHeight;
							 LinFilter.AddOffset = 0;
							 LinFilter.NormalizeParam = 0;
							 LinFilter.SignParam = 0;
							 SaveForUndo(hBitmap, LinFilter.szFilterName);
							 CalcLinFilter(hBitmap, LinFilter);
							 free(LinFilter.FilterMatrix);
               ShowCursor (FALSE);
               SetCursor (LoadCursor (NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
						   return 0;

					case IDM_LOCOP_GAUSS3X3:
							 if (!hBitmap)  return 0;
							 strcpy(LinFilter.szFilterName, TEXT("Gauss-Filter (3x3)"));
							 LinFilter.FilterHeight = 3;
							 LinFilter.FilterWidth  = 3;
							 LinFilter.FilterMatrix = (float*) malloc(LinFilter.FilterHeight*LinFilter.FilterWidth*sizeof(float));
							 if (!LinFilter.FilterMatrix)  return 0;
               SetCursor (LoadCursor (NULL, IDC_WAIT));
               ShowCursor (TRUE);
							 LinFilter.FilterMatrix[0]=1;
							 LinFilter.FilterMatrix[1]=2;
							 LinFilter.FilterMatrix[2]=1;
							 LinFilter.FilterMatrix[3]=2;
							 LinFilter.FilterMatrix[4]=4;
							 LinFilter.FilterMatrix[5]=2;
							 LinFilter.FilterMatrix[6]=1;
							 LinFilter.FilterMatrix[7]=2;
							 LinFilter.FilterMatrix[8]=1;
							 LinFilter.szFilterFile[0] = '\0';
							 LinFilter.BorderParam = -2;
							 LinFilter.DivBy = 0;
							 for (i=0; i<LinFilter.FilterWidth*LinFilter.FilterHeight; i++)  LinFilter.DivBy += LinFilter.FilterMatrix[i];
							 LinFilter.AddOffset = 0;
							 LinFilter.NormalizeParam = 0;
							 LinFilter.SignParam = 0;
							 SaveForUndo(hBitmap, LinFilter.szFilterName);
							 CalcLinFilter(hBitmap, LinFilter);
							 free(LinFilter.FilterMatrix);
               ShowCursor (FALSE);
               SetCursor (LoadCursor (NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
						   return 0;

					case IDM_LOCOP_GAUSS5X5:
							 if (!hBitmap)  return 0;
							 strcpy(LinFilter.szFilterName, TEXT("Gauss-Filter (5x5)"));
							 LinFilter.FilterHeight = 5;
							 LinFilter.FilterWidth  = 5;
							 LinFilter.FilterMatrix = (float*) malloc(LinFilter.FilterHeight*LinFilter.FilterWidth*sizeof(float));
							 if (!LinFilter.FilterMatrix)  return 0;
               SetCursor (LoadCursor (NULL, IDC_WAIT));
               ShowCursor (TRUE);
							 LinFilter.FilterMatrix[0]=1;
							 LinFilter.FilterMatrix[1]=4;
							 LinFilter.FilterMatrix[2]=6;
							 LinFilter.FilterMatrix[3]=4;
							 LinFilter.FilterMatrix[4]=1;
							 LinFilter.FilterMatrix[5]=4;
							 LinFilter.FilterMatrix[6]=16;
							 LinFilter.FilterMatrix[7]=24;
							 LinFilter.FilterMatrix[8]=26;
							 LinFilter.FilterMatrix[9]=4;
							 LinFilter.FilterMatrix[10]=6;
							 LinFilter.FilterMatrix[11]=24;
							 LinFilter.FilterMatrix[12]=36;
							 LinFilter.FilterMatrix[13]=24;
							 LinFilter.FilterMatrix[14]=6;
							 LinFilter.FilterMatrix[15]=4;
							 LinFilter.FilterMatrix[16]=16;
							 LinFilter.FilterMatrix[17]=24;
							 LinFilter.FilterMatrix[18]=16;
							 LinFilter.FilterMatrix[19]=4;
							 LinFilter.FilterMatrix[20]=1;
							 LinFilter.FilterMatrix[21]=4;
							 LinFilter.FilterMatrix[22]=6;
							 LinFilter.FilterMatrix[23]=4;
							 LinFilter.FilterMatrix[24]=1;
							 LinFilter.szFilterFile[0] = '\0';
							 LinFilter.BorderParam = -2;
							 LinFilter.DivBy = 0;
							 for (i=0; i<LinFilter.FilterWidth*LinFilter.FilterHeight; i++)  LinFilter.DivBy += LinFilter.FilterMatrix[i];
							 LinFilter.AddOffset = 0;
							 LinFilter.NormalizeParam = 0;
							 LinFilter.SignParam = 0;
							 SaveForUndo(hBitmap, LinFilter.szFilterName);
							 CalcLinFilter(hBitmap, LinFilter);
							 free(LinFilter.FilterMatrix);
               ShowCursor (FALSE);
               SetCursor (LoadCursor (NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
						   return 0;

					case IDM_LOCOP_UNSHARP3X3:
							 if (!hBitmap)  return 0;
							 strcpy(LinFilter.szFilterName, TEXT("Unscharfmaske (3x3)"));
							 LinFilter.FilterHeight = 3;
							 LinFilter.FilterWidth  = 3;
							 LinFilter.FilterMatrix = (float*) malloc(LinFilter.FilterHeight*LinFilter.FilterWidth*sizeof(float));
							 if (!LinFilter.FilterMatrix)  return 0;
               SetCursor (LoadCursor (NULL, IDC_WAIT));
               ShowCursor (TRUE);
							 for (i=0; i<LinFilter.FilterWidth*LinFilter.FilterHeight; i++)  LinFilter.FilterMatrix[i]=-1;
							 LinFilter.FilterMatrix[4] = 8;
							 LinFilter.szFilterFile[0] = '\0';
							 LinFilter.BorderParam = -2;
							 LinFilter.DivBy = 9;
							 LinFilter.AddOffset = 0;
							 LinFilter.NormalizeParam = 1;
							 LinFilter.SignParam = 2;
							 SaveForUndo(hBitmap, LinFilter.szFilterName);
							 CalcLinFilter(hBitmap, LinFilter);
							 free(LinFilter.FilterMatrix);
               ShowCursor (FALSE);
               SetCursor (LoadCursor (NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
						   return 0;

					case IDM_LOCOP_UNSHARP5X5:
							 if (!hBitmap)  return 0;
							 strcpy(LinFilter.szFilterName, TEXT("Unscharfmaske (5x5)"));
							 LinFilter.FilterHeight = 5;
							 LinFilter.FilterWidth  = 5;
							 LinFilter.FilterMatrix = (float*) malloc(LinFilter.FilterHeight*LinFilter.FilterWidth*sizeof(float));
							 if (!LinFilter.FilterMatrix)  return 0;
               SetCursor (LoadCursor (NULL, IDC_WAIT));
               ShowCursor (TRUE);
							 for (i=0; i<LinFilter.FilterWidth*LinFilter.FilterHeight; i++)  LinFilter.FilterMatrix[i]=-1;
							 LinFilter.FilterMatrix[12] = 24;
							 LinFilter.szFilterFile[0] = '\0';
							 LinFilter.BorderParam = -2;
							 LinFilter.DivBy = 25;
							 LinFilter.AddOffset = 0;
							 LinFilter.NormalizeParam = 1;
							 LinFilter.SignParam = 2;
							 SaveForUndo(hBitmap, LinFilter.szFilterName);
							 CalcLinFilter(hBitmap, LinFilter);
							 free(LinFilter.FilterMatrix);
               ShowCursor (FALSE);
               SetCursor (LoadCursor (NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
						   return 0;

					case IDM_LOCOP_DERIVHOR:
							 if (!hBitmap)  return 0;
							 strcpy(LinFilter.szFilterName, TEXT("Ableitungsfilter horizontal"));
							 LinFilter.FilterHeight = 3;
							 LinFilter.FilterWidth  = 3;
							 LinFilter.FilterMatrix = (float*) malloc(LinFilter.FilterHeight*LinFilter.FilterWidth*sizeof(float));
							 if (!LinFilter.FilterMatrix)  return 0;
               SetCursor (LoadCursor (NULL, IDC_WAIT));
               ShowCursor (TRUE);
							 for (i=0; i<LinFilter.FilterWidth*LinFilter.FilterHeight; i++)  LinFilter.FilterMatrix[i]=0;
							 LinFilter.FilterMatrix[3] =-1.0;
							 LinFilter.FilterMatrix[5] = 1.0;
							 LinFilter.szFilterFile[0] = '\0';
							 LinFilter.BorderParam = -1;
							 LinFilter.DivBy = 2;
							 LinFilter.AddOffset = 127;
							 LinFilter.NormalizeParam = 0;
							 LinFilter.SignParam = 0;
							 SaveForUndo(hBitmap, LinFilter.szFilterName);
							 CalcLinFilter(hBitmap, LinFilter);
							 free(LinFilter.FilterMatrix);
               ShowCursor (FALSE);
               SetCursor (LoadCursor (NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
						   return 0;

					case IDM_LOCOP_DERIVVER:
							 if (!hBitmap)  return 0;
							 strcpy(LinFilter.szFilterName, TEXT("Ableitungsfilter vertikal"));
							 LinFilter.FilterHeight = 3;
							 LinFilter.FilterWidth  = 3;
							 LinFilter.FilterMatrix = (float*) malloc(LinFilter.FilterHeight*LinFilter.FilterWidth*sizeof(float));
							 if (!LinFilter.FilterMatrix)  return 0;
               SetCursor (LoadCursor (NULL, IDC_WAIT));
               ShowCursor (TRUE);
							 for (i=0; i<LinFilter.FilterWidth*LinFilter.FilterHeight; i++)  LinFilter.FilterMatrix[i]=0;
							 LinFilter.FilterMatrix[1] =-1.0;
							 LinFilter.FilterMatrix[7] = 1.0;
							 LinFilter.szFilterFile[0] = '\0';
							 LinFilter.BorderParam = -1;
							 LinFilter.DivBy = 2;
							 LinFilter.AddOffset = 127;
							 LinFilter.NormalizeParam = 0;
							 LinFilter.SignParam = 0;
							 SaveForUndo(hBitmap, LinFilter.szFilterName);
							 CalcLinFilter(hBitmap, LinFilter);
							 free(LinFilter.FilterMatrix);
               ShowCursor (FALSE);
               SetCursor (LoadCursor (NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
						   return 0;

					case IDM_LOCOP_DERIVDIAG:
							 if (!hBitmap)  return 0;
							 strcpy(LinFilter.szFilterName, TEXT("Ableitungsfilter diagonal"));
							 LinFilter.FilterHeight = 3;
							 LinFilter.FilterWidth  = 3;
							 LinFilter.FilterMatrix = (float*) malloc(LinFilter.FilterHeight*LinFilter.FilterWidth*sizeof(float));
							 if (!LinFilter.FilterMatrix)  return 0;
               SetCursor (LoadCursor (NULL, IDC_WAIT));
               ShowCursor (TRUE);
							 for (i=0; i<LinFilter.FilterWidth*LinFilter.FilterHeight; i++)  LinFilter.FilterMatrix[i]=0;
							 LinFilter.FilterMatrix[0] =-1.0;
							 LinFilter.FilterMatrix[8] = 1.0;
							 LinFilter.szFilterFile[0] = '\0';
							 LinFilter.BorderParam = -1;
							 LinFilter.DivBy = 2;
							 LinFilter.AddOffset = 127;
							 LinFilter.NormalizeParam = 0;
							 LinFilter.SignParam = 0;
							 SaveForUndo(hBitmap, LinFilter.szFilterName);
							 CalcLinFilter(hBitmap, LinFilter);
							 free(LinFilter.FilterMatrix);
               ShowCursor (FALSE);
               SetCursor (LoadCursor (NULL, IDC_ARROW));
							 InvalidateRect(hwnd, NULL, FALSE);
							 if (!bChanged)
							 {
								 bChanged = TRUE;
								 SetCaption(hwnd, szFileName, bChanged);
							 }
						   return 0;

					case IDM_GLOBAL_SPECTRUM:
							 SpectrumParam.hInfoBox  = hInfoBox;
							 SpectrumParam.hInstance = hInstance;
							 SpectrumParam.hBitmap   = hBitmap;
					     hMenu = GetMenu(hwnd);
							 if (GetMenuState(hMenu,IDM_GLOBAL_AUTOSPECTSIZE, MF_BYCOMMAND) == MF_CHECKED)
							 {
								 // Mindestens 50% der Bitmap-Ausdehnung fr Rand (Padding) reservieren
							   SpectrumParam.Width  = bitmap.bmWidth + (bitmap.bmWidth>>1);
								 SpectrumParam.Height = bitmap.bmHeight + (bitmap.bmHeight>>1);
						     // Dan die Spektrum-Gre als nchsthhere Zweierpotenz berechnen
								 i=1;
								 while ((SpectrumParam.Width>>i) > 0)  i++;
						     SpectrumParam.Width = 1<<i;
						     i=1;
						     while ((SpectrumParam.Height>>i) > 0)  i++;
						     SpectrumParam.Height = 1<<i;
								 SpectrumParam.OffX = 0;
								 SpectrumParam.OffY = 0;
								 SpectrumParam.BorderParam = 0;
							 }
							 else
							 {
								 if (ManBitmapSize.NewWidth<bitmap.bmWidth  ||  ManBitmapSize.NewHeight<bitmap.bmHeight)
								 {
									 MessageBox(hwnd, TEXT("Die aktuelle Bitmap ist hher oder breiter als die ")
										                TEXT("entsprechende manuelle Angabe der Gre fr das Spektrum! ")
																		TEXT("Kontrollieren Sie bitte Ihre Angaben unter 'Gre ")
																		TEXT("manuell festlegen...'!")
										              , TEXT("Spektrum berechnen - Fehlerhafte Grenangabe"), MB_ICONINFORMATION | MB_OK);
									 return 0;
								 }
								 if ((!ManBitmapSize.OffXCenter && ManBitmapSize.NewWidth<bitmap.bmWidth+ManBitmapSize.OffX)
									   ||  (!ManBitmapSize.OffYCenter && ManBitmapSize.NewHeight<bitmap.bmHeight+ManBitmapSize.OffY))
								 {
									 MessageBox(hwnd, TEXT("Die in 'Gre manuell festlegen...' angegebenen Offsets ")
										                TEXT("sind zu gro. Bei dieser Einstellung wrde sich die aktuelle ")
																		TEXT("Bitmap zum Teil ausserhalb des darstellbaren Bereichs befinden ")
																		TEXT("und teilweise abgeschnitten! Kontrollieren Sie bitte ")
																		TEXT("Ihre Angaben unter 'Gre manuell festlegen...'!")
										              , TEXT("Spektrum berechnen - Fehlerhafte Offsetangabe"), MB_ICONINFORMATION | MB_OK);
									 return 0;
								 }
							   SpectrumParam.Width  = ManBitmapSize.NewWidth;
								 SpectrumParam.Height = ManBitmapSize.NewHeight;
								 SpectrumParam.OffX   = (ManBitmapSize.OffXCenter) ? (ManBitmapSize.NewWidth -bitmap.bmWidth )/2 : ManBitmapSize.OffX;
								 SpectrumParam.OffY   = (ManBitmapSize.OffYCenter) ? (ManBitmapSize.NewHeight-bitmap.bmHeight)/2 : ManBitmapSize.OffY;
								 SpectrumParam.BorderParam = ManBitmapSize.Border;
							 }
						   // Genug Speicher? Sonst abbrechen!
							 SpectrumParam.OldSpectRe = (float*)  malloc(SpectrumParam.Width*SpectrumParam.Height*sizeof(float));
							 if (!SpectrumParam.OldSpectRe)
							 {
								 MessageBox(hwnd, TEXT("Nicht gengend Speicher vorhanden!"), TEXT("Particle Tracer - Spektrum bearbeiten"), MB_OK | MB_ICONEXCLAMATION);
								 return 0;
							 }
							 SpectrumParam.OldSpectIm = (float*)  malloc(SpectrumParam.Width*SpectrumParam.Height*sizeof(float));
							 if (!SpectrumParam.OldSpectIm)
							 {
								 free (SpectrumParam.OldSpectRe);
								 MessageBox(hwnd, TEXT("Nicht gengend Speicher vorhanden!"), TEXT("Particle Tracer - Spektrum bearbeiten"), MB_OK | MB_ICONEXCLAMATION);
								 return 0;
							 }
							 SpectrumParam.FilterRe = (float*)  malloc(SpectrumParam.Width*SpectrumParam.Height*sizeof(float));
							 if (!SpectrumParam.FilterRe)
							 {
								 free (SpectrumParam.OldSpectRe);
								 free (SpectrumParam.OldSpectIm);
								 MessageBox(hwnd, TEXT("Nicht gengend Speicher vorhanden!"), TEXT("Particle Tracer - Spektrum bearbeiten"), MB_OK | MB_ICONEXCLAMATION);
								 return 0;
							 }
							 SpectrumParam.FilterIm = (float*)  malloc(SpectrumParam.Width*SpectrumParam.Height*sizeof(float));
							 if (!SpectrumParam.FilterIm)
							 {
								 free (SpectrumParam.OldSpectRe);
								 free (SpectrumParam.OldSpectIm);
								 free (SpectrumParam.FilterRe);
								 MessageBox(hwnd, TEXT("Nicht gengend Speicher vorhanden!"), TEXT("Particle Tracer - Spektrum bearbeiten"), MB_OK | MB_ICONEXCLAMATION);
								 return 0;
							 }
							 SpectrumParam.SpectRe = (float*)  malloc(SpectrumParam.Width*SpectrumParam.Height*sizeof(float));
							 if (!SpectrumParam.SpectRe)
							 {
								 free (SpectrumParam.OldSpectRe);
								 free (SpectrumParam.OldSpectIm);
								 free (SpectrumParam.FilterRe);
								 free (SpectrumParam.FilterIm);
								 MessageBox(hwnd, TEXT("Nicht gengend Speicher vorhanden!"), TEXT("Particle Tracer - Spektrum bearbeiten"), MB_OK | MB_ICONEXCLAMATION);
								 return 0;
							 }
							 SpectrumParam.SpectIm = (float*)  malloc(SpectrumParam.Width*SpectrumParam.Height*sizeof(float));
							 if (!SpectrumParam.SpectIm)
							 {
								 free (SpectrumParam.OldSpectRe);
								 free (SpectrumParam.OldSpectIm);
								 free (SpectrumParam.FilterRe);
								 free (SpectrumParam.FilterIm);
								 free (SpectrumParam.SpectRe);
								 MessageBox(hwnd, TEXT("Nicht gengend Speicher vorhanden!"), TEXT("Particle Tracer - Spektrum bearbeiten"), MB_OK | MB_ICONEXCLAMATION);
								 return 0;
							 }
							 SpectrumParam.Floats = (float*)  malloc(SpectrumParam.Width*SpectrumParam.Height*sizeof(float));
							 if (!SpectrumParam.Floats)
							 {
								 free (SpectrumParam.OldSpectRe);
								 free (SpectrumParam.OldSpectIm);
								 free (SpectrumParam.FilterRe);
								 free (SpectrumParam.FilterIm);
								 free (SpectrumParam.SpectRe);
								 free (SpectrumParam.SpectIm);
								 MessageBox(hwnd, TEXT("Nicht gengend Speicher vorhanden!"), TEXT("Particle Tracer - Spektrum bearbeiten"), MB_OK | MB_ICONEXCLAMATION);
								 return 0;
							 }
							 //  Neue Graustufen-DIB erzeugen, um Spektrum darzustellen (andere Gre als aktuelle Datei)
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_WAIT) );
							 SetCursor (LoadCursor (NULL, IDC_WAIT)) ;
							 SpectrumParam.hSpectBMP = hBitmap;
							 if (!ChangeBitmapSize(SpectrumParam.hSpectBMP, FALSE,
								                     SpectrumParam.Width, SpectrumParam.Height,
												      			 SpectrumParam.OffX, SpectrumParam.OffY,
																		 SpectrumParam.BorderParam)  )
							 {
								 free (SpectrumParam.OldSpectRe);
								 free (SpectrumParam.OldSpectIm);
								 free (SpectrumParam.FilterRe);
								 free (SpectrumParam.FilterIm);
								 free (SpectrumParam.SpectRe);
								 free (SpectrumParam.SpectIm);
								 free (SpectrumParam.Floats);
								 MessageBox(hwnd, TEXT("Nicht gengend Speicher vorhanden!"), TEXT("Particle Tracer - Spektrum bearbeiten"), MB_OK | MB_ICONEXCLAMATION);
								 return 0;
							 }
							 // Bitmap kopieren
							 GetObject(SpectrumParam.hSpectBMP, sizeof(BITMAP), &OtherBitmap);
							 {
								 BYTE* pBits;
								 pBits = (BYTE*) OtherBitmap.bmBits;
								 for (y=0; y<SpectrumParam.Height; y++)
								 {
									 for (x=0; x<SpectrumParam.Width; x++)
									 {
										 SpectrumParam.OldSpectRe[y*SpectrumParam.Width+x]=pBits[y*OtherBitmap.bmWidthBytes+x];
										 SpectrumParam.OldSpectIm[y*SpectrumParam.Width+x]=0;
									 }
								 }
							 }
							 // *********** Vorbereitungen fertig **********

							 // Berechne Spektrum
							 FFT2D(SpectrumParam.OldSpectRe, SpectrumParam.OldSpectIm, SpectrumParam.Width, SpectrumParam.Height, 1);
							 // Dialog: Spektrum bearbeiten
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_ARROW) );
							 SetCursor (LoadCursor (NULL, IDC_ARROW)) ;
							 if (!DialogBoxParam(hInstance, TEXT("FilterSpectDlg"), hwnd, FilterSpectDlgProc, (LPARAM) &SpectrumParam))
							 {
								 DeleteObject(SpectrumParam.hSpectBMP);  // Groe Bitmap (Spectrum) wieder freigeben
								 free (SpectrumParam.OldSpectRe);
								 free (SpectrumParam.OldSpectIm);
								 free (SpectrumParam.FilterRe);
								 free (SpectrumParam.FilterIm);
								 free (SpectrumParam.SpectRe);
								 free (SpectrumParam.SpectIm);
								 return 0;
							 }
							 // Rcktrafo
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_WAIT) );
							 SetCursor (LoadCursor (NULL, IDC_WAIT)) ;
							 FFT2D(SpectrumParam.SpectRe, SpectrumParam.SpectIm, SpectrumParam.Width, SpectrumParam.Height, -1);
							 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_ARROW) );
							 SetCursor (LoadCursor (NULL, IDC_ARROW)) ;
							 // Dialog:  Bearbeitetes, zurcktransformiertes Feld  ->  Bitmap
							 if (DialogBoxParam(hInstance, TEXT("InverseFFTDlg"), hwnd, InverseFFTDlgProc, (LPARAM) &SpectrumParam))
							 {
								 if (!bChanged)
								 {
									 bChanged = TRUE;
									 SetCaption(hwnd, szFileName, bChanged);
								 }
							 }
							 // Speicher wieder freigeben:
							 DeleteObject(SpectrumParam.hSpectBMP);
							 free (SpectrumParam.OldSpectRe);
							 free (SpectrumParam.OldSpectIm);
							 free (SpectrumParam.FilterRe);
							 free (SpectrumParam.FilterIm);
							 free (SpectrumParam.SpectRe);
							 free (SpectrumParam.SpectIm);
							 return 0;

 					case IDM_GLOBAL_AUTOSPECTSIZE:
						   hMenu = GetMenu(hwnd);
						   CheckMenuItem(hMenu, IDM_GLOBAL_AUTOSPECTSIZE,  MF_CHECKED );
						   CheckMenuItem(hMenu, IDM_GLOBAL_MANSPECTSIZE , MF_UNCHECKED);
							 if (!hBitmap)  return 0;
							 if (MessageBox(hwnd, TEXT("Die Gre eines berechneten Spektrums wird nun je nach Gre der jeweils ")
								                    TEXT("aktuellen Bitmap automatisch festgelegt. Dabei wird fr die Breite bzw. Hhe ")
											    					TEXT("des Spektrums die anderthalbfache Breite bzw. Hhe der aktuellen Bitmap auf ")
																		TEXT("die nchst grere Zweierpotenz erweitert und der entstehende Rand mit ")
																		TEXT("Nullen aufgefllt.\n\n")
													    			TEXT("Wollen Sie nun das Spektrum fr die aktuelle Bitmap berechnen?")
								                  , TEXT("Particle Tracer - Spektrum berechnen"), MB_YESNO | MB_ICONQUESTION) == IDYES)
								 SendMessage(hwnd, WM_COMMAND, IDM_GLOBAL_SPECTRUM, 0);
						   return 0;

 					case IDM_GLOBAL_MANSPECTSIZE:
							 if (!DialogBoxParam(hInstance, TEXT("BitmapSizeDlg"), hwnd, BitmapSizeDlgProc, (LPARAM) &ManBitmapSize))  return 0;
							 hMenu = GetMenu(hwnd);
							 CheckMenuItem(hMenu, IDM_GLOBAL_AUTOSPECTSIZE, MF_UNCHECKED);
							 CheckMenuItem(hMenu, IDM_GLOBAL_MANSPECTSIZE ,  MF_CHECKED );
							 if (!hBitmap)  return 0;
							 if (bitmap.bmWidth  > ManBitmapSize.NewWidth )  return 0;
							 if (bitmap.bmHeight > ManBitmapSize.NewHeight)  return 0;
							 if (!ManBitmapSize.OffXCenter  &&  bitmap.bmWidth  + ManBitmapSize.OffX > ManBitmapSize.NewWidth )  return 0;
							 if (!ManBitmapSize.OffYCenter  &&  bitmap.bmHeight + ManBitmapSize.OffY > ManBitmapSize.NewHeight)  return 0;
							 if (MessageBox(hwnd, TEXT("Die Parameter fr die Berechnung des Spektrums wurden gendert und gespeichert. ")
								                    TEXT("Sie werden so lange verwendet, bis Sie diese erneut ndern oder den Meneintrag ")
											    					TEXT("'Gre automatisch festlegen' auswhlen.\n\n")
													    			TEXT("Wollen Sie nun das Spektrum fr die aktuelle Bitmap berechnen?")
								                  , TEXT("Particle Tracer - Spektrum berechnen"), MB_YESNO | MB_ICONQUESTION) == IDYES)
								 SendMessage(hwnd, WM_COMMAND, IDM_GLOBAL_SPECTRUM, 0);
						   return 0;

					case IDM_NUMBERFORMAT_GERMAN:
						   hMenu = GetMenu(hwnd);
							 if (GetMenuState(hMenu, IDM_NUMBERFORMAT_GERMAN, MF_BYCOMMAND) & MF_CHECKED)
							   CheckMenuItem(hMenu, IDM_NUMBERFORMAT_GERMAN, MF_UNCHECKED);
							 else
							   CheckMenuItem(hMenu, IDM_NUMBERFORMAT_GERMAN, MF_CHECKED);
							 return 0;

					case IDM_PCOUNT_BINARISATION:
					case IDM_PCOUNT_MAXIMA:
					case IDM_PCOUNT_STREAKS:
						   if (!hBitmap)  return 0;
							 {
								 RGBQUAD OldRGBPal[256];

								 GetRGBPalette(hBitmap, OldRGBPal);
							   // InputBox: Grauwert-Cutoff usw. eingeben...
								 ParticleParam.hBitmap = hBitmap;
								 ParticleParam.OldRGBPal = OldRGBPal;
								 ParticleParam.Algorithm = LOWORD (wParam);
								 if (!DialogBoxParam(hInstance, TEXT("ParticleParamsDlg"), hwnd, ParticleParamsDlgProc, (LPARAM) &ParticleParam))
								 {
									 return 0;
								 }
								 UpdateWindow(hwnd);

								 // Speicher reservieren
							   Areas = (long*) malloc(bitmap.bmWidthBytes*bitmap.bmHeight*sizeof(long));
								 if (!Areas)
								 {
									 MessageBox(hwnd, TEXT("Nicht gengend Speicher vorhanden!"), TEXT("Particle Tracer - Maxima zhlen"), MB_OK | MB_ICONEXCLAMATION);
									 return 0;
								 }
								 pMemBits = (BYTE*) malloc(bitmap.bmWidthBytes*bitmap.bmHeight);
								 if (!pMemBits)
								 {
									 free(Areas);
									 MessageBox(hwnd, TEXT("Nicht gengend Speicher vorhanden!"), TEXT("Particle Tracer - Maxima zhlen"), MB_OK | MB_ICONEXCLAMATION);
									 return 0;
								 }
								 CopyMemory(pMemBits, pBits, bitmap.bmWidthBytes*bitmap.bmHeight);

								 // "Areas" vorbereiten  (Maxima oder Binarisation...)
								 if (LOWORD(wParam)==IDM_PCOUNT_MAXIMA)
									 SearchMaxima(hBitmap, (BYTE) ParticleParam.MinGrayVal, Areas);
								 else
									 AreaBinarisation(hBitmap, (BYTE) ParticleParam.MinGrayVal, Areas);
								 // Separation in "Areas"
								 i = AreaSeparation(hBitmap, Areas);
								 Particles = (PPARTICLEDATA) malloc(i*sizeof(PARTICLEDATA));
								 if (!Particles)
								 {
									 free(pMemBits);
									 free(Areas);
									 MessageBox(hwnd, TEXT("Nicht gengend Speicher vorhanden!"), TEXT("Particle Tracer - Maxima zhlen"), MB_OK | MB_ICONEXCLAMATION);
									 return 0;
								 }
								 // Flchen schrumpfen auf reprsentativen Punkt (Schwerpunkt) und darstellen
								 // AreaErosion(hBitmap, Areas, Particles, i);
								 AreaCenter(hBitmap, Areas, Particles, i);
								 CheckAreaSize(Particles, i, ParticleParam.MinArea, ParticleParam.MaxArea);
								 if (OldRGBPal[255].rgbRed>200  &&  OldRGBPal[255].rgbGreen<20  &&  OldRGBPal[255].rgbBlue<20)
									 SetSingleRGBPaletteEntry(hBitmap, 255, RGB(0,255,255));							 
								 else
									 SetSingleRGBPaletteEntry(hBitmap, 255, RGB(255,0,0));							 
								 ShowParticleCoord(hBitmap, Particles, i);
								 InvalidateRect(hwnd, NULL, FALSE);
								 UpdateWindow(hwnd);

								 // Daten exportieren ?
								 wsprintf(szTemp, TEXT("Insgesamt gezhlte Maxima: %li\n\nWollen Sie die Daten exportieren?"), i);
								 if (bTemp = (MessageBox(hwnd, szTemp, TEXT("Particle Tracer - Maxima zhlen"), MB_YESNO | MB_ICONQUESTION)==IDYES))
								 {
									 ofn.lpstrFilter = szFilter;
									 strcpy(szOtherFileName, szFileName);
									 x = strlen(szOtherFileName)-1;
									 while (x>0 && szOtherFileName[x]!='.') x--;
									 if (x>0)  szOtherFileName[x]='\0';
									 strcat(szOtherFileName, TEXT("_c"));
								   ofn.lpstrFile = szOtherFileName;
									 ofn.Flags = OFN_PATHMUSTEXIST;
									 do
									 {
										 bTemp = FALSE;
										 if (GetSaveFileName (&ofn))
										 {
											 PARTICLEDATAINFO ParticleDataInfo;
											 TCHAR OverwritePrompt[3*MAX_PATH + 1024];
											 strcpy(OverwritePrompt, TEXT("Achtung: Es werden mehrere Dateien (mit jeweils verschiedenen Dateiendungen) exportiert.\nAllerdings existieren die folgenden Dateien bereits:\n\n"));
											 x = strlen(szOtherFileName)-1;
											 while (x>0 && szOtherFileName[x]!='.') x--;
											 if (x>0)  szOtherFileName[x]='\0';
											 strcpy(ParticleDataInfo.szFileName, szOtherFileName);
											 strcat(ParticleDataInfo.szFileName, TEXT(".bmp"));
											 if (FileExists(ParticleDataInfo.szFileName))
											 {
												 strcat(OverwritePrompt, TEXT("     "));
												 strcat(OverwritePrompt, ParticleDataInfo.szFileName);
												 strcat(OverwritePrompt, TEXT("\n"));
												 bTemp = TRUE;
											 }
											 strcpy(ParticleDataInfo.szFileName, szOtherFileName);
											 strcat(ParticleDataInfo.szFileName, TEXT(".txt"));
											 if (FileExists(ParticleDataInfo.szFileName))
											 {
												 strcat(OverwritePrompt, TEXT("     "));
												 strcat(OverwritePrompt, ParticleDataInfo.szFileName);
												 strcat(OverwritePrompt, TEXT("\n"));
												 bTemp = TRUE;
											 }
											 strcpy(ParticleDataInfo.szFileName, szOtherFileName);
											 strcat(ParticleDataInfo.szFileName, TEXT(".dat"));
											 if (FileExists(ParticleDataInfo.szFileName))
											 {
												 strcat(OverwritePrompt, TEXT("     "));
												 strcat(OverwritePrompt, ParticleDataInfo.szFileName);
												 strcat(OverwritePrompt, TEXT("\n"));
												 bTemp = TRUE;
											 }
											 strcpy(ParticleDataInfo.szFileName, szOtherFileName);
											 if (bTemp)
											 {
												 strcat(OverwritePrompt, TEXT("\nSollen diese Dateien wirklich berschrieben werden?"));
												 if (MessageBox(hwnd, OverwritePrompt, TEXT("Daten exportieren"), MB_YESNO | MB_ICONEXCLAMATION) == IDYES)  bTemp = FALSE;
											 }
											 if (!bTemp)
											 {
												 hMenu = GetMenu(hwnd);
												 ParticleDataInfo.ExportGermanNumberFormat = GetMenuState(hMenu, IDM_NUMBERFORMAT_GERMAN, MF_BYCOMMAND) & MF_CHECKED;
												 ParticleDataInfo.hBitmap = hBitmap;
												 ParticleDataInfo.Algorithm = LOWORD(wParam);
												 ParticleDataInfo.GrayCutoff = (BYTE) ParticleParam.MinGrayVal;
												 ParticleDataInfo.MinArea = ParticleParam.MinArea;
												 ParticleDataInfo.MaxArea = ParticleParam.MaxArea;
												 ParticleDataInfo.Particles = Particles;
												 ParticleDataInfo.NumParticles = i;
												 j = ExportParticleData(ParticleDataInfo);
												 if (j)
												 {
													 if (j&1)  MessageBox(hwnd, TEXT("Fehler beim Speichern der '.bmp'-Datei!"), TEXT("Daten exportieren"), MB_OK | MB_ICONEXCLAMATION);
													 if (j&2)  MessageBox(hwnd, TEXT("Fehler beim Speichern der '.dat'-Datei!"), TEXT("Daten exportieren"), MB_OK | MB_ICONEXCLAMATION);
													 if (j&4)  MessageBox(hwnd, TEXT("Fehler beim Speichern der '.txt'-Datei!"), TEXT("Daten exportieren"), MB_OK | MB_ICONEXCLAMATION);
													 if (MessageBox(hwnd, TEXT("Wollen Sie nochmal versuchen, die Dateien unter einem anderen Namen oder Laufwerk abzuspeichern?"), TEXT("Daten exportieren"), MB_YESNO | MB_ICONQUESTION) != IDNO)
														 bTemp = TRUE;
												 }
												 // Wenn nicht geklappt, mu bTemp hier wieder TRUE sein!!!!
											 }
										 }
									 } while (bTemp);
								 }
	
//							 SaveBitsForUndo(pMemBits, bitmap.bmWidth, bitmap.bmWidthBytes, bitmap.bmHeight, TEXT("Maxima suchen"));
								 CopyMemory(pBits, pMemBits, bitmap.bmWidthBytes*bitmap.bmHeight);
								 SetRGBPalette(hBitmap, OldRGBPal);
								 InvalidateRect(hwnd, NULL, FALSE);
								 free(Areas);
								 free(Particles);
								 free(pMemBits);
								}
						   return 0;
					
					case IDM_HELP_ABOUTBOX:
						   DialogBox(hInstance, TEXT("AboutBox"), hwnd, AboutBoxProc);
							 return 0;
					
					case IDM_EXIT:
						   SendMessage (hwnd, WM_CLOSE, 0, 0);
							 return 0;
					}
          return 0;

		case WM_KEYDOWN:
			   if (!hBitmap)  return 0;
				 if (bScroll)  return 0;
			   hMenu = GetMenu(hwnd);
				 if (wParam == VK_ESCAPE  &&  GetMenuState(hMenu,IDM_TOOLS_SELECT, MF_BYCOMMAND) == MF_CHECKED  &&  (bSelecting  ||  bSelMoving))
				 {
					 ReleaseCapture();
					 bSelecting = FALSE;
					 bSelMoving = FALSE;
					 SelRect = OldSelRect;
					 InvalidateRect(hwnd, NULL, FALSE);
				 }
			   if (wParam == VK_CONTROL  &&  GetMenuState(hMenu,IDM_TOOLS_MOVE, MF_BYCOMMAND) == MF_CHECKED)
				 {
					 bZoom = TRUE;
					 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(hInstance, TEXT("zoomglass")) );
					 SetCursor(LoadCursor(hInstance, TEXT("zoomglass")));
				 }
			   return 0;

		case WM_KEYUP:
			   if (wParam == VK_CONTROL  &&  bZoom)
				 {
					 bZoom = FALSE;
					 if (hBitmap)
					 {
						 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(hInstance, TEXT("crosscursor")) );
					   SetCursor(LoadCursor(hInstance, TEXT("crosscursor")));
					 }
					 else
					 {
						 SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_ARROW) );
					   SetCursor(LoadCursor(NULL, IDC_ARROW));
					 }
				 }
			   return 0;

		case WM_MOUSEWHEEL:
          iResult = 0;
			    if (hBitmap)
					{
				    if ((short) HIWORD(wParam)>0  &&  ShowSizeFactor<20)
						{
							ShowSizeFactor += 1.0f;
							iResult = +1;
						}
						if ((short) HIWORD(wParam)<0  &&  ShowSizeFactor>-10)
						{
							ShowSizeFactor -= 1.0f;
							iResult = -1;
						}
            if (iResult == 0)  break;

						Factor = ((double) (cxClient/2-xShow)) / ((double) ShowWidth);
						ShowWidth = ShowSizeFactor<0 ? long(-bitmap.bmWidth/(ShowSizeFactor-1)) : long(bitmap.bmWidth*(ShowSizeFactor+1));
						xShow = int(cxClient/2 - Factor*ShowWidth);
						Factor = ((double) (cyClient/2-yShow)) / ((double) ShowHeight);
						ShowHeight = ShowSizeFactor<0 ? long(-bitmap.bmHeight/(ShowSizeFactor-1)) : long(bitmap.bmHeight*(ShowSizeFactor+1));
						yShow = int(cyClient/2 - Factor*ShowHeight);
						
						CheckBorder(xShow, yShow, cxClient, cyClient, ShowWidth, ShowHeight);
						InvalidateRect(hwnd, NULL, FALSE);
					}
					return 0;

     case WM_LBUTTONDOWN:
					xMousePos = LOWORD(lParam);
					yMousePos = HIWORD(lParam);
					if (!hBitmap)  return 0;
			    hMenu = GetMenu(hwnd);
					if (GetMenuState(hMenu, IDM_TOOLS_SETORIGIN, MF_BYCOMMAND) == MF_CHECKED)
					{
						Factor = (double) ShowWidth / bitmap.bmWidth;
						x = long((xMousePos-xShow) / Factor);
						Factor = (double) ShowHeight / bitmap.bmHeight;
						y = long((yShow+ShowHeight-1-yMousePos) / Factor);
						if (x<0  ||  x>=bitmap.bmWidth  ||  y<0  ||  y>=bitmap.bmHeight)  return 0;
						SaveForUndo(hBitmap, TEXT("Ursprung verschieben"));
						SetBMPOrigin(hBitmap, x, y);
						InvalidateRect(hwnd, NULL, FALSE);
					}
					if (GetMenuState(hMenu, IDM_TOOLS_SELECT, MF_BYCOMMAND) == MF_CHECKED)
					{
						Factor = (double) ShowWidth / bitmap.bmWidth;
						x = long((xMousePos-xShow) / Factor);
						Factor = (double) ShowHeight / bitmap.bmHeight;
						y = long((yShow+ShowHeight-1-yMousePos) / Factor);
						if (x<0)  x=0;
						if (x>=bitmap.bmWidth)  x=bitmap.bmWidth-1;
						if (y<0)  y=0;
						if (y>=bitmap.bmHeight) y=bitmap.bmHeight-1;
						OldSelRect = SelRect;
						XORSelRect.left = XORSelRect.bottom = 0;
						XORSelRect.right = bitmap.bmWidth-1;
						XORSelRect.top = bitmap.bmHeight-1;
						bTemp = SelRect.left==0  &&  SelRect.bottom==0  &&  SelRect.right==bitmap.bmWidth-1  &&  SelRect.top==bitmap.bmHeight-1;
						bTemp = bTemp || x<SelRect.left  ||  x>SelRect.right  ||  y<SelRect.bottom  ||  y>SelRect.top;
						if (bTemp)
						{
							SelRect.left = SelRect.right = x;
							SelRect.top = SelRect.bottom = y;
							bSelecting = TRUE;
							InvalidateRect(hwnd, NULL, FALSE);
						}
						else
						{
							bSelMoving = TRUE;
							wsprintf(szTemp, "X:  %li (%li)", SelRect.left, abs(SelRect.right-SelRect.left)+1);
							SetDlgItemText(hInfoBox, IDC_X_POS, szTemp);
							wsprintf(szTemp, "Y:  %li (%li)", SelRect.bottom, abs(SelRect.top-SelRect.bottom)+1);
							SetDlgItemText(hInfoBox, IDC_Y_POS, szTemp);
							x = pBits[y*bitmap.bmWidthBytes + x];
							wsprintf(szTemp, "  (Markierung)");
							SetDlgItemText(hInfoBox, IDC_GRAYVALUE, szTemp);
						}
						SetCapture(hwnd);
					}
					if (GetMenuState(hMenu, IDM_TOOLS_DRAW, MF_BYCOMMAND) == MF_CHECKED  &&  !bDrawLB  &&  !bDrawRB)
					{
						Factor = (double) ShowWidth / bitmap.bmWidth;
						x = long((xMousePos-xShow) / Factor);
						Factor = (double) ShowHeight / bitmap.bmHeight;
						y = long((yShow+ShowHeight-1-yMousePos) / Factor);
						if (x>=0  &&  x<bitmap.bmWidth  &&  y>=0  &&  y<bitmap.bmHeight)
						{
							SaveForUndo(hBitmap, TEXT("Zeichnen"));
							pMemBits = (BYTE*) malloc(bitmap.bmWidthBytes*bitmap.bmHeight);
							bDrawingLine = FALSE;
							if (!bChanged)
							{
							  bChanged = TRUE;
							  SetCaption(hwnd, szFileName, bChanged);
							}
							XORSelRect.left = XORSelRect.right = x;
							XORSelRect.bottom = XORSelRect.top = y;
							DrawLine(hBitmap, SelRect, XORSelRect, PenWidth, PenColorParam.LButtonGrayVal, pMemBits);
							InvalidateRect(hwnd, NULL, FALSE);
							bDrawLB = TRUE;
							SetCapture(hwnd);
						}
					}

					if (GetMenuState(hMenu, IDM_TOOLS_MOVE, MF_BYCOMMAND) == MF_CHECKED)
					{
						if ((wParam & MK_CONTROL)  &&  ShowSizeFactor<20  &&  bZoom)
						{
					    ShowSizeFactor += 1.0f;
							Factor = ((double) (xMousePos-xShow)) / ((double) ShowWidth);
							ShowWidth = ShowSizeFactor<0 ? long(-bitmap.bmWidth/(ShowSizeFactor-1)) : long(bitmap.bmWidth*(ShowSizeFactor+1));
							xShow = int(cxClient/2 - Factor*ShowWidth);
							Factor = ((double) (yMousePos-yShow)) / ((double) ShowHeight);
							ShowHeight = ShowSizeFactor<0 ? long(-bitmap.bmHeight/(ShowSizeFactor-1)) : long(bitmap.bmHeight*(ShowSizeFactor+1));
							yShow = int(cyClient/2 - Factor*ShowHeight);
							CheckBorder(xShow, yShow, cxClient, cyClient, ShowWidth, ShowHeight);
							InvalidateRect(hwnd, NULL, FALSE);
						}
						if (!(wParam & MK_CONTROL))
						{
							SetCapture(hwnd);
							SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(hInstance, TEXT("closedhand")) );
							SetCursor(LoadCursor(hInstance, TEXT("closedhand")) );
							bScroll = TRUE;
						}
					}
			    return 0;

		 case WM_LBUTTONUP:
					xMousePos = LOWORD(lParam);
					yMousePos = HIWORD(lParam);
					if (!hBitmap)  return 0;
			    hMenu = GetMenu(hwnd);
					if (GetMenuState(hMenu,IDM_TOOLS_SELECT, MF_BYCOMMAND) == MF_CHECKED)
					{
						if (bSelecting)
						{
							bSelecting = FALSE;
						  ReleaseCapture();
							Factor = (double) ShowWidth / bitmap.bmWidth;
							x = long((xMousePos-xShow) / Factor);
							Factor = (double) ShowHeight / bitmap.bmHeight;
							y = long((yShow+ShowHeight-1-yMousePos) / Factor);
							if (x<0)  x=0;
							if (x>=bitmap.bmWidth)  x=bitmap.bmWidth-1;
							if (y<0)  y=0;
							if (y>=bitmap.bmHeight) y=bitmap.bmHeight-1;
							SelRect.right = x;
							SelRect.bottom = y;
					    if (SelRect.right<SelRect.left)
							{
								x = SelRect.right;
								SelRect.right = SelRect.left;
								SelRect.left = x;
							}
					    if (SelRect.top<SelRect.bottom)
							{
								x = SelRect.bottom;
								SelRect.bottom = SelRect.top;
								SelRect.top = x;
							}
							bDelOldSelRect = TRUE;
							InvalidateRect(hwnd, NULL, FALSE);
						}
						if (bSelMoving)
						{
							bSelMoving = FALSE;
							ReleaseCapture();
							InvalidateRect(hwnd, NULL, FALSE);
						}
					}
					if (bDrawLB)
					{
						ReleaseCapture();
						bDrawLB = FALSE;
						// Original wiederherstellen?
						//DrawLine(hBitmap, XORSelRect, PenWidth, -1);
						Factor = (double) ShowWidth / bitmap.bmWidth;
						XORSelRect.right = long(((signed short) xMousePos-xShow) / Factor);
						Factor = (double) ShowHeight / bitmap.bmHeight;
						XORSelRect.top = long((yShow+ShowHeight-1-(signed short) yMousePos) / Factor);
						DrawLine(hBitmap, SelRect, XORSelRect, PenWidth, PenColorParam.LButtonGrayVal, pMemBits);
						InvalidateRect(hwnd, NULL, FALSE);
						if (pMemBits)  free(pMemBits);
					}
					if (GetMenuState(hMenu,IDM_TOOLS_MOVE, MF_BYCOMMAND) == MF_CHECKED  &&  bScroll)
					{
					  ReleaseCapture();
						bScroll = FALSE;
						SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(hInstance, TEXT("openhand")) );
						SetCursor(LoadCursor(hInstance, TEXT("openhand")) );
					}
					break;			    

		 case WM_RBUTTONDOWN:
					if (!hBitmap)  return 0;
	  	    hMenu = GetMenu(hwnd);
					if (GetMenuState(hMenu, IDM_TOOLS_SETORIGIN, MF_BYCOMMAND) == MF_CHECKED)
					{
						Factor = (double) ShowWidth / bitmap.bmWidth;
						x = long((xMousePos-xShow) / Factor);
						Factor = (double) ShowHeight / bitmap.bmHeight;
						y = long((yShow+ShowHeight-1-yMousePos) / Factor);
						if (x<0  ||  x>=bitmap.bmWidth  ||  y<0  ||  y>=bitmap.bmHeight)  return 0;
						SaveForUndo(hBitmap, TEXT("Ursprung verschieben"));
						SetBMPOrigin(hBitmap, bitmap.bmWidth-x, bitmap.bmHeight-y);
						InvalidateRect(hwnd, NULL, FALSE);
					}
					if (GetMenuState(hMenu,IDM_TOOLS_SELECT, MF_BYCOMMAND) == MF_CHECKED  &&  !bSelecting)
					{
						if (!(wParam & MK_CONTROL))
						{
							SelRect.bottom = SelRect.left = 0;
							SelRect.top = bitmap.bmHeight-1;
							SelRect.right = bitmap.bmWidth-1;
							InvalidateRect(hwnd, NULL, FALSE);
						}
						else
						{
							Factor = (double) ShowWidth / bitmap.bmWidth;
							x = long((LOWORD(lParam)-xShow) / Factor);
							Factor = (double) ShowHeight / bitmap.bmHeight;
							y = long((yShow+ShowHeight-1-HIWORD(lParam)) / Factor);
							if (x<0)  x=0;
							if (x>=bitmap.bmWidth)  x=bitmap.bmWidth-1;
							if (y<0)  y=0;
							if (y>=bitmap.bmHeight) y=bitmap.bmHeight-1;
							SelRect.right = x;
							SelRect.bottom = y;
					    if (SelRect.right<SelRect.left)
							{
								x = SelRect.right;
								SelRect.right = SelRect.left;
								SelRect.left = x;
							}
					    if (SelRect.top<SelRect.bottom)
							{
								x = SelRect.bottom;
								SelRect.bottom = SelRect.top;
								SelRect.top = x;
							}
							InvalidateRect(hwnd, NULL, FALSE);
						}
					}
					if (GetMenuState(hMenu, IDM_TOOLS_DRAW, MF_BYCOMMAND) == MF_CHECKED  &&  !bDrawLB  &&  !bDrawRB)
					{
						Factor = (double) ShowWidth / bitmap.bmWidth;
						x = long((xMousePos-xShow) / Factor);
						Factor = (double) ShowHeight / bitmap.bmHeight;
						y = long((yShow+ShowHeight-1-yMousePos) / Factor);
						if (x>=0  &&  x<bitmap.bmWidth  &&  y>=0  &&  y<bitmap.bmHeight)
						{
							SaveForUndo(hBitmap, TEXT("Zeichnen"));
							pMemBits = (BYTE*) malloc(bitmap.bmWidthBytes*bitmap.bmHeight);
							bDrawingLine = FALSE;
							if (!bChanged)
							{
							  bChanged = TRUE;
							  SetCaption(hwnd, szFileName, bChanged);
							}
							XORSelRect.left = XORSelRect.right = x;
							XORSelRect.bottom = XORSelRect.top = y;
							DrawLine(hBitmap, SelRect, XORSelRect, PenWidth, PenColorParam.RButtonGrayVal, pMemBits);
							InvalidateRect(hwnd, NULL, FALSE);
							bDrawRB = TRUE;
							SetCapture(hwnd);
						}
					}
					if (GetMenuState(hMenu,IDM_TOOLS_MOVE, MF_BYCOMMAND) == MF_CHECKED)
					{
						if ((wParam & MK_CONTROL)  &&  ShowSizeFactor>-10)
						{
							ShowSizeFactor -= 1.0f;
							Factor = ((double) (xMousePos-xShow)) / ((double) ShowWidth);
							ShowWidth = ShowSizeFactor<0 ? long(-bitmap.bmWidth/(ShowSizeFactor-1)) : long(bitmap.bmWidth*(ShowSizeFactor+1));
							xShow = int(cxClient/2 - Factor*ShowWidth);
							Factor = ((double) (yMousePos-yShow)) / ((double) ShowHeight);
							ShowHeight = long(ShowSizeFactor<0 ? -bitmap.bmHeight/(ShowSizeFactor-1) : bitmap.bmHeight*(ShowSizeFactor+1));
							yShow = int(cyClient/2 - Factor*ShowHeight);
							CheckBorder(xShow, yShow, cxClient, cyClient, ShowWidth, ShowHeight);
							InvalidateRect(hwnd, NULL, FALSE);
						}
						if (!(wParam & MK_CONTROL))
						{
						  xShow=(cxClient-ShowWidth)/2;
						  yShow=(cyClient-ShowHeight)/2;
							InvalidateRect(hwnd,0,FALSE);
						}
					}
			    return 0;

		 case WM_RBUTTONUP:
					if (bDrawRB)
					{
						ReleaseCapture();
						bDrawRB = FALSE;
						// Original wiederherstellen?
						//DrawLine(hBitmap, XORSelRect, PenWidth, -1);
						Factor = (double) ShowWidth / bitmap.bmWidth;
						XORSelRect.right = long(((signed short) xMousePos-xShow) / Factor);
						Factor = (double) ShowHeight / bitmap.bmHeight;
						XORSelRect.top = long((yShow+ShowHeight-1-(signed short) yMousePos) / Factor);
						DrawLine(hBitmap, SelRect, XORSelRect, PenWidth, PenColorParam.RButtonGrayVal, pMemBits);
						InvalidateRect(hwnd, NULL, FALSE);
						if (pMemBits)  free(pMemBits);
					}
			    return 0;

		 case WM_MOUSEMOVE:
					x = (signed short) LOWORD(lParam);
					y = (signed short) HIWORD(lParam);
					if (!(wParam & MK_CONTROL))  bZoom = FALSE;
					if (!hBitmap)
					{
						SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(NULL, IDC_ARROW) );
						SetDlgItemText (hInfoBox, IDC_X_POS, TEXT("X: (Kein Bild)"));
						SetDlgItemText (hInfoBox, IDC_Y_POS, TEXT("Y: (Kein Bild)"));
						SetDlgItemText (hInfoBox, IDC_GRAYVALUE, TEXT("Grauwert: ---"));
						xMousePos = x;
						yMousePos = y;
						return 0;
					}
					if (bDrawLB  ||  bDrawRB)
					{
						if (bDrawingLine)	DrawLine(hBitmap, SelRect, XORSelRect, PenWidth, -1, pMemBits);
						if (!bDrawingLine  &&  (wParam & MK_CONTROL))
						{
							if (pMemBits)
							{
								CopyMemory(pMemBits, pBits, bitmap.bmWidthBytes*bitmap.bmHeight);
								bDrawingLine = TRUE;
							}
						}
						if (bDrawingLine  &&  !(wParam & MK_CONTROL))
						{
							bDrawingLine = FALSE;
							if (bDrawLB)
								DrawLine(hBitmap, SelRect, XORSelRect, PenWidth, PenColorParam.LButtonGrayVal, pMemBits);
							else
								DrawLine(hBitmap, SelRect, XORSelRect, PenWidth, PenColorParam.RButtonGrayVal, pMemBits);
							XORSelRect.left = XORSelRect.right;
							XORSelRect.bottom = XORSelRect.top;
						}
						Factor = (double) ShowWidth / bitmap.bmWidth;
						XORSelRect.right = long((x-xShow) / Factor);
						Factor = (double) ShowHeight / bitmap.bmHeight;
						XORSelRect.top = long((yShow+ShowHeight-1-y) / Factor);
						if (bDrawLB)
							DrawLine(hBitmap, SelRect, XORSelRect, PenWidth, PenColorParam.LButtonGrayVal, pMemBits);
						else
							DrawLine(hBitmap, SelRect, XORSelRect, PenWidth, PenColorParam.RButtonGrayVal, pMemBits);
						if (!bDrawingLine)
						{
							XORSelRect.left = XORSelRect.right;
							XORSelRect.bottom = XORSelRect.top;
						}
						InvalidateRect(hwnd, NULL, FALSE);
					}
					if (bSelMoving)
					{
						x = x - xMousePos;
						Factor = (double) ShowWidth / bitmap.bmWidth;
						x = long(x/Factor);
						if (OldSelRect.right+x>bitmap.bmWidth-1)  x=bitmap.bmWidth-1-OldSelRect.right;
						if (OldSelRect.left+x<0)   x = -OldSelRect.left;
						y = yMousePos - y;
						Factor = (double) ShowHeight / bitmap.bmHeight;
						y = long(y/Factor);
						if (OldSelRect.top+y>bitmap.bmHeight-1)  y=bitmap.bmHeight-1-OldSelRect.top;
						if (OldSelRect.bottom+y<0)   y = -OldSelRect.bottom;
						SelRect.left   = OldSelRect.left   + x;
						SelRect.right  = OldSelRect.right  + x;
						SelRect.bottom = OldSelRect.bottom + y;
						SelRect.top    = OldSelRect.top    + y;
						bDelOldSelRect = TRUE;
						InvalidateRect(hwnd, NULL, FALSE);
					}
					if (bSelecting)
					{
						Factor = (double) ShowWidth / bitmap.bmWidth;
						SelRect.right = long((x-xShow) / Factor);
						Factor = (double) ShowHeight / bitmap.bmHeight;
						SelRect.bottom = long((yShow+ShowHeight-1-y) / Factor);
						if (SelRect.right<0)  SelRect.right=0;
						if (SelRect.right>=bitmap.bmWidth)  SelRect.right=bitmap.bmWidth-1;
						if (SelRect.bottom<0)  SelRect.bottom=0;
						if (SelRect.bottom>=bitmap.bmHeight) SelRect.bottom=bitmap.bmHeight-1;
						bDelOldSelRect = TRUE;
						InvalidateRect(hwnd, NULL, FALSE);
					}
					if (bScroll)
					{
						if (cxClient<ShowWidth)
						{
							if (xShow + x - xMousePos > 0)  xMousePos = xShow + x;
						  if (xShow + x - xMousePos < cxClient - ShowWidth)  xMousePos = xShow + x - cxClient + ShowWidth;
						}
						else
						{
							if (xShow + x - xMousePos < 0)  xMousePos = xShow + x;
							if (xShow + x - xMousePos > cxClient - ShowWidth)  xMousePos = xShow + x - cxClient + ShowWidth;
						}
						if (cyClient<ShowHeight)
						{
							if (yShow + y - yMousePos > 0)  yMousePos = yShow + y;
						  if (yShow + y - yMousePos < cyClient - ShowHeight)  yMousePos = yShow + y - cyClient + ShowHeight;
						}
						else
						{
							if (yShow + y - yMousePos < 0)  yMousePos = yShow + y;
							if (yShow + y - yMousePos > cyClient - ShowHeight)  yMousePos = yShow + y - cyClient + ShowHeight;
						}
						
						ScrollWindow (hwnd, x - xMousePos, y - yMousePos, NULL, NULL);
						xShow += x - xMousePos;
						yShow += y - yMousePos;
						UpdateWindow(hwnd);
					}
					else if (bZoom)
						SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(hInstance, TEXT("zoomglass")) );
					else
						SetClassLong(hwnd, GCL_HCURSOR, (long) LoadCursor(hInstance, TEXT("crosscursor")) );

					if (!bSelMoving)
					{
						xMousePos = x;
						yMousePos = y;
					}
			    //Pixelposition in InfoBox anzeigen...
					if (bSelecting)
					{
						wsprintf(szTemp, "Breite:  %li", abs(SelRect.right-SelRect.left)+1);
						SetDlgItemText(hInfoBox, IDC_X_POS, szTemp);
						wsprintf(szTemp, "Hhe:  %li", abs(SelRect.top-SelRect.bottom)+1);
						SetDlgItemText(hInfoBox, IDC_Y_POS, szTemp);
						wsprintf(szTemp, "  (der Markierung)");
						SetDlgItemText(hInfoBox, IDC_GRAYVALUE, szTemp);
					}
					else if (xMousePos<xShow || xMousePos>=xShow+ShowWidth || yMousePos<yShow || yMousePos>=yShow+ShowHeight)
					{
						SetDlgItemText (hInfoBox, IDC_X_POS, TEXT("X: (auerhalb)"));
						SetDlgItemText (hInfoBox, IDC_Y_POS, TEXT("Y: (auerhalb)"));
						SetDlgItemText (hInfoBox, IDC_GRAYVALUE, TEXT("Grauwert: ---"));
					}
					else if (bSelMoving)
					{
						wsprintf(szTemp, "X:  %li (%li)", SelRect.left, abs(SelRect.right-SelRect.left)+1);
						SetDlgItemText(hInfoBox, IDC_X_POS, szTemp);
						wsprintf(szTemp, "Y:  %li (%li)", SelRect.bottom, abs(SelRect.top-SelRect.bottom)+1);
						SetDlgItemText(hInfoBox, IDC_Y_POS, szTemp);
						wsprintf(szTemp, "  (Markierung)");
						SetDlgItemText(hInfoBox, IDC_GRAYVALUE, szTemp);
					}
					else
					{
						Factor = (double) ShowWidth / bitmap.bmWidth;
						x = long((xMousePos-xShow) / Factor);
						Factor = (double) ShowHeight / bitmap.bmHeight;
						y = long((yShow+ShowHeight-1-yMousePos) / Factor);
						wsprintf(szTemp, "X:  %li  (%li)", x, bitmap.bmWidth);
						SetDlgItemText(hInfoBox, IDC_X_POS, szTemp);
						wsprintf(szTemp, "Y:  %li  (%li)", y, bitmap.bmHeight);
						SetDlgItemText(hInfoBox, IDC_Y_POS, szTemp);
						x = pBits[y*bitmap.bmWidthBytes + x];
						wsprintf(szTemp, "Grauwert:  %li", x);
						SetDlgItemText(hInfoBox, IDC_GRAYVALUE, szTemp);
					}
					break;

		 case WM_PAINT:
			    if (!hBitmap)  InvalidateRect(hwnd, NULL, TRUE);
          hdc = BeginPaint (hwnd, &ps);
          if (hBitmap)
          {    // Bitmap im Anwendungsbereich zeichnen
					  if ((XORSelRect.left!=0  ||  XORSelRect.right!=bitmap.bmWidth-1  ||  XORSelRect.bottom!=0  ||  XORSelRect.top!=bitmap.bmHeight-1)  &&  bDelOldSelRect)
						{
							DrawSelectionRect(hdc, bitmap, xShow, yShow, ShowWidth, ShowHeight, XORSelRect, TRUE);
						}
						if (!bSelecting && !bSelMoving)
						{
							if (xShow>0)
							{
								rect.top = rect.left = 0;
								rect.right = xShow;
								rect.bottom = cyClient;
								FillRect(hdc, &rect, (HBRUSH) GetClassLong(hwnd, GCL_HBRBACKGROUND));
							}
							if (xShow+ShowWidth<cxClient)
							{
								rect.top = 0;
								rect.left = xShow+ShowWidth;
								rect.right = cxClient;
								rect.bottom = cyClient;
								FillRect(hdc, &rect, (HBRUSH) GetClassLong(hwnd, GCL_HBRBACKGROUND));
							}
							if (yShow>0)
							{
								rect.top = rect.left = 0;
								rect.right = cxClient;
								rect.bottom = yShow;
								FillRect(hdc, &rect, (HBRUSH) GetClassLong(hwnd, GCL_HBRBACKGROUND));
							}
							if (yShow+ShowHeight<cyClient)
							{
								rect.top = yShow+ShowHeight;
								rect.left = 0;
								rect.right = cxClient;
								rect.bottom = cyClient;
								FillRect(hdc, &rect, (HBRUSH) GetClassLong(hwnd, GCL_HBRBACKGROUND));
							}
							PaintTheBitmap(hdc, hBitmap, xShow, yShow, ShowWidth, ShowHeight, cxClient, cyClient);
						}
					  if (SelRect.left!=0  ||  SelRect.right!=bitmap.bmWidth-1  ||  SelRect.bottom!=0  ||  SelRect.top!=bitmap.bmHeight-1  ||  bDelOldSelRect)
						{
							bTemp = bSelecting || bSelMoving;
							DrawSelectionRect(hdc, bitmap, xShow, yShow, ShowWidth, ShowHeight, SelRect, bTemp);
							if (bSelecting  ||  bSelMoving)  XORSelRect = SelRect;
							bDelOldSelRect = FALSE;
						}
					}
          EndPaint (hwnd, &ps) ;
          return 0 ;
     
		 case WM_CLOSE:
				  if (hBitmap)
					{
					  if (bChanged)
						{
						  if (MessageBox(hwnd, TEXT("Die Datei wurde bearbeitet und noch nicht gespeichert! Wollen Sie Particle Tracer wirklich beenden?"),
							               TEXT("Particle Tracer beenden"), MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2) != IDYES)
										 return 0;
						}
					}
					break;

     case WM_DESTROY:
          if (hBitmap)  DeleteObject (hBitmap) ;
					ClrUndoBuf();
          PostQuitMessage (0) ;
          return 0 ;
     }
		 if (MSWHEEL_ROLLMSG!=0  &&  message == MSWHEEL_ROLLMSG)
			 SendMessage(hwnd, WM_MOUSEWHEEL, LOWORD(wParam)<<16, lParam);
     return DefWindowProc (hwnd, message, wParam, lParam) ;
}




void SetCaption(HWND hwnd, TCHAR *szFileName, BOOL bChanged)
{
	TCHAR szCaption[100 + MAX_PATH];

  if (!szFileName)
	{
		strcpy(szCaption, szLongAppName);
	}
	else
	{
		if (szFileName[0])
		  wsprintf (szCaption, TEXT("%s - %s"), szLongAppName, szFileName);
		else
			wsprintf (szCaption, TEXT("%s - %s"), szLongAppName, TEXT("Unbenannt"));
		if (bChanged)  strcat(szCaption, TEXT("  (gendert)"));
	}
 	SetWindowText (hwnd, szCaption);
}



BOOL CheckBorder(int &xShow, int &yShow, int cxClient, int cyClient, long ShowWidth, long ShowHeight)
{
	int xShowOld, yShowOld;
	
	xShowOld = xShow;
	yShowOld = yShow;
	if (cxClient<ShowWidth)
	{
		if (xShow > 0)  xShow = 0;
	  if (xShow < cxClient - ShowWidth)  xShow = cxClient - ShowWidth;
	}
	else
	{
		if (xShow < 0)  xShow = 0;
		if (xShow > cxClient - ShowWidth)  xShow = cxClient - ShowWidth;
	}
	if (cyClient<ShowHeight)
	{
		if (yShow > 0)  yShow = 0;
	  if (yShow < cyClient - ShowHeight)  yShow = cyClient - ShowHeight;
	}
	else
	{
		if (yShow < 0)  yShow = 0;
		if (yShow > cyClient - ShowHeight)  yShow = cyClient - ShowHeight;
	}
	if (xShowOld!=xShow  ||  yShowOld!=yShow)
		return TRUE;
	else
		return FALSE;
}



BOOL PaintTheBitmap(HDC hdc, HBITMAP hBitmap, int xShow, int yShow, long ShowWidth, long ShowHeight, int cxClient, int cyClient)
{
	BITMAP bitmap;
	HDC    hdcMem;
	BOOL   bPainted;
	int    xOffset, yOffset, bX, bY, bW, bH, nr;
	double Factor;

	xOffset = xShow;
	yOffset = yShow;
	GetObject (hBitmap, sizeof (BITMAP), &bitmap);
	bX = 0;
	bY = 0;
	bW = bitmap.bmWidth;
	bH = bitmap.bmHeight;

	Factor = (double) ShowWidth / bitmap.bmWidth;
	if (xOffset<0)
	{
		nr = int(-xOffset / Factor);
		xOffset += int(nr * Factor);
		ShowWidth -= long(nr * Factor);
		bX = nr;
		bW -= nr;
	}
	if (xOffset+ShowWidth-1 > cxClient)
	{
		nr = int((xOffset+ShowWidth-1-cxClient) / Factor);
		ShowWidth -= long(nr * Factor);
		bW -= nr;
	}

	Factor = (double) ShowHeight / bitmap.bmHeight;
	if (yOffset<0)
	{
		nr = int(-yOffset / Factor);
		yOffset += int(nr * Factor);
		ShowHeight -= long(nr * Factor);
		bY = nr;
		bH -= nr;
	}
	if (yOffset+ShowHeight-1 > cyClient)
	{
		nr = int((yOffset+ShowHeight-1-cyClient) / Factor);
		ShowHeight -= long(nr * Factor);
		bH -= nr;
	}

	hdcMem = CreateCompatibleDC (hdc);
	SelectObject (hdcMem, hBitmap);
  
	SetStretchBltMode(hdc, HALFTONE);
	bPainted = StretchBlt (hdc, xOffset, yOffset, ShowWidth, ShowHeight,
		                     hdcMem, bX, bY, bW, bH, SRCCOPY);
	DeleteDC (hdcMem);

	return bPainted;
}



void DrawSelectionRect(HDC hdc, BITMAP bitmap, int xShow, int yShow, long ShowWidth, long ShowHeight, RECT SelRect, BOOL bSelecting)
{
	HPEN   hPen, hOldPen;
	long   tmpLong;
	double FactorX, FactorY;

	FactorX = (double) ShowWidth / bitmap.bmWidth;
	FactorY = (double) ShowHeight / bitmap.bmHeight;

	if (SelRect.left!=0  ||  SelRect.right!=bitmap.bmWidth-1  ||  SelRect.bottom!=0  ||  SelRect.top!=bitmap.bmHeight-1)
	{
    if (SelRect.right<SelRect.left)
		{
			tmpLong = SelRect.right;
			SelRect.right = SelRect.left;
			SelRect.left = tmpLong;
		}
    if (SelRect.top<SelRect.bottom)
		{
			tmpLong = SelRect.bottom;
			SelRect.bottom = SelRect.top;
			SelRect.top = tmpLong;
		}
		hPen = CreatePen(PS_SOLID,0,RGB(64,64,255));
		hOldPen=(HPEN) SelectObject(hdc, hPen);
		if (bSelecting)
			SetROP2(hdc, R2_XORPEN);
		else
			SetROP2(hdc, R2_COPYPEN);
		MoveToEx(hdc, int(xShow+SelRect.left*FactorX-1),    int(yShow+(bitmap.bmHeight-1-SelRect.top)*FactorY-1), NULL);
		LineTo  (hdc, int(xShow+(SelRect.right+1)*FactorX), int(yShow+(bitmap.bmHeight-1-SelRect.top)*FactorY-1));
		LineTo  (hdc, int(xShow+(SelRect.right+1)*FactorX), int(yShow+(bitmap.bmHeight-SelRect.bottom)*FactorY));
		LineTo  (hdc, int(xShow+SelRect.left*FactorX-1),    int(yShow+(bitmap.bmHeight-SelRect.bottom)*FactorY));
		LineTo  (hdc, int(xShow+SelRect.left*FactorX-1),    int(yShow+(bitmap.bmHeight-1-SelRect.top)*FactorY-1));
		DeleteObject(SelectObject(hdc, hOldPen));
	}
}



void DrawLine(HBITMAP hBitmap, RECT SelRect, RECT DrawRect, int PenWidth, int GrayVal, BYTE *pMemBits)
{
	BITMAP bitmap;
	BYTE   *pBits;
	long   x, y, i, j;
	double f;

	if (!hBitmap)  return;
	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	if (DrawRect.right-DrawRect.left==0  &&  DrawRect.top-DrawRect.bottom==0)
	{
		for (j=DrawRect.top-PenWidth/2; j<=DrawRect.top+(PenWidth-1)/2; j++)
		{
			if (j>=SelRect.bottom  &&  j<=SelRect.top)
			{
				for (i=DrawRect.left-(PenWidth-1)/2; i<=DrawRect.left+PenWidth/2; i++)
				{
					if (i>=SelRect.left  &&  i<=SelRect.right)
						if (GrayVal<0)
							pBits[j*bitmap.bmWidthBytes+i] = pMemBits[j*bitmap.bmWidthBytes+i];
						else
							pBits[j*bitmap.bmWidthBytes+i] = GrayVal;
				}
			}
		}
		return;
	}
	if (abs(DrawRect.right-DrawRect.left)>abs(DrawRect.top-DrawRect.bottom))
	{
		if (DrawRect.left>DrawRect.right)
		{
			x = DrawRect.left;
			DrawRect.left = DrawRect.right;
			DrawRect.right = x;
			y = DrawRect.bottom;
			DrawRect.bottom = DrawRect.top;
			DrawRect.top = y;
		}
		f = (double) (DrawRect.top-DrawRect.bottom) / (DrawRect.right-DrawRect.left);
		for (x=DrawRect.left; x<=DrawRect.right; x++)
		{
			y = long(DrawRect.bottom + f*(x-DrawRect.left) + 0.5);
			for (j=y-PenWidth/2; j<=y+(PenWidth-1)/2; j++)
			{
				if (j>=SelRect.bottom  &&  j<=SelRect.top)
				{
					for (i=x-(PenWidth-1)/2; i<=x+PenWidth/2; i++)
					{
						if (i>=SelRect.left  &&  i<=SelRect.right)
							if (GrayVal<0)
								pBits[j*bitmap.bmWidthBytes+i] = pMemBits[j*bitmap.bmWidthBytes+i];
							else
								pBits[j*bitmap.bmWidthBytes+i] = GrayVal;
					}
				}
			}
		}
	}
	else
	{
		if (DrawRect.bottom>DrawRect.top)
		{
			x = DrawRect.left;
			DrawRect.left = DrawRect.right;
			DrawRect.right = x;
			y = DrawRect.bottom;
			DrawRect.bottom = DrawRect.top;
			DrawRect.top = y;
		}
		f = (double) (DrawRect.right-DrawRect.left) / (DrawRect.top-DrawRect.bottom);
		for (y=DrawRect.bottom; y<=DrawRect.top; y++)
		{
			x = long(DrawRect.left + f*(y-DrawRect.bottom) + 0.5);
			for (j=y-PenWidth/2; j<=y+(PenWidth-1)/2; j++)
			{
				if (j>=SelRect.bottom  &&  j<=SelRect.top)
				{
					for (i=x-(PenWidth-1)/2; i<=x+PenWidth/2; i++)
					{
						if (i>=SelRect.left  &&  i<=SelRect.right)
							if (GrayVal<0)
								pBits[j*bitmap.bmWidthBytes+i] = pMemBits[j*bitmap.bmWidthBytes+i];
							else
								pBits[j*bitmap.bmWidthBytes+i] = GrayVal;
					}
				}
			}
		}
	}
}




/*******************************************************
 *                                                     *
 *              DIALOGPROZEDUR:   AboutBox             *
 *                                                     *
 *******************************************************/

BOOL CALLBACK AboutBoxProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
		{
		case WM_INITDIALOG:
			return TRUE;

		case WM_COMMAND:
			switch (LOWORD (wParam))
			{
			case IDOK:
			case IDCANCEL:
				EndDialog (hDlg, 0);
				return TRUE;
			}
			break;
	}
	return FALSE;
}




/*******************************************************
 *                                                     *
 *              DIALOGPROZEDUR:    InfoBox             *
 *                                                     *
 *******************************************************/

BOOL CALLBACK InfoBoxProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
		{
		case WM_INITDIALOG:
			return TRUE;

		case WM_COMMAND:
			switch (LOWORD (wParam))
			{
			case IDOK:
			case IDCANCEL:
				return TRUE;
			}
			break;
	}
	return FALSE;
}
