#include "DlgProcs.h"

#define BORDER 10


void CalcHisto(HBITMAP hBitmap, float* Histo);
void ArrangeButtons(HWND hDlg);
BOOL LoadNewPal(TCHAR *szFileName, BYTE *LoadPal);
BOOL SaveNewPal(TCHAR *szFileName, BYTE NewPal[]);


/*******************************************************
 *                                                     *
 *              DIALOGPROZEDUR:   PaletteDlg           *
 *                                                     *
 *******************************************************/

BOOL CALLBACK PaletteDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	PPALETTEPARAM    pPaletteParam;
	static HINSTANCE hInstance;
	static HWND      hInfoBox;
	TCHAR            szBuffer[80], *szEndConv;
	static OPENFILENAME ofn;
  static TCHAR     szFileName [MAX_PATH], szTitleName [MAX_PATH] ;
  static TCHAR     szFilter[] = TEXT ("Paletten-Dateien (*.PAL)\0*.pal\0")
                                TEXT ("Alle Dateien (*.*)\0*.*\0\0");
	static HBITMAP   hBitmap;
	BITMAP           bitmap;
	static BYTE      *pBits = NULL, NewPal[256];
	static float	   Histo[256];
	HDC						   hdc;
	PAINTSTRUCT      ps;
	RECT             rect;
	PENCOLORPARAM    PenColorParam;
	static int       x1, y1, x2, y2, SelBegin = -1, SelEnd = -1;
	int              iResult;
	static BOOL      bDrawing = FALSE, bDrawOutOfSel = FALSE, bSelecting = FALSE, bColorMarked = FALSE, bChanged;
	BOOL             bTmp;
	long             x, y, Offset;
	double           f, fExp, fOff;
	BYTE             bMin, bMax, *LoadPal;

	
	switch (message)
		{
		case WM_INITDIALOG:
			pPaletteParam = (PPALETTEPARAM) lParam;
			hInstance = pPaletteParam->hInstance;
			hBitmap = pPaletteParam->hBitmap;
			hInfoBox = pPaletteParam->hInfoBox;
			SetDlgItemText(hInfoBox, IDC_GRAYVALUE, TEXT("  (Grauwerte)"));
			if (!hBitmap) 
			{
				EndDialog(hDlg, FALSE);
				return TRUE;
			}
		 	for (Offset = 0; Offset < 256; Offset++)  NewPal[Offset]=(BYTE) Offset;
			GetObject (hBitmap, sizeof (BITMAP), &bitmap);
			pBits = (BYTE*) bitmap.bmBits;
			CalcHisto(hBitmap, Histo);
	    ofn.lStructSize       = sizeof (OPENFILENAME);
      ofn.hwndOwner         = hDlg;
      ofn.hInstance         = NULL;
      ofn.lpstrFilter       = szFilter;
      ofn.lpstrCustomFilter = NULL;
      ofn.nMaxCustFilter    = 0;
      ofn.nFilterIndex      = 0;
      ofn.lpstrFile         = szFileName;
      ofn.nMaxFile          = MAX_PATH;
      ofn.lpstrFileTitle    = szTitleName;
      ofn.nMaxFileTitle     = MAX_PATH;
      ofn.lpstrInitialDir   = NULL;
      ofn.lpstrTitle        = NULL;
      ofn.Flags             = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST;
      ofn.nFileOffset       = 0;
      ofn.nFileExtension    = 0;
      ofn.lpstrDefExt       = TEXT ("bmp");
      ofn.lCustData         = 0;
      ofn.lpfnHook          = NULL;
      ofn.lpTemplateName    = NULL;
			bChanged = FALSE;
			SetDlgItemInt(hDlg, IDC_EXPONENT, 1, FALSE);
			ArrangeButtons(hDlg);
			return TRUE;

		case WM_COMMAND:
			switch (LOWORD (wParam))
			{
			case IDM_FILE_OPEN:
			case IDC_LOADPAL:
				LoadPal = (BYTE*) malloc(256);
				if (!LoadPal)  return FALSE;
				ofn.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST;
				if (!GetOpenFileName(&ofn))
				{
					free(LoadPal);
					return 0;
				}
				if (!LoadNewPal(szFileName, LoadPal))
				{
					MessageBox(hDlg, TEXT("Beim Laden ist ein Fehler aufgetreten! Falsches Dateiformat?"),
						            TEXT("Particle Tracer - Palette laden"), MB_OK | MB_ICONEXCLAMATION);
					free(LoadPal);
					return TRUE;
				}
				if (SelBegin<0  ||  SelEnd<0)
				{
					for (Offset=0; Offset<256; Offset++)  NewPal[Offset] = LoadPal[Offset];
				}
				else
				{
					iResult = MessageBox(hDlg, TEXT("Sie haben einen Farbbereich markiert. Die neu eingeladene Farbpalette wird\n")
						                         TEXT("grundstzlich nur in diesem Bereich eingefgt.\n\n")
																		 TEXT("Wollen Sie die neu eingeladene Palette auf die Breite der Markierung stauchen?"), TEXT("Particle Tracer - Palette einladen"), MB_YESNOCANCEL | MB_ICONQUESTION);
					if (iResult == IDCANCEL)
					{
						free(LoadPal);
						return TRUE;
					}
					if (iResult == IDYES)
					{
						f = 255.0;
						if (SelEnd-SelBegin>0)  f /= SelEnd-SelBegin;
						for (Offset=SelBegin; Offset<=SelEnd; Offset++)  NewPal[Offset] = LoadPal[(long) ((Offset-SelBegin)*f)];
					}
					else
					{
						if (SelBegin>0)
							iResult = MessageBox(hDlg, TEXT("Da die neu eingeladene Palette nicht gestaucht werden soll, beinhaltet sie mehr ")
								                         TEXT("neue Farbwerte\nals die Markierung breit ist. Alle Eintrge, die sich auerhalb ")
																				 TEXT("dieser Markierung befinden, werden\nabgeschnitten.\n\nSoll der Anfang der neuen ")
																				 TEXT("Palettenwerte an den Anfang der Markierung verschoben werden?"), TEXT("Particle Tracer - Palette einladen"), MB_YESNOCANCEL | MB_ICONQUESTION);
						else
							iResult = IDNO;
						if (iResult == IDCANCEL)
						{
							free(LoadPal);
							return TRUE;
						}
            if (iResult == IDYES)  x = SelBegin;  else  x = 0;
						for (Offset=SelBegin; Offset<=SelEnd; Offset++)
						{
							NewPal[Offset] = LoadPal[Offset-x];
						}
					}
				}
				free(LoadPal);
				SetGrayPalette(hBitmap, NewPal);
				InvalidateRect(hDlg, NULL, FALSE);
				InvalidateRect(GetParent(hDlg), NULL, FALSE);
				return TRUE;

			case IDM_FILE_SAVE:
			case IDC_SAVEPAL:
        ofn.Flags = OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT;
		    if (!GetSaveFileName (&ofn))  return 0;
				if (!SaveNewPal(szFileName, NewPal))
				{
					MessageBox(hDlg, TEXT("Beim Speichern ist ein Fehler aufgetreten! Versuchen Sie einen anderen Dateinamen oder ein anderes Verzeichnis!"),
						            TEXT("Particle Tracer - Palette speichern"), MB_OK | MB_ICONEXCLAMATION);
				}
				return TRUE;

			case IDM_SEL_MANUAL:
				if (SelBegin<0  ||  SelEnd<0)
				{
					PenColorParam.LButtonGrayVal = 0;
					PenColorParam.RButtonGrayVal = 255;
				}
				else
				{
					PenColorParam.LButtonGrayVal = SelBegin;
					PenColorParam.RButtonGrayVal = SelEnd;
				}
				if (DialogBoxParam(hInstance, TEXT("PenColorDlg"), hDlg, PenColorDlgProc, (LPARAM) &PenColorParam))
				{
					if (PenColorParam.LButtonGrayVal>PenColorParam.RButtonGrayVal)
					{
						SelBegin = PenColorParam.RButtonGrayVal;
						SelEnd = PenColorParam.LButtonGrayVal;
					}
					else
					{
						SelBegin = PenColorParam.LButtonGrayVal;
						SelEnd = PenColorParam.RButtonGrayVal;
					}
					InvalidateRect(hDlg, NULL, FALSE);
				}
				return TRUE;
			
			case IDM_SEL_DELETE:
				SelBegin = SelEnd = -1;
				InvalidateRect(hDlg, NULL, FALSE);
				return TRUE;

			case IDC_NORMALIZE:
				GetObject (hBitmap, sizeof (BITMAP), &bitmap);
				if (SelBegin>-1  &&  SelBegin<256  &&  SelEnd>-1  &&  SelEnd<256)
				{
					bMin = SelBegin;
					bMax = SelEnd;
					if (SelBegin>SelEnd)  MessageBeep(0);
				}
				else
				{
					bMin = 0;
					bMax = 255;
				}
				while (Histo[bMin] == 0  &&  bMin<255)	bMin++;
				while (Histo[bMax] == 0  &&  bMax>0)	bMax--;

				for (Offset = 0; Offset < 256; Offset++)
				{
					if (Offset<=bMin)  NewPal[Offset] = 0;
					else if (Offset>=bMax)  NewPal[Offset]=255;
					else
					{
						f = 256.0 / (bMax-bMin);
						NewPal[Offset] = BYTE((Offset-bMin) * f);
					}
				}

				SetGrayPalette(hBitmap, NewPal);
				InvalidateRect(hDlg, NULL, FALSE);
				InvalidateRect(GetParent(hDlg), NULL, FALSE);
				return TRUE;

			case IDC_EXPONENT:
				if (HIWORD(wParam) == EN_CHANGE)
				{
					GetDlgItemText(hDlg, IDC_EXPONENT, szBuffer, 80);
					fExp = strtod(szBuffer, &szEndConv);
					if (fExp<=0 || strlen(szEndConv) != 0)
						EnableWindow(GetDlgItem(hDlg, IDC_POWER), FALSE);
					else
						EnableWindow(GetDlgItem(hDlg, IDC_POWER), TRUE);
				}
				return TRUE;
			
			case IDOK:
				if (GetFocus() != GetDlgItem(hDlg, IDC_EXPONENT))  return TRUE;
			case IDC_POWER:
				GetDlgItemText(hDlg, IDC_EXPONENT, szBuffer, 80);
				fExp = strtod(szBuffer, &szEndConv);
				if (fExp<=0 || strlen(szEndConv) != 0)
				{
					MessageBeep(MB_ICONHAND);
					return TRUE;
				}
				if (SelBegin<0  ||  SelEnd<0)
				{
					f = 255.0/pow(255.0, fExp);
					fOff = 0.49;
					for (Offset = 0; Offset < 256; Offset++)
					  NewPal[Offset] = BYTE(f*pow((double) Offset, fExp)+fOff);
				}
				else
				{
					if (SelEnd-SelBegin<3)  return FALSE;
					f = (NewPal[SelEnd]-NewPal[SelBegin])/pow((double) SelEnd-SelBegin, fExp);
					fOff = NewPal[SelBegin] + 0.49;
					for (Offset = SelBegin; Offset <= SelEnd; Offset++)
					  NewPal[Offset] = BYTE(f*pow((double) Offset-SelBegin, fExp)+fOff);
				}
				SetGrayPalette(hBitmap, NewPal);
				InvalidateRect(hDlg, NULL, FALSE);
				InvalidateRect(GetParent(hDlg), NULL, FALSE);
				return TRUE;

			case IDC_USE_NEW_PAL:
				GetObject(hBitmap, sizeof (BITMAP), &bitmap);
        SetStandardPalette(hBitmap);
				SaveForUndo(hBitmap, TEXT("Palette"));
				bChanged = TRUE;
				for (y = 0; y < bitmap.bmHeight; y++)
				{
				  Offset = y * bitmap.bmWidthBytes;
				  for (x = 0; x < bitmap.bmWidth; x++)
					{
						pBits[Offset] = NewPal[pBits[Offset]];
					  Offset++;
					}
				}
				for (Offset = 0; Offset < 256; Offset++)  NewPal[Offset]=(BYTE) Offset;
        CalcHisto(hBitmap, Histo);
				SelBegin = SelEnd = -1;
				InvalidateRect(hDlg, NULL, FALSE);
				InvalidateRect(GetParent(hDlg), NULL, FALSE);
				return TRUE;

			case IDCANCEL:
				SetStandardPalette(hBitmap);
				InvalidateRect(GetParent(hDlg), NULL, FALSE);
				EndDialog(hDlg, bChanged);
				return TRUE;
			}
			break;

			case WM_LBUTTONDOWN:
				x1 = (signed short) LOWORD(lParam)-BORDER;
				y1 = 255 - (HIWORD(lParam)-BORDER);
				if (x1>-20  &&  x1<276  &&  y1>-20  &&  y1<276)
				{
					if (wParam & MK_SHIFT)
					{
						bSelecting = TRUE;
						SelBegin = x1;
						if (SelBegin<0)  SelBegin = 0;
						if (SelBegin>255)  SelBegin = 255;
						SelEnd = SelBegin;
						return TRUE;
					}
					bDrawing = TRUE;
					if (wParam & MK_CONTROL)
						bDrawOutOfSel = TRUE;
					else
						bDrawOutOfSel = FALSE;
					Offset = x1;
					if (Offset<0)  Offset = 0;
					if (Offset>255)  Offset = 255;
					y = y1;
					if (y<0)  y = 0;
					if (y>255)  y = 255;
					bTmp = FALSE;
					if (SelBegin<0  ||  SelEnd<0)
					{
						NewPal[Offset] = (BYTE) y;
						bTmp = TRUE;
					}
					else
					{
						if (Offset>=SelBegin  &&  Offset<=SelEnd)
						{
							if (!bDrawOutOfSel)
							{
								NewPal[Offset] = (BYTE) y;
								bTmp = TRUE;
							}
						}
						else
						{
							if (bDrawOutOfSel)
							{
								NewPal[Offset] = (BYTE) y;
								bTmp = TRUE;
							}
						}
					}
					if (bTmp)
					{
						SetGrayPalette(hBitmap, NewPal);
						InvalidateRect(hDlg, NULL, FALSE);
						InvalidateRect(GetParent(hDlg), NULL, FALSE);
					}
					SetCapture(hDlg);
				}
				return TRUE;

			case WM_MOUSEMOVE:
				x2 = ((signed short) LOWORD(lParam))-BORDER;
				y2 = 255 - (((signed short) HIWORD(lParam))-BORDER);
				if (x2>-20  &&  x2<276  &&  y2>-20  &&  y2<276)
				{
					x = x2;
					if (x<0)  x=0;
					if (x>255)  x=255;
					y = y2;
					if (y<0)  y=0;
					if (y>255)  y=255;
					wsprintf(szBuffer, TEXT("ALT:  %li"), x);
					SetDlgItemText(hInfoBox, IDC_X_POS, szBuffer);
					wsprintf(szBuffer, TEXT("NEU:  %li"), y);
					SetDlgItemText(hInfoBox, IDC_Y_POS, szBuffer);
				}
				else
				{
					SetDlgItemText(hInfoBox, IDC_X_POS, TEXT("ALT: (auerhalb)"));
					SetDlgItemText(hInfoBox, IDC_Y_POS, TEXT("NEU: (auerhalb)"));
				}
				if (bColorMarked  && !(wParam&MK_SHIFT))
				{
					SetGrayPalette(hBitmap, NewPal);
					InvalidateRect(GetParent(hDlg), NULL, FALSE);
				}
				if (!bSelecting  &&  !bDrawing  &&  wParam&MK_SHIFT)
				{
					SetGrayPalette(hBitmap, NewPal);
					SetSingleRGBPaletteEntry(hBitmap, (BYTE) x, RGB(0, 255, 0));
					bColorMarked = TRUE;
					InvalidateRect(GetParent(hDlg), NULL, FALSE);
				}
				if (bSelecting)
				{
					SetGrayPalette(hBitmap, NewPal);
					if (wParam & MK_SHIFT)
					{
						SelEnd = x2;
						if (SelEnd<0)  SelEnd = 0;
						if (SelEnd>255)  SelEnd = 255;
						SetSingleRGBPaletteEntry(hBitmap, SelBegin, RGB(0, 255, 0));
						SetSingleRGBPaletteEntry(hBitmap, SelEnd, RGB(255,0,0));
						InvalidateRect(hDlg, NULL, FALSE);
					}
					else
					{
						bSelecting = FALSE;
						if (SelEnd<SelBegin)
						{
							x = SelEnd;
							SelEnd = SelBegin;
							SelBegin = x;
						}
						SetGrayPalette(hBitmap, NewPal);
					}
					InvalidateRect(GetParent(hDlg), NULL, FALSE);
				}

				if (!bDrawing)  return FALSE;
				bTmp = FALSE;
				if (x2<-40  ||  x2>296  ||  y2<-40  ||  y2>296)
				{
					ReleaseCapture();
					bDrawing = FALSE;
					return FALSE;
				}
				if (x1==x2)
				{
					Offset = x2;
					if (Offset<0)  Offset = 0;
					if (Offset>255)  Offset = 255;
					y = y2;
					if (y<0)  y = 0;
					if (y>255)  y = 255;
					if (SelBegin<0  ||  SelEnd<0)
					{
						NewPal[Offset] = (BYTE) y;
						bTmp = TRUE;
					}
					else
					{
						if (Offset>=SelBegin  &&  Offset<=SelEnd)
						{
							if (!bDrawOutOfSel)
							{
								NewPal[Offset] = (BYTE) y;
								bTmp = TRUE;
							}
						}
						else
						{
							if (bDrawOutOfSel)
							{
								NewPal[Offset] = (BYTE) y;
								bTmp = TRUE;
							}
						}
					}
				}
				else
				{
					f = (double) (y2-y1) / (x2-x1);
					x = x1;
					do
					{
						Offset = x;
						if (Offset<0)  Offset = 0;
						if (Offset>255)  Offset = 255;
						y = long(y1 + f*(x-x1));
						if (y<0)  y = 0;
						if (y>255)  y = 255;
						if (SelBegin<0  ||  SelEnd<0)
						{
							NewPal[Offset] = (BYTE) y;
							bTmp = TRUE;
						}
						else
						if (Offset>=SelBegin  &&  Offset<=SelEnd)
						{
							if (!bDrawOutOfSel)
							{
								NewPal[Offset] = (BYTE) y;
								bTmp = TRUE;
							}
						}
						else
						{
							if (bDrawOutOfSel)
							{
								NewPal[Offset] = (BYTE) y;
								bTmp = TRUE;
							}
						}
						if (x1>x2)  x--;  else  x++;
					} while (x != x2);
				}
				x1 = x2;
				y1 = y2;
				if (bTmp)
				{
					SetGrayPalette(hBitmap, NewPal);
					InvalidateRect(hDlg, NULL, FALSE);
					InvalidateRect(GetParent(hDlg), NULL, FALSE);
				}
				return TRUE;

			case WM_LBUTTONUP:
				ReleaseCapture();
				bDrawing = FALSE;
				if (bSelecting)
				{
					SetGrayPalette(hBitmap, NewPal);
					SelEnd = (signed short) LOWORD(lParam);
					SelEnd -= BORDER;
					if (SelEnd>255) SelEnd=255;
					if (SelEnd<0)   SelEnd=0;
					if (SelEnd<SelBegin)
					{
						x = SelBegin;
						SelBegin = SelEnd;
						SelEnd = x;
					}
					InvalidateRect(hDlg, NULL, FALSE);
					InvalidateRect(GetParent(hDlg), NULL, FALSE);
				}
				bSelecting = FALSE;
				return TRUE;

			case WM_RBUTTONUP:
        if (bDrawing || bSelecting)  return FALSE;
				SelBegin = SelEnd = -1;
				InvalidateRect(hDlg, NULL, FALSE);
				return TRUE;

			case WM_PAINT:
				InvalidateRect(hDlg, NULL, FALSE);
				hdc = BeginPaint(hDlg, &ps);
				Rectangle(hdc, BORDER-1, BORDER-1, BORDER+257, BORDER+257);
				SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(100,100,255)));
				for (x = 0; x < 256; x++)
				{
					MoveToEx(hdc, BORDER+x, BORDER+255, NULL);
					LineTo(hdc, BORDER+x, int(BORDER+254-Histo[x]));
					SetPixel(hdc, BORDER+x, BORDER+255-NewPal[x], 0);
				}
				if (SelBegin>=0  &&  SelEnd>=0)
				{
				  rect.top    = BORDER;
					rect.bottom = BORDER + 256;
					rect.left   = BORDER;
					rect.left  += SelBegin<SelEnd ? SelBegin : SelEnd;
					rect.right  = BORDER + 1;
					rect.right += SelBegin<SelEnd ? SelEnd : SelBegin;
					InvertRect(hdc, &rect);
				}
				DeleteObject(SelectObject(hdc, GetStockObject(BLACK_PEN)));
				EndPaint(hDlg, &ps);
				return TRUE;
			
		case WM_DESTROY:
			return TRUE;
	}
	return FALSE;
}




void CalcHisto(HBITMAP hBitmap, float* Histo)
{
  BITMAP bitmap;
	long   x, y, Offset, Counts[256];
	BYTE   *pBits;
	double f;
	
	GetObject(hBitmap, sizeof (BITMAP), &bitmap);
	pBits = (BYTE*) bitmap.bmBits;
	for (Offset = 0; Offset<256; Offset++)  Counts[Offset]=0;
	for (y = 0; y < bitmap.bmHeight; y++)
	{
	  Offset = y * bitmap.bmWidthBytes;
	  for (x = 0; x < bitmap.bmWidth; x++)
		{
			Counts[pBits[Offset]]++;
		  Offset++;
		}
	}
	y = 0;
	for (Offset = 0; Offset<256; Offset++)  if (Counts[Offset]>y)  y = Counts[Offset];
	f = (double) 255 / y;
	for (Offset = 0; Offset<256; Offset++)  Histo[Offset] = float(f*Counts[Offset]);
}



void ArrangeButtons(HWND hDlg)
{
	RECT rect;
	int  Width, Height, BorderWidth, BorderHeight, cxClient, cyClient;
	long x, y;

	GetWindowRect(GetDlgItem(hDlg, IDCANCEL), &rect);
	Width = abs(rect.right-rect.left);
	Height = abs(rect.bottom-rect.top);
	GetClientRect(hDlg, &rect);
	BorderWidth = abs(rect.right-rect.left);
	BorderHeight = abs(rect.bottom-rect.top);
	GetWindowRect(hDlg, &rect);
	BorderWidth = abs(rect.right-rect.left) - BorderWidth;
	BorderHeight = abs(rect.bottom-rect.top) - BorderHeight;
	cxClient = BORDER+257+3*Width/2;
	cyClient = BORDER+257+2*Height;
	if (cxClient<3*Width)  cxClient = 3*Width;
	if (cyClient<5*Height) cyClient = 5*Height;
	MoveWindow(hDlg, rect.left, rect.top, cxClient+BorderWidth, cyClient+BorderHeight, FALSE);
	MoveWindow(GetDlgItem(hDlg, IDC_USE_NEW_PAL), (cxClient-3*Width)/2, BORDER+257+Height/2, Width, Height, FALSE);
	MoveWindow(GetDlgItem(hDlg, IDCANCEL), (cxClient+Width)/2, BORDER+257+Height/2, Width, Height, FALSE);
	GetWindowRect(GetDlgItem(hDlg, IDC_LOADPAL), &rect);
	Height = abs(rect.bottom-rect.top);
  x = BORDER+257+Width/4;
	y = BORDER;
	MoveWindow(GetDlgItem(hDlg, IDC_LOADPAL), x, y, Width, Height, FALSE);
	y += 3*Height/2;
	MoveWindow(GetDlgItem(hDlg, IDC_SAVEPAL), x, y, Width, Height, FALSE);
	y += 2*Height;
	MoveWindow(GetDlgItem(hDlg, IDC_NORMALIZE), x, y, Width, Height, FALSE);
	y += 3*Height/2;
	MoveWindow(GetDlgItem(hDlg, IDC_EXP_CAPTION), x, y+Height/4, Width, Height*3/4, FALSE);
	y += Height;
	MoveWindow(GetDlgItem(hDlg, IDC_EXPONENT), x, y, Width, Height*3/4, FALSE);
	y += Height;
	MoveWindow(GetDlgItem(hDlg, IDC_POWER), x, y, Width, Height, FALSE);
}



BOOL SaveNewPal(char *szFileName, BYTE NewPal[])
{
	HANDLE hFile;
	char   szBuf[256];
	DWORD  iResult;
	int    i;

	hFile = CreateFile(szFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
		                 FILE_ATTRIBUTE_ARCHIVE | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE)  return FALSE;
	strcpy(szBuf, TEXT("Particle Tracer - Palette:\r\n"));
	if (!WriteFile(hFile, szBuf, strlen(szBuf), &iResult, NULL))
	{
		CloseHandle(hFile);
		return FALSE;
	}
  for (i=0; i<256; i++)
	{
		itoa(NewPal[i], szBuf, 10);
		strcat(szBuf, "\r\n");
		if (!WriteFile(hFile, szBuf, strlen(szBuf), &iResult, NULL))
		{
			CloseHandle(hFile);
			return FALSE;
		}
	}
	CloseHandle(hFile);
	return TRUE;
}



BOOL LoadNewPal(char *szFileName, BYTE *LoadPal)
{
	HANDLE hFile;
	char   *szBuf;
	DWORD  dwLength, Offset;
	int    i, iTmp;

	hFile = CreateFile(szFileName, GENERIC_READ, 0, NULL, OPEN_EXISTING,
		                 FILE_SHARE_READ, NULL);
	if (hFile == INVALID_HANDLE_VALUE)  return FALSE;
	Offset = GetFileSize(hFile, NULL);
	if (Offset == 0xFFFFFFFF)   // Fehler?
	{
		CloseHandle(hFile);
		return FALSE;
	}
	szBuf = (char*) malloc(Offset+1);
	if (!szBuf)
	{
		CloseHandle(hFile);
		return FALSE;
	}
	if (!ReadFile(hFile, szBuf, Offset, &dwLength, NULL))
	{
		CloseHandle(hFile);
		free(szBuf);
		return FALSE;
	}
	CloseHandle(hFile);
	if (dwLength != Offset)
	{
		free(szBuf);
		return FALSE;
	}
	szBuf[dwLength] = '\0';
	if (strncmp(szBuf, TEXT("Particle Tracer - Palette:\r\n"), 28) != 0)
	{
		free(szBuf);
		return FALSE;
	}
	Offset = 28;
	i = 0;
	while (1)
	{
		iTmp = atoi(szBuf+Offset);
		if (iTmp<0  ||  iTmp>255)
		{
			free(szBuf);
			return FALSE;
		}
		LoadPal[i] = iTmp;
		if (i==255)  break;
		while (szBuf[Offset]!='\r'  &&  szBuf[Offset]!='\n')
		{
			Offset++;
			if (Offset>=dwLength)
			{
				free(szBuf);
				return FALSE;
			}
		}
		while (szBuf[Offset]=='\r'  ||  szBuf[Offset]=='\n')
		{
			Offset++;
			if (Offset>=dwLength)
			{
				free(szBuf);
				return FALSE;
			}
		}
		i++;
	}
	free(szBuf);
	return TRUE;
}
